package clarion.acs;
import clarion.system.*;
import clarion.tools.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is a base class for ACS action format and classes implementing
 * different concrete types of ACS actions needs to extend this class.
 *
 * ACS action format - first part is TYPE dimension, second part is SUB-ACTION
 * dimensions, third part is PARAMETER dimensions.
 * NON-TYPE part - Both Second and third part are called NON-TYPE part.
 * Each SUB-ACTION dimension has only one active value at a time.
 * Whether each PARAMETER dimension has one or multiple active values depends
 * on the parameter dimension type.
 * We add one SIGNAL value at the end of each PARAMETER dimension.

 * <li> summary of the important methods. <br>
 * <ol>
 * <li> public void decideActiveDimVals(): Decides active values for all NON-TYPE
 * dimensions.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * This class implemnts the new version of general ACS action format in which
 * multiple active values are allowed in each dimension.
 * <li> Usage of this class. <br>
 * It is instantiated inside an <em>AcsNetComp</em> object.
 * </ul>
 * @author Xi Zhang
 *
 */

public class AcsAction{
   final int ACTION_IDX = 0, PARAM_IDX = 1;
   /** global settings. */
   protected Global global;
   /** # of dimensions of each NON-TYPE part. */
   protected short[] dimNums;
   /** # of allowable values of each dimension of NON-TYPE part. */
   protected short[][] dimValNums;
   /** the full expansion length of NON-TYPE parts in 1-dimensional format. */
   protected int fullLength;
   /** action type : EX, GS, WM, NACS-control. */
   protected int type;
   /** what the action will do in term of a set of subactions. */
   protected short[] subActions;
   /** relevant parameters to the specific action. */
   protected VCGSet[] parameters;
   /** the # of dimensions of NON-TYPE parts. */
   protected int nonTypeDimNum;
   /** the start position of PARAMETER part in the array : activations. */
   protected int paramStart;
   /** currently active values in each dimension of NON-TYPE part. */
   protected short[][] activeDimVals;
   /** rule support for this action. */
   protected double ruleSupport;
   /** the activations of the action. */
   protected double[] activations;

   /** Constructor for initializing <tt>this AcsAction</tt> object using the given
    *  parameters.
    *  @param global the global variables.
    *  @param type action type.
    *  @param fullLength the full length of the NON-TYPE parts of the action in
    *  one dimensional format.
    *  @param dimNums the # of dimesions of each NON-TYPE part.
    *  @param dimValNums the # of allowable values in each dimension of NON-TYPE part. */
   public AcsAction( Global global, int type, int fullLength, short[] dimNums, short[][] dimValNums )
   {
      this.global = global;
      this.type = type;
      this.fullLength = fullLength;
      this.dimNums = dimNums;
      this.dimValNums = dimValNums;
      subActions = new short[ dimNums[ACTION_IDX] ];
      parameters = new VCGSet[ dimNums[PARAM_IDX] ];
      for( int i = 0; i < parameters.length; i++ ){
           // neeed be further revision.
           int dimType = global.getOutputMultiValueProperty( type )[i];
           if( dimType == Global.SINGLE_ACTIVE_VALUED ){
               short[] elemNums = new short[2];
               elemNums[0] = dimValNums[PARAM_IDX][i];
               elemNums[1] = 2;
               parameters[i] = new VCGSet( dimType, 2, elemNums );
           }
           else{ // if dimType is MULTIPLE_ACTIVE_VALUED.
               // vcgNum considers the signal VCG in the end of the dim.
               int vcgNum = dimValNums[PARAM_IDX][i] + 1;
               short[] elemNums = new short[vcgNum];
               for( int j = 0; j < vcgNum; j++ )
                    elemNums[j] = 2;
               parameters[i] = new VCGSet( dimType, vcgNum, elemNums );
           }
      }
      nonTypeDimNum = dimNums[ACTION_IDX] + dimNums[PARAM_IDX];
      paramStart = 0;
      for( int i = 0; i < dimValNums[ACTION_IDX].length; i++ )
           paramStart += dimValNums[ACTION_IDX][i];
      activeDimVals = new short[ nonTypeDimNum ][];
      activations = new double[ fullLength ];
   }

   /** Returns the type.
    *  @return the type. */
   public int getType()
   {
      return type;
   }

   /** Returns the active values.
    *  @return the active values. */
   public short[][] getActiveValues()
   {
      return activeDimVals;
   }

   /** Returns the sub-actions.
    *  @return the sub-actions. */
   public short[] getSubActions()
   {
      return subActions;
   }

   /** Returns the parameters.
    *  @return the parameters. */
   public VCGSet[] getParameters()
   {
      return parameters;
   }

   /** Sets currently active values for all ACS action dimensions.
    *  @param activeVals: currently active values of all ACS action dimensions.
    *  first dimension indicates dimension indices in the ACS action.
    *  second dimension stores all of currently active values in the specific
    *  dimension.
    *  called by TL methods. */
   public void setActiveDimVals( short[][] activeVals )
   {
      activeDimVals = new short[activeVals.length][];
      for( int i = 0; i < activeVals.length; i++ ){
           activeDimVals[i] = new short[ activeVals[i].length ];
           for( int j = 0; j < activeVals[i].length; j++ )
                activeDimVals[i][j] = activeVals[i][j];
      }
      for( int i = 0; i < dimNums[ACTION_IDX]; i++ )
           subActions[i] = activeDimVals[i][0];
      for( int i = 0; i < parameters.length; i++ )
           parameters[i].setActiveValues( activeVals[ dimNums[ACTION_IDX] + i ] );
   }

   /** Sets activations for this ACS action.
    *  @param activations: current activations calculated from BL level.
    *  called by BL methods, the input 'activations' include SIGNAL values. */
   public void setActivations( double[] activations )
   {
      this.activations = activations;
      double[][] paramActivations = new double[ dimNums[PARAM_IDX] ][];
      int offset = paramStart;
      for( int i = 0; i < dimNums[PARAM_IDX]; i++ ){
           int length = parameters[i].getFullLength();
           paramActivations[i] = new double[ length ];
           for( int j = 0; j < length; j++ )
                paramActivations[ i ][ j ] = activations[ offset + j ];
           parameters[i].setActivations( paramActivations[i] );
           offset += length;
      }
   }

   /** Decides active values for all NON-TYPE dimensions.
    *  called after activations is set by BL methods. */
   public void decideActiveDimVals()
   {
      activeDimVals = new short[ nonTypeDimNum ][];
      // decide the active values from SUB-ACTION dimensions.
      StochasticDecider decider = new StochasticDecider( 0.1 );
      int start = 0;
      for( int i = 0; i < dimNums[ACTION_IDX]; i++ ){
           activeDimVals[i] = new short[1];
           int end = start + dimValNums[ACTION_IDX][i];
           activeDimVals[i][0] = (short)decider.getRecommendation( activations, start, end );
           start = end;
      }
      // decide the active values from PARAMETER dimensions.
      int dimOffset = dimNums[ACTION_IDX];
      for( int i = 0; i < dimNums[PARAM_IDX]; i++ )
           activeDimVals[dimOffset + i] = parameters[i].decideActiveValues();
   }
}