package clarion.acs.gs;
import java.util.Vector;
import clarion.system.*;
import clarion.common.*;
import clarion.tools.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is to implement one concrete goal structure: goal list described in
 * CLARION model. It extends the abstract class <em>GoalStructure</em>.
 * <li> summary of the important methods. <br>
 * <ol>
 * <li> public boolean actionSet( int idx ): the set action.
 * <li> public boolean actionReset( int idx ): the reset action.
 * <li> public boolean goalCompete(): the process of goal competition to be
 * active.
 * <li> public void updateState( double[] state ): updates current goal state as
 * input to goal aciton-decision network.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * This class implements the concept of goal list which is more ideal for simulating
 * human's cognitive process.
 * <li> Usage of this class. <br>
 * It is instantiated inside <em>GSNet</em> object.
 * </ul>
 * @author Xi Zhang
 *
 */

public class GoalList implements GoalStructure {

  /** global settings. */
  protected Global global;
  /** task-specific clarion settings. */
  protected TaskClarion taskClarion;
  /** net index. */
  protected int    netIdx;
  /** current active goal index. */
  protected int    curActiveGoalIdx;
  /** all of the possible goals. */
  protected short  goalReps[][];
  /** each parameter dimension of a goal. */
  protected short  gsParamDVs[];
  /** size of goal chunk. */
  protected int    size_gc;
  /** size of goal structure. */
  protected int    size_gs;
  /** GS action by top level only or by both. */
  protected int    gsActionMadeMode;
  /** goal threshold. */
  protected double threshold_GS;
  /** BLA option: constant or recency-based. */
  protected int    gsBlaOption;
  /** BLA of GS. */
  protected double gsBla;
  /** parameters of recency-based BLA. */
  protected double gsBlaParams[];
  /** current time for calculating BLA. */
  protected long   currentTime;
  /** coordination of external vs goal actions. */
  protected double P_GS;
  /** coordination of external vs goal actions. */
  protected double P_EXT;
  /** recording current existing goals. */
  protected GoalChunk[] goalItems;
  /** used for selecting active goal. */
  protected StochasticDecider decider;

  /** constructor.
   *  @param g the global settings. */
  public GoalList( Global g )
  {
      global = g;
      taskClarion = global.getTaskClarion();
      netIdx = global.getGsNetIdx();
      size_gc = global.getGoalChunkSize();
      size_gs = global.getGoalStructSize();
      gsActionMadeMode = global.getGoalDecisionMode();
      threshold_GS = global.getGoalActivationThreshold();
      gsBlaOption = global.getGoalBlaOption();
      if( gsBlaOption == global.CONSTANT_OPTION )
          gsBla = global.getGoalBla();
      else
          gsBlaParams = global.getGoalBlaParas();
      P_GS = global.getGoalCorConstant();
      P_EXT = global.getExternCorConstant();
      goalReps = taskClarion.getGoals();
      gsParamDVs = taskClarion.getGoalDimDVs();
      // initialize all of the goal items.
      goalItems = new GoalChunk[ size_gs ];
      for( int i = 0; i < size_gs; i++ ){
           Feature[] features = new Feature[ size_gc + 1 ];
           for( int j = 0; j < size_gc + 1; j++ )
                if( j < size_gc ){    // parameter dimensions.
                    short[] con = new short[ gsParamDVs[j] ];
                    for( int k = 0; k < con.length; k++ )
                         con[k] = 0;
                    con[goalReps[i][j]] = 1;   // goal parameter value in this dimension.
                    features[j] = new Feature( "goal param dim " + j, (short)Global.FT_GS,
                                               (short)j, con, global );
                }
                else{ // goal dimension.
                    short[] con = new short[ size_gs ];
                    for( int k = 0; k < size_gs; k++ )
                         con[k] = 0;
                    con[i] = 1;    // the ith goal.
                    features[j] = new Feature( global.getFeatureNames(Global.FT_GS, j), (short)Global.FT_GS,
                                               (short)j, con, global );
                }
           goalItems[ i ] = new GoalChunk( Global.CHK_GOAL, "goal item " + i,
                                           features, global );
      }
      decider = new StochasticDecider( 0.1 );
      global.setGS( this );
  }

  /** Returns if this goal stack is empty or not. */
  public boolean isEmpty()
  {
     boolean emptyFlag = true;
     for( int i = 0; i < goalItems.length; i++ )
          if( !goalItems[i].isRemoved() ){
              emptyFlag = false;
              break;
          }
     return emptyFlag;
  }

  /** Returns the number of goals in this goal stack. */
  public int getSize()
  {
     return size_gs;
  }

  /** Returns the currently active goal in terms of goal index. */
  public int getActiveGoal()
  {
     int currentGoal = getActiveItem().getFeature(size_gc).getValues()[0];
     return currentGoal;
  }

  /** Returns the currently active goal item in terms of a chunk. */
  public GoalChunk getActiveItem()
  {
      return goalItems[ curActiveGoalIdx ];
  }

  /** Returns the goal threshold.
   *  @return the goal threshold. */
  public double getThreshlod()
  {
     return threshold_GS;
  }

  /** Sets current time.
   *  @param time the time to set. */
  public void setCurrentTime( long time )
  {
     currentTime = time;
  }

  /** 'set' action.
   *  @param idx goal idx. */
  public boolean actionSet( int idx )
  {
     GoalChunk gc = goalItems[ idx ];
     if( gc.isRemoved() )
         return false;
     if( !gc.isActivated() )
         gc.setActivated( true );
     gc.increaseUsage( currentTime );
     return true;
  }

 /** 'reset' action.
   *  @param idx goal idx. */
  public boolean actionReset( int idx )
  {
     GoalChunk gc = goalItems[ idx ];
     gc.setRemoved( true );
     return true;
  }

  /** routine of no-goal-action. */
  public boolean actionNoGoal()
  {
     return false;
  }

  /** The process of goal competition to be active. */
  public boolean goalCompete()
  {
     int activatedGoalNum = 0;
     int indices[] = new int[ size_gs ];
     double inputs[] = new double[ size_gs ];
     for( int i = 0; i < goalItems.length; i++ ){
          GoalChunk gc = goalItems[ i ];
          if( gc.isRemoved() || !gc.isActivated() )
              continue;
          gc.calcBLA( currentTime );
          double bla = gc.getBLA();
          if( bla > threshold_GS ){
              inputs[ activatedGoalNum ] = bla;
              indices[ activatedGoalNum++ ] = i;
          }
     }
     if( activatedGoalNum == 0 ){
         if( curActiveGoalIdx == global.EMPTY )
             return false;
         else
             return true;
     }
     int idx = decider.getRecommendation( inputs, activatedGoalNum );
     if( curActiveGoalIdx == indices[ idx ] )
         return false;
     else{
         curActiveGoalIdx = indices[ idx ];
         return true;
     }
  }

 /** routine when a goal is completed, no automatic goal pop.
   *  @param popFlag indicate a pop action is done or not. */
  public void goalFinished( boolean popFlag )
  {
     GoalChunk gc = goalItems[ curActiveGoalIdx ];
     gc.setRemoved( true );
  }

  /** Updates current goal state as input to GSNet.
   *  @param state current state to update. */
  public void updateState( double[] state )
  {
     goalItems[ curActiveGoalIdx ].getContent( state );
  }
}