package clarion.acs.gs;
import  java.util.Vector;
import  java.util.Stack;
import clarion.system.*;
import clarion.common.*;
import clarion.tools.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is to implement one concrete goal structure: goal stack described in
 * CLARION model. It extends the abstract class <em>GoalStructure</em>.
 * <li> summary of the important methods. <br>
 * <ol>
 * <li> public boolean actionSet( int idx ): the push action.
 * <li> public boolean actionReset( int idx ): the pop action.
 * <li> public void updateState( double[] state ): updates current goal state as
 * input to goal aciton-decision network.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * This class implements the concept of goal stack which is in the center of
 * CLARION model.
 * <li> Usage of this class. <br>
 * It is instantiated inside <em>GSNet</em> object.
 * </ul>
 * @author Xi Zhang
 *
 */

public class GoalStack implements GoalStructure {

  /** global settings. */
  protected Global global;
  /** task-specific clarion settings. */
  protected TaskClarion taskClarion;
  /** net index. */
  protected int    netIdx;
  /** current active goal index. */
  protected int    curActiveGoalIdx;
  /** representations of all of the possible goals. */
  protected short  goalReps[][];
  /** d-v info of parameter dimensions. */
  protected short  gsParamDVs[];
  /** size of goal chunk, actually the # of parameter dims. */
  protected int    size_gc;
  /** size of goal structure. */
  protected int    size_gs;
  /** GS action by top level only or by both. */
  protected int    gsActionMadeMode;
  /** coordination of external vs goal actions. */
  protected double P_GS;
  /** coordination of external vs goal actions. */
  protected double P_EXT;
  /** the goal items. */
  protected GoalChunk goalItems[];
  /** stack for goals. */
  protected Stack  stack;
  /** used for selecting active goal. */
  protected StochasticDecider decider;

  /** constructor.
   *  @param g the global settings. */
  public GoalStack( Global g )
  {
      global = g;
      taskClarion = global.getTaskClarion();
      netIdx = global.getGsNetIdx();
      // actually the # of parameter dims.
      size_gc = global.getGoalChunkSize();
      // the # of goals.
      size_gs = global.getGoalStructSize();
      // representaions of all of the possible goals.
      goalReps = taskClarion.getGoals();
      // d-v info of all of parameter dimensions.
      gsParamDVs = taskClarion.getGoalDimDVs();
      gsActionMadeMode = global.getGoalDecisionMode();
      P_GS = global.getGoalCorConstant();
      P_EXT = global.getExternCorConstant();
      // initialize all of the goal items.
      goalItems = new GoalChunk[ size_gs ];
      for( int i = 0; i < size_gs; i++ ){
           Feature[] features = new Feature[ size_gc + 1 ];
           for( int j = 0; j < size_gc + 1; j++ )
                if( j < size_gc ){    // parameter dimensions.
                    short[] con = new short[ gsParamDVs[j] ];
                    for( int k = 0; k < con.length; k++ )
                         con[k] = 0;
                    con[goalReps[i][j]] = 1;   // goal parameter value in this dimension.
                    features[j] = new Feature( "goal param dim " + j, (short)Global.FT_GS,
                                               (short)j, con, global );
                }
                else{ // goal dimension.
                    short[] con = new short[ size_gs ];
                    for( int k = 0; k < size_gs; k++ )
                         con[k] = 0;
                    con[i] = 1;    // the ith goal.
                    features[j] = new Feature( global.getFeatureNames(Global.FT_GS, j), (short)Global.FT_GS,
                                               (short)j, con, global );
                }
           goalItems[ i ] = new GoalChunk( Global.CHK_GOAL, "goal item " + i,
                                           features, global );
      }
      stack = new Stack();
      //stack.setSize( size_gs );
      decider = new StochasticDecider( 0.1 );
      global.setGS( this );
  }

  /** Returns if this goal stack is empty or not. */
  public boolean isEmpty()
  {
     return stack.empty();
  }

  /** Returns the number of goals in this goal stack. */
  public int getSize()
  {
     return size_gs;
  }

  /** Returns the currently active goal in terms of goal index. */
  public int getActiveGoal()
  {
     if( stack.empty() )
         return -1;
     int currentGoal = getActiveItem().getFeature(size_gc).getValues()[0];
     return currentGoal;
  }

  /** Returns the currently active goal item in terms of a chunk. */
  public GoalChunk getActiveItem()
  {
     if( stack.empty() )
         return null;
     int idx = ((Integer)stack.peek()).intValue();
     return goalItems[idx];
  }

  /** 'set' action. in goal stack, it is a  push action.
   *  @param idx the index of a goal. */
  public boolean actionSet( int idx )
  {
     if( !stack.empty() && curActiveGoalIdx == idx )
         return false;
     curActiveGoalIdx = idx;
     stack.push( new Integer( curActiveGoalIdx ) );
     return true;
  }

  /** 'reset' action. in goal stack, it is a pop action.
   *  @param idx the index of a goal. */
  public boolean actionReset( int idx )
  {
     if( stack.empty() ){
         curActiveGoalIdx = Global.EMPTY;
         return false;
     }
     curActiveGoalIdx = ((Integer)( stack.pop() )).intValue();
     return true;
  }

  /** routine of no-goal-action. */
  public boolean actionNoGoal()
  {
     return false;
  }

  /** routine when a goal is completed, no automatic goal pop.
   *  @param popFlag indicate a pop action is done or not. */
  public void goalFinished( boolean popFlag )
  {
     if( popFlag )
         actionReset( 0 );
  }

  /** Updates current goal state as input to GSNet.
   *  @param state current state to update. */
  public void updateState( double[] state )
  {
     goalItems[ curActiveGoalIdx ].getContent( state );
  }
}