package clarion.acs.net;
import clarion.acs.*;
import clarion.system.*;
import clarion.common.*;
import clarion.nacs.*;
import clarion.tools.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is to implement Implicit Decision Network. It extends the more
 * generic class <em>AcsNetComp</em>. In this class, it contains one <em>BPNet</em>
 * object which implements the backpropagation network.
 * <li> summary of the important methods. <br>
 * <ol>
 * <li> public short[][] decideAction(): decide action by this IDN.
 * <li> public void reasoning(): the reasoning routine.
 * <li> public void update( double reward ): update <tt>this</tt> IDN with currently
 *  received reward.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * This class is to implement Implicit Decision Network, the bottom level of an
 * ACS action-decision network.
 * <li> Usage of this class. <br>
 * It is instantiated inside an <AcsANet> object.
 * @author Xi Zhang
 *
 */

public class IDN extends AcsNetComp{

  /** used for filling in Q value arrays. */
  final int MODE_Q1 = 0, MODE_Q2 = 1;
  /** learning option: q-learning, simplified q, BP, fix or by MCS. */
  protected int learnOption;
  /** variables for the net. */
  protected BPNet net;
  /** current state including EX, GS, WM. */
  protected double[] state;
  /** current desired output. */
  protected Feature desiredOutput[];
  /** VCGSets for action-deciion making. */
  protected VCGSet[][] vcgSets;
  /** the # of each type of output:  EX, WM, GS, NACS-CONTROL if any. */
  protected short[] outputNums;
  /** used to store the desired output converted from chosen action. */
  protected double[] overallOutput;
  /** used to locate each type of output in the overall outputs.
   *  EX, WM, GS, NACS-CONTROL if any. */
  protected short[] outputLocations;
  /** Q(x,a)s and maxQ(y,b)s (with multiple dims). */
  protected double qValue1, qValue2;
  /** Q(x,a)s and maxQ(y,b)s (with multiple dims).
   *  the first dim: ACTION_NORMAL OR ACTION_CONTROL.
   *  the second dim: each output dim.
   *  the third dim: the activations of active values in one output dim. */
  protected double qValue1s[][][], qValue2s[][][];
  /** used for updating PM and NM. */
  protected boolean positivity;
  /** current time used for computing BLA. */
  protected int currentTime;
  /** constant used for MCS's minotoring conclusions. */
  protected int c24;
  /** constant used for MCS's minotoring conclusions. */
  protected Chunk[] blResults;

  /** constructor for initializing <tt>this</tt> IDN with inputNum, network index and
   *  the associated <tt>Global</tt> object. */
  public IDN( int subsysIdx, int inputNum, int nIdx, Global g )
  {
     this.subsysIdx = subsysIdx;
     INPUT_NUM = inputNum;
     netIdx = nIdx;
     global = g;
     taskClarion = global.getTaskClarion();
     learnOption = global.getIdnLearningOption( netIdx );
     if( learnOption == Global.BP )
         net = new BPNet( inputNum, netIdx, global, Global.ACS );
     else
         net = new QBPNet( inputNum, netIdx, global );
     state = new double[ INPUT_NUM ];
     //OUTPUT_NUM = global.getIdnOutputNum( netIdx );
     OUTPUT_DIM_NUM = global.getIdnOutputDimNum( netIdx );
     outputDVs = global.getIdnOutputDimVals( netIdx );
     outputOffsets = new short[OUTPUT_DIM_NUM];
     CommonMethods.calcOffsets( outputOffsets, outputDVs );
     if( subsysIdx == Global.ACS ){
         netType = global.isExNet(netIdx) ? Global.EX_TYPE :(global.isWmNet(netIdx) ? Global.WM_TYPE : Global.GS_TYPE );
         FULL_OUTPUT_NUM = global.getAcsPhysicOutputNum( netIdx );
         FULL_OUTPUT_DIM_NUM = global.getAcsPhysicOutputDimNum( netIdx );
         fullOutputDVs = global.getAcsPhysicOutputDimVals( netIdx );
         fullOutputOffsets = new short[fullOutputDVs.length];
         CommonMethods.calcOffsets( fullOutputOffsets, fullOutputDVs );
         // initialize variables for locations of each type of output in overall outputs.
         outputNums = global.getAcsPhysicOutputNums();
         outputLocations = new short[ outputNums.length ];
         CommonMethods.calcOffsets( outputLocations, outputNums );
     }
     else{  // MCS.
     }
     CONTROL_OUTPUT_NUM = global.getNacsControlOutputNum();
     CONTROL_OUTPUT_DIM_NUM = Global.ACT_DIM_NUM_NACS_CONTROL;
     controlOutputDVs = global.getNacsControlOutputDimVals();
     controlOutputOffsets = new short[CONTROL_OUTPUT_DIM_NUM];
     CommonMethods.calcOffsets( controlOutputOffsets, controlOutputDVs );
     FULL_CONTROL_OUTPUT_NUM = global.getNacsControlPhysicOutputNum();
     FULL_CONTROL_OUTPUT_DIM_NUM = Global.ACT_DIM_NUM_NACS_CONTROL;
     fullControlOutputDVs = global.getNacsControlPhysicOutputDimVals();
     fullControlOutputOffsets = new short[FULL_CONTROL_OUTPUT_DIM_NUM];
     CommonMethods.calcOffsets( fullControlOutputOffsets, fullControlOutputDVs );
     chosenAction = new short[ 2 ][][];
     chosenAction[Global.ACTION_NORMAL] = new short[ OUTPUT_DIM_NUM ][];
     if( global.isNacsOn() && subsysIdx == Global.ACS )
         chosenAction[Global.ACTION_CONTROL] = new short[ CONTROL_OUTPUT_DIM_NUM ][];
     suggestedAction = new short[ 2 ][][];
     suggestedAction[Global.ACTION_NORMAL] = new short[ OUTPUT_DIM_NUM ][];
     if( global.isNacsOn() && subsysIdx == Global.ACS )
         suggestedAction[Global.ACTION_CONTROL] = new short[ CONTROL_OUTPUT_DIM_NUM ][];
     fullSuggestedAction = new short[ 2 ][][];
     fullSuggestedAction[Global.ACTION_NORMAL] = new short[ FULL_OUTPUT_DIM_NUM ][];
     if( global.isNacsOn() && subsysIdx == Global.ACS )
         fullSuggestedAction[Global.ACTION_CONTROL] = new short[ FULL_CONTROL_OUTPUT_DIM_NUM ][];
     outputActivations = new double[ 2 ][];
     outputActivations[Global.ACTION_NORMAL] = new double[ FULL_OUTPUT_NUM ];
     if( global.isNacsOn() && subsysIdx == Global.ACS )
         outputActivations[Global.ACTION_CONTROL] = new double[ FULL_CONTROL_OUTPUT_NUM ];
     vcgSets = new VCGSet[2][];
     vcgSets[Global.ACTION_NORMAL] = createVCGSets( netType );
     if( global.isNacsOn() && subsysIdx == Global.ACS )
         vcgSets[Global.ACTION_CONTROL] = createVCGSets( Global.CONTROL_TYPE );
     // initialize the overall output.
     overallOutput = new double[ net.getnOutput() ];
     // initialize q-values arrays.
     qValue1s = new double[2][][];
     qValue1s[Global.ACTION_NORMAL] = this.createQvalues( Global.ACTION_NORMAL );
     if( global.isNacsOn() && subsysIdx == Global.ACS )
         qValue1s[Global.ACTION_CONTROL] = this.createQvalues( Global.ACTION_CONTROL );
     qValue2s = new double[2][][];
     qValue2s[Global.ACTION_NORMAL] = this.createQvalues( Global.ACTION_NORMAL );
     if( global.isNacsOn() && subsysIdx == Global.ACS )
         qValue2s[Global.ACTION_CONTROL] = this.createQvalues( Global.ACTION_CONTROL );
     // initialize the conclusions for monitoring.
     if( global.isMcsOn() ){
         int c24 = global.getMonitorAcsBottomConcNum();
         blResults = new Chunk[ c24 ];
     }
     //temperature = global.getIdnActionSelectionTemp( netIdx );
     //actionDecider = new StochasticDecider( OUTPUT_NUM, temperature );
  }

   /** Returns all the activated actions' activations with its full-bottom format.
    *  @returns the activations with its full-bottom format. */
   public double[][] getOutputActivations()
   {
       net.getOutput( outputActivations[Global.ACTION_NORMAL], outputLocations[netType], outputNums[netType] );
       if( subsysIdx == Global.ACS && global.isNacsOn() )
           net.getOutput( outputActivations[Global.ACTION_CONTROL], outputLocations[Global.CONTROL_TYPE],
                          outputNums[Global.CONTROL_TYPE] );
       return  CommonMethods.copy( outputActivations );
   }

   /** Returns all the activated actions' activations.
    *  @param arr the array to store the activations. */
   public void getOutputActivations( double[][] arr )
   {
      double[][] activations = getOutputActivations();
      for( int i = 0; i < arr.length; i++ )
           arr[ i ] = activations[ i ];
   }

  /** Returns the outputs.
   *  @return the outputs. */
  public void getOutput( double[] output )
  {
     net.getOutput( output );
  }

  /** Returns the success rate.
   *  @return the success rate. */
  public double getSuccRate()
  {
     return net.getSuccRate();
  }

  /** Returns the number of positive matches.
   *  @return the number of positive matches. */
  public double getPM()
  {
     return net.getPM();
  }

  /** Returns the number of negative matches.
   *  @return the number of negative matches. */
  public double getNM()
  {
     return net.getNM();
  }

  /** Returns all conclusions by this IDN in terms of chunks.
   *  @return all conclusions. */
  public Chunk[] getAllConclusions()
  {
     Chunk[] chunks = new Chunk[1];
     int type = Global.CHK_ACTION;
     String label = Global.CHUNK_TYPE_LABELS[type];
     Feature[] featList = new Feature[ OUTPUT_DIM_NUM ];
     for( int i = 0; i < OUTPUT_DIM_NUM; i++ ){
          int typeIdx = global.getIdnOutputFeatureInfo( netIdx )[i][0];
          int featIdx = global.getIdnOutputFeatureInfo( netIdx )[i][1];
          String featLabel = global.getFeatureNames( typeIdx, featIdx );
          int featLen = global.getFeatureLength( typeIdx, featIdx );
          short[] values = CommonMethods.copy( suggestedAction[Global.ACTION_NORMAL][i] );
          featList[i] = new Feature( featLabel, (short)typeIdx, (short)featIdx, (short)featLen,
                                     values, global );
     }
     chunks[0] = new Chunk( type, label, featList, global );
     return chunks;
  }

  /** Returns the response time given Bottom Level Activation Time.
   *  @param BLAT Bottom Level Activation Time.
   *  @return the response time. */
  public int getResponseTime( int BLAT )
  {
     calcRT( BLAT );
     return net.getResponseTime();
  }

  /** Returns <tt>true</tt> if the specified rule is positive
   *  <tt>false</tt>, otherwise.
   *  @param ruleType rule type such as RER, IRL or FR.
   *  @threshold threshold for checking positivity.
   *  @return if the specified rule is positive or not.  */
  public boolean getPositivity( int ruleType, double threshold )
  {
     boolean posFlag = false;
     double reward = (net instanceof QBPNet)?((QBPNet)net).getReinforcement():0.0;
     double qDiscount = (net instanceof QBPNet)?((QBPNet)net).getQDiscount():0.0;
     if( ruleType == Global.RER )
         posFlag = taskClarion.getRerPositivity(learnOption,reward,qDiscount,qValue2,qValue1,threshold);
     if( ruleType == Global.IRL )
         posFlag = taskClarion.getIrlPositivity(learnOption,reward,qDiscount,qValue2,qValue1,threshold);
     if( ruleType == Global.FR )
         posFlag = taskClarion.getFrPositivity(learnOption,reward,qDiscount,qValue2,qValue1,threshold);
     return posFlag;
     /*
     if( learnOption == Global.QL ) {
         QBPNet qNet = (QBPNet)net;
         if( qNet.getReinforcement() + qNet.getQDiscount() * qValue2 - qValue1 > value )
             positivity = true;
         else
             positivity = false;
     }
     else if( learnOption == Global.SQL )
          {
              QBPNet qNet = (QBPNet)net;
              if( qNet.getReinforcement() > value )
                  positivity = true;
              else
                  positivity = false;
          }
          else {
              if( value > 0 )
                  positivity = true;
              else
                  positivity = false;
          }
     return positivity;
     */
   }


  /** Sets the states.
   *  @param exState external state.
   *  @param gsState goal structure state.
   *  @param wmState working memory state. */
  public void setStates( double[] state, double[] exState, double[] gsState, double[] wmState )
  {
      for( int i = 0; i < INPUT_NUM; i++ )
           this.state[ i ] = state[ i ];
      if( learnOption != Global.QL )
          net.setInput( this.state );
      else
          ((QBPNet)net).setState( this.state );
  }

  /** Sets the state chunk as input to this IDN.
   *  @param chunk the state chunk as input to this IDN. */
  public void setState( Chunk chunk )
  {
     int dimNum = chunk.getDimNum();
     int offset = 0;
     for( int i = 0; i < dimNum; i++ ) {
          short[] arr = chunk.getFeature(i).getContent();
          for( int j = 0; j < arr.length; j++ )
               state[ offset + j ] = arr[ j ];
          offset += arr.length;
     }
     if( learnOption != Global.QL )
         net.setInput( state );
     else
         ((QBPNet)net).setState( state );
  }

   /** Sets the chosen action.
    *  @param action the action to fill in the chosen action.
    *  In new design, this parameter has full # of output dimensions.
    *  */
   public void setChosenAction( short[][][] fullAction )
   {
      if( subsysIdx == Global.ACS && !global.isExNet( netIdx ) ){
           chosenAction = CommonMethods.copy( fullAction );
           return;
      }
      // convert the action with fill format into the partial format.
      short[][] dimFeatInfo = global.getIdnOutputFeatureInfo( netIdx );
      for( int i = 0; i <  OUTPUT_DIM_NUM; i++ ){
           int idx = dimFeatInfo[i][1];
            // don't need relocation.
            if( subsysIdx == Global.ACS && netType != Global.EX_TYPE )
                idx = i;
           chosenAction[Global.ACTION_NORMAL][i] = CommonMethods.copy(fullAction[Global.ACTION_NORMAL][idx]);
      }
    if( subsysIdx == Global.ACS && global.isNacsOn() )
      chosenAction[Global.ACTION_CONTROL] = CommonMethods.copy(fullAction[Global.ACTION_CONTROL]);
   }

  /** Sets current desired output. It is used for updating IDN with BP learning.
   *  @param desiredOutput the desired output to set. */
  public void setCurDesiredOutput( Feature[] desiredOutput )
  {
     net.setDesiredOutput( desiredOutput );
  }

  /** Sets current time.
   *  @param curTime current time to set. */
  public void setCurrentTime( int curTime )
  {
     currentTime = curTime;
  }

  /** the reasoning routine. */
  public void reasoning()
  {
     net.forwardPass();
  }

  /** decide action with its full format. If <tt>this</tt> methods is called, that
   *  means IDN decides the action at current step.
   *  @return the decided ation by this IDN. */
  public short[][][] decideAction()
  {
     this.getOutputActivations();
     fillInVCGSets( Global.ACTION_NORMAL );
     for( int i = 0; i < FULL_OUTPUT_DIM_NUM; i++ )
          fullSuggestedAction[Global.ACTION_NORMAL][i] = vcgSets[Global.ACTION_NORMAL][i].decideActiveValues();
     if( subsysIdx == Global.ACS && global.isNacsOn() ){
         fillInVCGSets( Global.ACTION_CONTROL );
         for( int i = 0; i < FULL_CONTROL_OUTPUT_DIM_NUM; i++ )
              fullSuggestedAction[Global.ACTION_CONTROL][i] = vcgSets[Global.ACTION_CONTROL][i].decideActiveValues();
     }
     // retrieve the partial suggested action.
     suggestedAction[Global.ACTION_CONTROL] = CommonMethods.copy( fullSuggestedAction[Global.ACTION_CONTROL] );
     if( subsysIdx == Global.ACS && !global.isExNet( netIdx ) )
         suggestedAction[Global.ACTION_NORMAL] = CommonMethods.copy( fullSuggestedAction[Global.ACTION_NORMAL] );
     else
         for( int i = 0; i < OUTPUT_DIM_NUM; i++ ){
              int idx = global.getIdnOutputFeatureInfo( netIdx )[i][1];
              suggestedAction[Global.ACTION_NORMAL][i] = CommonMethods.copy(fullSuggestedAction[Global.ACTION_NORMAL][idx] );
         }

     return CommonMethods.copy( fullSuggestedAction );
  }

  /** update <tt>this</tt> IDN with currently received reward. */
  public void update( double reward )
  {
     // qvalues of Q(x,a).
     if( learnOption != Global.BP ){
         fillInVCGSets( Global.ACTION_NORMAL );
         fillInQVals( Global.ACTION_NORMAL, MODE_Q1, chosenAction[Global.ACTION_NORMAL], qValue1s[Global.ACTION_NORMAL] );
         if( subsysIdx == Global.ACS && global.isNacsOn() ){
             fillInVCGSets( Global.ACTION_CONTROL );
             fillInQVals( Global.ACTION_CONTROL, MODE_Q1, chosenAction[Global.ACTION_CONTROL], qValue1s[Global.ACTION_CONTROL] );
        }
     }
     // qvalues of maxQ(y,b)
     if( learnOption == Global.QL ){
         net.forwardPass();
         fillInVCGSets( Global.ACTION_NORMAL );
         fillInQVals( Global.ACTION_NORMAL, MODE_Q2, chosenAction[Global.ACTION_NORMAL], qValue2s[Global.ACTION_NORMAL] );
         if( subsysIdx == Global.ACS && global.isNacsOn() ){
             fillInVCGSets( Global.ACTION_CONTROL );
             fillInQVals( Global.ACTION_CONTROL, MODE_Q2, chosenAction[Global.ACTION_CONTROL], qValue2s[Global.ACTION_CONTROL] );
         }
     }
     // calcluate qvalue1 and qvalue2 used for positivity measurement.
     calcPositiveMeatureParas();
     // update BL network.
     switch( learnOption  ){
        case Global.QL:
             ((QBPNet)net).update( chosenAction, reward, qValue1s, qValue2s );
             break;
        case Global.SQL:
             ((QBPNet)net).update( chosenAction, reward, qValue1s );
             break;
        case Global.BP:
             net.backwardPass();
             break;
     }
  }

  /** update for offline training using AEM. */
  public void update( double reward, Chunk[] nextStates, double[] stateDistribution )
  {
     // qvalue of Q(x,a).
     if( learnOption != Global.BP ){
         fillInVCGSets( Global.ACTION_NORMAL );
         fillInQVals( Global.ACTION_NORMAL, MODE_Q1, chosenAction[Global.ACTION_NORMAL], qValue1s[Global.ACTION_NORMAL] );
         if( subsysIdx == Global.ACS && global.isNacsOn() ){
             fillInVCGSets( Global.ACTION_CONTROL );
             fillInQVals( Global.ACTION_CONTROL, MODE_Q1, chosenAction[Global.ACTION_CONTROL], qValue1s[Global.ACTION_CONTROL] );
        }
     }
     // qvalue of maxQ(y,b)
     if( learnOption == Global.QL ){
         net.forwardPass();
         fillInVCGSets( Global.ACTION_NORMAL );
         fillInQVals( Global.ACTION_NORMAL, MODE_Q2, chosenAction[Global.ACTION_NORMAL], qValue2s[Global.ACTION_NORMAL] );
         if( subsysIdx == Global.ACS && global.isNacsOn() ){
             fillInVCGSets( Global.ACTION_CONTROL );
             fillInQVals( Global.ACTION_CONTROL, MODE_Q2, chosenAction[Global.ACTION_CONTROL], qValue2s[Global.ACTION_CONTROL] );
         }
     }
     // calcluate qvalue1 and qvalue2 used for positivity measurement.
     calcPositiveMeatureParas();
     // update QL using frequency estimates represented by the AEM.
     if( learnOption == Global.QL ){
           // qvalue of next potential states.
           double[][][][] qValues = new double[nextStates.length][2][][];
           for( int i = 0; i < nextStates.length; i++ ) {
                // set current state.
                setState( nextStates[i] );
                net.forwardPass();
               //((QBPNet)net).getMaxQVals( qValues[i] );
               qValues[i][Global.ACTION_NORMAL] = this.createQvalues( Global.ACTION_NORMAL );
               fillInVCGSets( Global.ACTION_NORMAL );
               fillInQVals( Global.ACTION_NORMAL, MODE_Q2, chosenAction[Global.ACTION_NORMAL],
                            qValues[i][Global.ACTION_NORMAL] );
               if( subsysIdx == Global.ACS && global.isNacsOn() ){
                   qValues[i][Global.ACTION_CONTROL] = this.createQvalues( Global.ACTION_CONTROL );
                   fillInVCGSets( Global.ACTION_CONTROL );
                   fillInQVals( Global.ACTION_CONTROL, MODE_Q2, chosenAction[Global.ACTION_CONTROL],
                                qValues[i][Global.ACTION_CONTROL] );
               }
           }
           for( int i = 0; i < qValues.length; i++ )
             for( int j = 0; j < qValues[i].length; j++ )
               for( int k = 0; k < qValues[i][j].length; k++ )
                  for( int m = 0; m < qValues[i][j][k].length; m++ )
                       qValue2s[j][k][m] += stateDistribution[i] * qValues[i][j][k][m] / qValues.length;
          ((QBPNet)net).update( chosenAction, reward, qValue1s, qValue2s );
     }
     // update SQL.
     if( learnOption == Global.SQL )
         ((QBPNet)net).update( chosenAction, reward, qValue1s );
     // update BP.
     if( learnOption == Global.BP ){
         CommonMethods.resetArray( overallOutput, 0.0 );
         mappingOutput( netType, chosenAction[Global.ACTION_NORMAL] );
         if( subsysIdx == Global.ACS && global.isNacsOn() )
             mappingOutput( Global.CONTROL_TYPE, chosenAction[Global.ACTION_CONTROL] );
         //net.setDesiredOutput( chosenAction );
         net.setDesiredOutput( overallOutput );
         net.backwardPass();
     }
  }

  /** Updates with flag indicating current step is successful or not. */
  public void update( boolean posFlag )
  {
  }

  /** Calculates the response time given BLAT.
   *  @param BLAT Bottom Level Activation Time. */
  public void calcRT( int BLAT )
  {
     net.calcRT( BLAT );
  }

/******************************* private methods ******************************/
  /** overrided if neccessary, using averaging Q values to get qValue2( maxQ(y,b) )
      and qValue1( Q(x,a) ). ( r + Qdiscount * maxQ(y,b) - Q(x,a) ) */
  private void calcPositiveMeatureParas()
  {
     qValue1 = qValue2 = 0.0;
     qValue1 = CommonMethods.averageArray( qValue1s );
     qValue2 = CommonMethods.averageArray( qValue2s );
  }

  /** outputType: EX, WM, GS, NACS-CONTROL. */
  private double[][] createQvalues( int outputType )
  {
     int dimNum = (outputType == Global.ACTION_NORMAL) ? OUTPUT_DIM_NUM : CONTROL_OUTPUT_DIM_NUM;
     short[] oDVs = (outputType == Global.ACTION_NORMAL) ? fullOutputDVs : fullControlOutputDVs;
     short[][] dimFeatInfo = (outputType == Global.ACTION_NORMAL) ? global.getIdnOutputFeatureInfo( netIdx ):
                                          global.getNacsControlFeatureInfo();
     int outType = (outputType == Global.ACTION_NORMAL) ? netType: Global.CONTROL_TYPE;
     double[][] qVals = new double[dimNum][];

     //System.out.println( "outType = " + outType );

     for( int i = 0; i < dimNum; i++ ){
          int idx = dimFeatInfo[i][1];
            // don't need relocation.
          if( subsysIdx == Global.ACS && outType != Global.EX_TYPE )
              idx = i;

          //System.out.println( "idx = " + idx );


          int type = global.getOutputMultiValueProperty( outType )[idx];
          if( type == Global.SINGLE_ACTIVE_VALUED )
              qVals[i] = new double[2];
          else {
              int length = oDVs[idx] / 2;
              qVals[i] = new double[length];
          }
     }
     return qVals;
  }

  /** outputType: ACTION_NORMAL or ACTION_CONTROL. */
  private void fillInQVals( int outputType, int mode, short[][] action, double[][] qVals )
  {
     VCGSet[] sets = vcgSets[outputType];
     int dimNum = (outputType == Global.ACTION_NORMAL) ? OUTPUT_DIM_NUM : CONTROL_OUTPUT_DIM_NUM;
     short[][] dimInfo = (outputType == Global.ACTION_NORMAL) ? global.getIdnOutputFeatureInfo( netIdx ):
                                          global.getNacsControlFeatureInfo();

     for( int i = 0; i < dimNum; i++ ){
          int idx = dimInfo[i][1];
          if( subsysIdx == Global.ACS && !global.isExNet(netIdx) || outputType == Global.ACTION_CONTROL)
              idx = i;
          if( sets[idx].getType() == Global.SINGLE_ACTIVE_VALUED ){
              int vcgNum = sets[idx].getVCGNum();
              if( mode == MODE_Q1){
                  // get the activation of the active value.

                  //if( i >= qVals.length || i >= action.length || idx >= sets.length )
                   //   System.out.println( "outputType = " + outputType + ", i = " + i + ", qVals[i] len = " +
                   //       qVals[i].length + ", action[i] len = " + action[i].length + ", idx = " + idx + ", sets' length = " + sets.length );


                  qVals[i][0] = (action[i]==null)? 0.0 : sets[idx].getActivation( 0, action[i][0] );
                  // get the activtaion of the signal SINGLE_ACTIVE_VALUED.
                  qVals[i][1] = sets[idx].getActivation( 1, Global.SINGLE_ACTIVE_VALUED );
              }
              else{
                  // get the MAX activations.
                  for( int j = 0; j < vcgNum; j++ )
                       qVals[i][j] = sets[idx].getMaxActivation( j );
              }
          }
          else {
            int vcgNum = sets[idx].getVCGNum();
            if( mode == MODE_Q1 ){
                // initialize the q values.
                for( int j = 0; j < vcgNum; j++ )
                     qVals[i][j] = sets[idx].getActivation( j, Global.ELEM_INACTIVE );
                // get the activations of all of the active values.
                if( action[i] != null ){
                    for( int j = 0; j < action[i].length; j++ ){
                         int loc = action[i][j];
                         qVals[i][loc] = sets[idx].getActivation( loc, Global.ELEM_ACTIVE );
                    }
                    // get the activtaion of the signal MULTI_ACTIVE_VALUED.
                    if( action[i].length > 1 )
                        qVals[i][vcgNum-1] = sets[idx].getActivation( vcgNum-1, Global.MULTI_ACTIVE_VALUED );
                    else
                        qVals[i][vcgNum-1] = sets[idx].getActivation( vcgNum-1, Global.SINGLE_ACTIVE_VALUED );
                }
            }
            else{
                  // get the MAX activations.
                  for( int j = 0; j < vcgNum; j++ )
                       qVals[i][j] = sets[idx].getMaxActivation( j );
            }
         }
     }
  }

  private VCGSet[] createVCGSets( int outputType )
  {
     int dimNum = (outputType != Global.CONTROL_TYPE) ? FULL_OUTPUT_DIM_NUM : FULL_CONTROL_OUTPUT_DIM_NUM;
     short[] oDVs = (outputType != Global.CONTROL_TYPE) ? fullOutputDVs  : fullControlOutputDVs;
     short[] offsets = (outputType != Global.CONTROL_TYPE) ? fullOutputOffsets  : fullControlOutputOffsets;
    // short[][] dimInfo = (outputType != Global.CONTROL_TYPE) ? global.getIdnOutputFeatureInfo( netIdx ):
     //                                     global.getNacsControlFeatureInfo();

     //System.out.println( "outputType = " + outputType + ", dimNum = " + dimNum  + ", dimInfo len = " + dimInfo.length );

     VCGSet[] sets = new VCGSet[ dimNum ];
     for( int i = 0; i < dimNum; i++ ){
          int type = global.getOutputMultiValueProperty( outputType )[i];
          if( type == Global.SINGLE_ACTIVE_VALUED ){
              int vcgNum = 2;
              short[] elemNums = new short[vcgNum];
              elemNums[0] = (short)(oDVs[i] - 2);
              elemNums[1] = 2;
              sets[i] = new VCGSet( type, vcgNum, elemNums );
          }
          else {
              int vcgNum = (short)oDVs[i] / 2;
              short[] elemNums = new short[vcgNum];
              for( int j = 0; j < vcgNum; j++ )
                   elemNums[j] = 2;
              sets[i] = new VCGSet( type, vcgNum, elemNums );
          }
     }
     return sets;
  }

  /** outputType: ACTION_NORMAL or ACTION_CONTROL. */
  private void fillInVCGSets( int outputType )
  {
     VCGSet[] sets = vcgSets[outputType];
     int dimNum = (outputType == Global.ACTION_NORMAL) ? FULL_OUTPUT_DIM_NUM : FULL_CONTROL_OUTPUT_DIM_NUM;
     short[] oDVs = (outputType == Global.ACTION_NORMAL) ? fullOutputDVs  : fullControlOutputDVs;
     short[] offsets = (outputType == Global.ACTION_NORMAL) ? fullOutputOffsets  : fullControlOutputOffsets;

     //System.out.println("in IDN dimNum = " + dimNum + ", outputType = " + outputType );
     //System.out.println("in IDN output activation len = " + outputActivations[outputType].length );

     for( int i = 0; i < dimNum; i++ ){
          int length = oDVs[i];
          double[] activations = new double[ length ];

          //System.out.println("in IDN offsets[i] = " + offsets[i] + ", length = " + length );


          System.arraycopy( outputActivations[outputType], offsets[i], activations, 0, length );
          sets[i].setActivations( activations );
     }
  }

  /** outputType: EX_TYPE, WM_TYPE, GS_TYPE or CONTROL_TYPE. */
  private void mappingOutput( int outputType, short[][] action )
  {
     int dimNum = (outputType != Global.CONTROL_TYPE) ? OUTPUT_DIM_NUM : CONTROL_OUTPUT_DIM_NUM;
     short[] oDVs = (outputType != Global.CONTROL_TYPE) ? fullOutputDVs  : fullControlOutputDVs;
     short[] offsets = (outputType != Global.CONTROL_TYPE) ? fullOutputOffsets  : fullControlOutputOffsets;
     short[][] dimInfo = (outputType != Global.CONTROL_TYPE) ? global.getIdnOutputFeatureInfo( netIdx ):
                                          global.getNacsControlFeatureInfo();
     for( int i = 0; i < dimNum; i++ ){
          int idx = dimInfo[i][1];
          if( subsysIdx == Global.ACS && outputType != Global.EX_TYPE )
              idx = i;
          if( global.getOutputMultiValueProperty( outputType )[idx] == Global.SINGLE_ACTIVE_VALUED ){
              for( int j = 0; j < 2; j++ ){
                   int offset = outputLocations[outputType] + offsets[idx] + action[i][0];
                   if( j == 1 )// SINGLE_ACTIVE_VALUED.
                       offset = outputLocations[outputType] + offsets[idx] + oDVs[idx] - 2;
                   overallOutput[ offset ] = 1.0;
              }
          }
          else{
              int length = oDVs[idx] / 2;
              short[] flags = new short[length-1];
              CommonMethods.resetArray( flags, 0 );
              for( int j = 0; j < action[i].length; j++ )
                   flags[action[i][j]] = 1;
              for( int j = 0; j < length-1; j++ ){
                   int offset = outputLocations[outputType] + offsets[idx] + 2*j;
                   if( flags[i] == 1 )
                       offset += Global.ELEM_ACTIVE;
                   overallOutput[ offset ] = 1.0;
              }
              int offset = outputLocations[outputType] + offsets[idx] + 2*length-2;
              if( action[i].length > 1 )
                  offset += Global.MULTI_ACTIVE_VALUED;
              overallOutput[ offset ] = 1.0;
          }
     }
  }
}