package clarion.acs.net;
import java.util.Vector;
import java.util.TreeMap;
import clarion.acs.net.auxiliary.*;
import clarion.system.*;
import clarion.tools.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is to implement IRL (Independent Rule Learning) Rule set. It extends
 * the more generic rule set class <em>TLRuleSet</em>.
 * <li> summary of the important methods. <br>
 * <ol>
 * <li> public void update( boolean successFlag ): Updates with flag indicating
 * current step is successful or not.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * This class implements the IRL rules which is one of the three rule knowledge
 * types in the top level of an ACS action-decision network.
 * <li> Usage of this class. <br>
 * It is instantiated inside an <em>AcsANet</em> object.
 * @author Xi Zhang
 *
 */

public class IrlRuleSet extends TLRuleSet{

   /** decide which IRL subset is used. */
   protected int groupId;
   /** currently used rule set. */
   protected int curRuleSetIdx;
   /** # of IRL rule subsets. */
   protected int SET_NUM;
   /** check if all rules are used up. */
   protected boolean useupFlag;
   /** vector of acitve rules. */
   protected Vector activeRuleList;

  /** constructor.
    *  @param netIdx the network index.
    *  @param global the associated global settings. */
   public IrlRuleSet( int subsysIdx, int netIdx, Global global )
   {
       super( subsysIdx, netIdx, Global.IRL, global );
       groupId = global.getAgentGroupId();
       MAX_AGE = (int)( 1 / global.getIrlDensity( netIdx ) );
       SET_NUM = global.getIrlSetNum( netIdx );
       initNewSet();
   }

   /** Updates with flag indicating current step is successful or not.
    *  @param successFlag the flag indicating current step is successful or not. */
   public void update( boolean successFlag )
   {
      super.update( successFlag );
      // check usefulness of each rule.
      int idx = 0;
      while( idx < activeRuleList.size() ){
           int ruleIdx = ((Integer)activeRuleList.elementAt( idx )).intValue();
           IrlRule rule = (IrlRule)ruleList[ruleIdx];
           if( rule.isUseless() ){

               //System.out.println( "rule set = " + curRuleSetIdx + ", the deleted rule's pm " + rule.getPM()
               //                    + ", the deleted rule's nm " + rule.getNM() );

               ruleList[ruleIdx] = null;
               activeRuleList.remove( idx );
           }
           else{
               idx++;
               if( rule.canGeneralize() )
                   rule.generalize();
               if( rule.canSpecialize() )
                   rule.specialize();

           }
      }
      if( activeRuleList.isEmpty() ){
          if( (++curRuleSetIdx) < SET_NUM )
              initNewSet();
          else if( !useupFlag ){
                   useupFlag = true;
                   System.out.println( "NO IRL RULES!" );
                 }
      }
      // discount statistics.
      trialIdx++;
      for( int i = 0; i < activeRuleList.size(); i++ ){
           idx = ((Integer)activeRuleList.elementAt( i )).intValue();
           ruleList[idx].aging();
           if( trialIdx >= episodeLength )
               ruleList[idx].discountStatistics();
      }
      if( trialIdx >= episodeLength )
          trialIdx = 0;
   }

/*************************** PRIVATE METHODS **********************************/
   /** initialize new rule set. */
   private void initNewSet()
   {
       //System.out.println( "net idx = " + netIdx + ", cur set idx = " + curRuleSetIdx );
       //System.out.println( "set num = " +  global.getIrlRuleNums( netIdx ).length );

       RULE_NUM = global.getIrlRuleNums( netIdx )[ curRuleSetIdx ];
       UTL_FLAG = global.isRuleUtilityOn( netIdx, global.IRL, curRuleSetIdx );
       ruleList = new IrlRule[ RULE_NUM ];
       for( int i = 0; i < RULE_NUM; i++ )
            ruleList[ i ] = new IrlRule( groupId, subsysIdx, netIdx, curRuleSetIdx, i, global );
       activeRuleList = new Vector( RULE_NUM );
       for( int i = 0; i < RULE_NUM; i++ )
            activeRuleList.addElement( new Integer(i) );
       ruleUtls = new double[ RULE_NUM ];
       recommendRules = new short[ RULE_NUM ];
       if( global.isActionChunkBlaOn( netIdx, global.IRL, curRuleSetIdx ) )
           actionBLAoption = global.getAcsChunkBlaOption( netIdx, global.IRL, curRuleSetIdx );
       else
           actionBLAoption = global.EMPTY;
   }
}