package clarion.acs.net.auxiliary;
import java.util.HashSet;
import java.util.Iterator;
import clarion.system.*;
import clarion.acs.net.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is to implement an active rule in RER ruletype in top level of ACS.
 * <li> summary of the important methods. <br>
 * <ol>
 * <li> public void addChildren( ActiveRule activeRule ): adds children into the
 * children list.
 * <li> public RerRule[] removeChildren(): removes the eligible children from
 * the children list.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * In RER rule type, there are two kinds of rules: one is active rules involving in
 * the process of action recommendation and the other is the inactive rules those
 * are in the children list of the active rules. The process of rule merge and
 * specialization may change the active rules and its inactive children. See the
 * tutorial on the details of this structure. This class is used to implement this
 * structure of managing the rules.
 *
 * <li> Usage of this class. <br>
 * It is instantiated inside an <em>RerRuleSet</em> object.
 * @author Xi Zhang
 *
 */

public class ActiveRule {

  /** the active rule in the format of <em>RerRule</em>. */
  protected RerRule rule;
  /** the children list of this rule. */
  protected HashSet children;
  /** the network index. */
  protected int netIdx;
  /** the task-specific CLARION settings. */
  protected TaskClarion taskClarion;

  /** constructor.
   *  @param rule a RER rule.
   *  @param netidx the network index.
   *  @param taskClarion task-specific CLARION settings. */
  public ActiveRule( RerRule rule, int netIdx, TaskClarion taskClarion )
  {
      this.rule = rule;
      children = new HashSet();
      this.netIdx = netIdx;
      this.taskClarion = taskClarion;
  }

  /** Returns the number of children. */
  public int getCount()
  {
      return children.size();
  }

  /** Return this rule. */
  public RerRule getRule()
  {
      return rule;
  }

  /** Sets a rule. */
  public void setRule( RerRule rule )
  {
     this.rule = rule;
  }

  /** Returns the children.
   *  @return the children. */
  public Object[] getChildren()
  {
      return children.toArray();
  }

  /** Returns the positive matches of this rule.
   *  @return the positive matches. */
  public double getPM()
  {
     return rule.getPM();
  }

  /** Returns the negative matches of this rule.
   *  @return the negative matches. */
  public double getNM()
  {
     return rule.getNM();
  }

  /** Returns the overall positive matches based on both this rule and its children.
   *  @return the positive matches. */
  public double getOverallPM()
  {
     double pm = getPM();
     Iterator iter = children.iterator();
     while( iter.hasNext() ){
         RerRule chd = (RerRule)iter.next();
         pm += chd.getPM();
     }
     return pm;
  }

  /** Returns the overall positive matches based on both this rule and its children.
   *  @return the positive matches. */
  public double getOverallNM()
  {
     double pm = getNM();
     Iterator iter = children.iterator();
     while( iter.hasNext() ){
         RerRule chd = (RerRule)iter.next();
         pm += chd.getNM();
     }
     return pm;
  }

  /** Adds children into the children list.
   *  @param activeRule the children to add. */
  public void addChildren( ActiveRule activeRule )
  {
      /*
      System.out.println( "before, added activeRule = \n" + activeRule );
      System.out.println( "before, key rule = \n" + rule );
      System.out.println( "before, children = " );
      Object[] chds = children.toArray();
      for( int i = 0; i < children.size(); i++ )
           System.out.print( chds[i] );
       */

      addChild( activeRule.getRule() );
      if( activeRule.getCount() == 0 )
          return;
      Object[] rules = activeRule.getChildren();
      for( int i = 0; i < rules.length; i++ )
           addChild( (RerRule)rules[i] );

      /*
      System.out.println( "after, children = " );
      chds = children.toArray();
      for( int i = 0; i < children.size(); i++ )
           System.out.print( chds[i] );
      */
  }

  /** Removes the eligible children from the children list.
   *  @return the eligible children. */
  public RerRule[] removeChildren()
  {
     RerRule[] rules = new RerRule[ children.size() ];
     int count = 0;
     Iterator iter = children.iterator();
     while( iter.hasNext() ){
          RerRule chd = (RerRule)iter.next();
          if( rule.cover( chd ) == -2 ){ // one-direction cover.
              if( chd.cover( rule ) == -2 ){  // two-direction cover.
                  iter.remove();
                  double ig = taskClarion.getRerIG( netIdx, rule.getMatchRate()[0], rule.getMatchRate()[1],
                                                    chd.getMatchRate()[0], chd.getMatchRate()[1] );
                  if( ig < 0.0 )
                      this.setRule( chd );
              }
          }
          else{
             rules[count++] = chd;
             iter.remove();
          }
     }
     // fill in the restored children.
     RerRule[] restoreChildren = new RerRule[ count ];
     for( int i = 0; i < count; i++ )
          restoreChildren[i] = rules[i];
     return restoreChildren;
  }

  /** Checks if this rule covers the specific rule or not.
   *  @param otherRule the rule to check.
   *  @return
   *   -2: cover completely;
   *   -1: more than one dim is different
   * >= 0: the different dim index. */
  public int cover( ActiveRule otherRule )
  {
     return rule.cover( otherRule.getRule() );
  }

  /** methods from class <em>Object</em> */
  public String toString()
  {
     String str = "ACTIVE RULE: " + rule;
     Iterator iter = children.iterator();
     while( iter.hasNext() ){
         RerRule chd = (RerRule)(iter.next());
         str += "-----------INACTIVE RULE: " + chd;
     }
     return str;
  }

  /** methods from class <em>Object</em> */
  public int hashCode()
  {
     return 0;
  }

  /** methods from class <em>Object</em> */
  public boolean equals( Object obj )
  {
     if( !(obj instanceof ActiveRule) || obj == null )
         return false;
     return rule.equals( ((ActiveRule)obj).getRule() );
  }

  /** Adds a child into the children list.
   *  @param child the child to add. */
  private boolean addChild( RerRule child )
  {
      if( child == null || rule.equals( child ) ){
          //System.out.println( "the child cannot be added !" );
          return false;
      }
      boolean containFlag = children.contains( child );
      /*
      boolean equalFlag = false;
      for( int i = 0; i < children.size(); i++ )
           if( child.equals( children.toArray()[i] ) ){
               equalFlag = true;
               break;
           }
      if( equalFlag != containFlag ){
          System.out.println( "SOME ERRORS IN HASHSET CONTAIN" );
          System.out.println( "child = " + child );
          System.exit( 1 );
      }
      */
      //if( !equalFlag  )
      if( !containFlag )
          children.add( child );
      else{
          //System.out.println( "child is one of the children, child = \n" + child );

          Iterator iter = children.iterator();
          while( iter.hasNext() ){
              RerRule chd = (RerRule)iter.next();
              if( chd.equals( child ) ){
                  double ig = taskClarion.getRerIG( netIdx, chd.getMatchRate()[0], chd.getMatchRate()[1],
                                                    child.getMatchRate()[0], child.getMatchRate()[1] );
                  if( ig < 0.0 ){
                      iter.remove();
                      children.add( child );
                  }
                  break;
              }
          }
      }
      return true;
  }
}