 package clarion.acs.net.auxiliary;
 import java.util.Arrays;
 import clarion.system.*;
 import clarion.common.*;
 import clarion.tools.*;

 /**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is to represent a generic rule of a particular rule type in top
 * level of ACS.
 * <li> summary of the important methods. <br>
 * <ol>
 * <li> public boolean decideAction( short[] state, short[] exState, short[] gsState,
 * short[] wmState ): Decides an action by this IRL rule.
 * <li> public void generalize(): the generalization routine.
 * <li> public void specialize(): the specialization routine.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * This class implements a generic rule from a particular type in the top level
 * of an ACS action-decision network. A concrete rule from a particular type
 * (rule set) such as RER, IRL and FR will extends it.
 * <li> Usage of this class. <br>
 * It is instantiated inside an <em>TLRuleSet</em> object.
 * @author Xi Zhang
 *
 */

 public class TLRule{
   /** used for udpating statistics. */
   protected final static double DISCOUNT = 0.9;
   /** global settings. */
   protected Global global;
   /** the task-specific CLARION settings. */
   protected TaskClarion taskClarion;
   /** the subsystem index: ACS or MCS. */
   protected int subsysIdx;
   /** the net index. */
   protected int netIdx;
   /** the rule type. */
   protected int ruleType;
   /** the rule subset index. */
   protected int setIdx;
   /** the number of input units. */
   protected int INPUT_NUM;
   /** the number of output units. */
   protected int OUTPUT_NUM;
   /** the number of input dimensions. */
   protected int INPUT_DIM_NUM;
   /** the number of output dimensions. */
   protected int OUTPUT_DIM_NUM;
   /** the numbers of values in the input dimensions. */
   protected short inputDVs[];
   /** the numbers of values in the output dimensions. */
   protected short outputDVs[];
   /** the start location of each input dimension in the overall input with a format
    *  of one dimensional array. */
   protected short inputOffsets[];
   /** the start location of each input dimension in the overall output with a format
    *  of one dimensional array. */
   protected short outputOffsets[];
   /** the length of NACS-control outputs. */
   protected int CONTROL_OUTPUT_NUM;
   /** the # of NACS-control dims. */
   protected int CONTROL_OUTPUT_DIM_NUM;
   /** the dim-val info of NACS-control outputs. */
   protected short[] controlOutputDVs;
   /** the start offsets of each NACS-control dim in an one-dimenisonal array. */
   protected short[] controlOutputOffsets;
   /** the rule condition */
   protected short conditions[][];
   /** the suggested action by this rule. */
   protected short suggestedAction[][][];
   /** input state in the format of dim-vals (First dimension: each input dimension,
    *  Second dimension: the active values in one input dimension). */
   protected short inputConditions[][];
   /** the index to the dimension with difference, used for general match. */
   protected int   DIFF_DIM_IDX;
   /** flag for partial match is on or not. */
   protected boolean partialMatchFlag;
   /** rule support. */
   protected double  ruleSupport;
   /** BLA option: constant or variable. */
   protected int blaOption;
   /** the rule cost. */
   protected double ruleCost;
   /** the rule attributes. */
   protected RuleAttributes attributes;

  /** constructor.
   *  @param netIdx network index.
   *  @param ruleType rule type.
   *  @param setIdx rule subset index.
   *  @param glboal the associated global settings. */
   public TLRule( int subsysIdx, int netIdx, int ruleType, int setIdx, Global global )
   {
       this.subsysIdx = subsysIdx;
       this.netIdx = netIdx;
       this.ruleType = ruleType;
       this.setIdx = setIdx;
       this.global = global;
       taskClarion = global.getTaskClarion();
       INPUT_NUM = global.getARInputNum( netIdx, ruleType );
       OUTPUT_NUM = global.getAROutputNum( netIdx, ruleType );
       INPUT_DIM_NUM = global.getARInputDimNum( netIdx, ruleType );
       OUTPUT_DIM_NUM = global.getAROutputDimNum( netIdx, ruleType );
       inputDVs = global.getARInputDimVals( netIdx, ruleType );
       outputDVs = global.getAROutputDimVals( netIdx, ruleType );
       inputOffsets = new short[ INPUT_DIM_NUM ];
       outputOffsets = new short[ OUTPUT_DIM_NUM ];
       CommonMethods.calcOffsets( inputOffsets, inputDVs );
       CommonMethods.calcOffsets( outputOffsets, outputDVs );
       CONTROL_OUTPUT_NUM = global.getNacsControlOutputNum();
       CONTROL_OUTPUT_DIM_NUM = Global.ACT_DIM_NUM_NACS_CONTROL;
       controlOutputDVs = global.getNacsControlOutputDimVals();
       controlOutputOffsets = new short[CONTROL_OUTPUT_DIM_NUM];
       CommonMethods.calcOffsets( controlOutputOffsets, controlOutputDVs );
       suggestedAction = new short[ 2 ][][];
       suggestedAction[Global.ACTION_NORMAL] = new short[ OUTPUT_DIM_NUM ][];
       if( subsysIdx == Global.ACS && global.isNacsOn() )
           suggestedAction[Global.ACTION_CONTROL] = new short[ CONTROL_OUTPUT_DIM_NUM ][];
       inputConditions = new short[ INPUT_DIM_NUM ][];
       if( global.isRuleBlaOn( netIdx, ruleType, setIdx ) )
           blaOption = global.getActionRuleBlaOption( netIdx, ruleType, setIdx );
       else
           blaOption = global.EMPTY;
       // initialize rule attributes.
       attributes = new RuleAttributes( netIdx, ruleType, setIdx, global );
       partialMatchFlag = global.isRulePartMatchOn( netIdx, ruleType, setIdx );
   }

/*###############################################################################################
   This method and several others need be revised, DIM INFOS of state(IDN input) may be different
   from those of rules since DIMS of rules are a subset of DIMS of IDN.
################################################################################################*/
   /** Returns the number of dimensions current state mismatch the rule condition and
    *  calculate the rule support. Also get the index of mismatched dimension if only
    *  one dimension is mismatched.
    *  @param input current input.
    *  @return the # of mismatched dimensions.  */
   public int isConditionMatched( short[] input )
   {
      convertToInputConditions( input );
      DIFF_DIM_IDX = -1;
      ruleSupport = 0.0;
      double[] activations = new double[INPUT_DIM_NUM];
      for( int i = 0; i < INPUT_DIM_NUM; i++ )
           activations[i] = 0.0;
      int misMatchNum = 0;
      for( int i = 0; i < INPUT_DIM_NUM; i++ ){
           boolean dimMatchFlag = true;
           for( int j = 0; j < inputConditions[i].length; j++ ){
                int condIdx = inputConditions[i][j];
                if( conditions[i][condIdx] == 0 ){
                    dimMatchFlag = false;
                    break;
                }
           }
           if( dimMatchFlag )
               activations[i] = 1.0;
           else{
               misMatchNum++;
               DIFF_DIM_IDX = i;
           }
      }
      // calculate rule support.
      if( partialMatchFlag ){
          double[] weights = taskClarion.getRerWeights(netIdx);
          switch( ruleType ){
                  case Global.RER:
                       weights = taskClarion.getRerWeights(netIdx);
                       break;
                  case Global.IRL:
                       weights = taskClarion.getIrlWeights(netIdx);
                       break;
                  case Global.FR:
                       weights = taskClarion.getFrWeights(netIdx);
                       break;
          }
          for( int i = 0; i < INPUT_DIM_NUM; i++ )
               ruleSupport += activations[i] * weights[i];
      }
      else{
          if( misMatchNum == 0 )
              ruleSupport = 1.0;
          else
              ruleSupport = 0.0;
      }
      return misMatchNum;
   }

   /** Returns if the performed action matches the suggested action by this rule.
    *  @return <tt>true</tt> if matched.
    *  <tt>false>, otherwise. */
   public boolean isActionMatched( short[][][] action )
   {
      int retV0 = CommonMethods.isEqual( suggestedAction[Global.ACTION_NORMAL], action[Global.ACTION_NORMAL] );
      if( subsysIdx == Global.ACS && global.isNacsOn() ){
          int retV1 = CommonMethods.isEqual( suggestedAction[Global.ACTION_CONTROL], action[Global.ACTION_CONTROL] );
          return retV0 == action[Global.ACTION_NORMAL].length && retV1 == action[Global.ACTION_CONTROL].length;
      }
      return retV0 == action[Global.ACTION_NORMAL].length;
   }

  /** Decides the action. this may be overrided.
    *  @param state curent state.
    *  @param exState external state.
    *  @param gsState goal structure state.
    *  @param wmState working memory state.
    *  @return <tt>true</tt> if this rule is matched.
    *  <tt>false>, otherwise. */
   public boolean decideAction( short[] state, short[] exState, short[] gsState, short[] wmState )
   {
      boolean matchFlag = (isConditionMatched(state) == 0);
      return matchFlag;
   }

   /** Returns the rule condition.
    *  @return the rule condition. */
   public short[][] getCondition()
   {
      return CommonMethods.copy( conditions );
   }

   /** Returns the condition in a specific dimension.
    *  @param dimeIdx the index to a specific dimension.
    *  @return the conditionin a specific dimension. */
   public short[] getCondition( int dimIdx )
   {
      return CommonMethods.copy( conditions[dimIdx] );
   }

   /** Returns the rule suggested action.
    *  @return the rule suggested action. */
   public short[][][] getAction()
   {
      return CommonMethods.copy( suggestedAction );
   }

   /** Returns the rule suggested action in terms of a chunk.
    *  @return the rule suggested action in terms of a chunk. */
   public Chunk getActionChunk()
   {
      int type = Global.CHK_ACTION;
      String label = Global.CHUNK_TYPE_LABELS[type];
      Feature[] featList = new Feature[ OUTPUT_DIM_NUM ];
      for( int i = 0; i < OUTPUT_DIM_NUM; i++ ){
           int typeIdx = global.getAROutputFeatureInfo( netIdx, ruleType )[i][0];
           int featIdx = global.getAROutputFeatureInfo( netIdx, ruleType )[i][1];
           String featLabel = global.getFeatureNames( typeIdx, featIdx );
           int featLen = global.getFeatureLength( typeIdx, featIdx );
           int valNum = suggestedAction[Global.ACTION_NORMAL][i].length;
           short[] values = new short[valNum];
           for( int j = 0; j < valNum; j++ )
                values[j] = suggestedAction[Global.ACTION_NORMAL][i][j];
           featList[i] = new Feature( featLabel, (short)typeIdx, (short)featIdx, (short)featLen,
                                      values, global );
      }
      Chunk chunk = new Chunk( type, label, featList, global );
      return chunk;
   }

   /** Returns the rule age since last use.
    *  @return the rule age since last use. */
   public int getAge()
   {
       return attributes.getAge();
   }

   /** Returns the number of positive matches.
    *  @return the number of positive matches. */
   public double getPM()
   {
       return attributes.getPM();
   }

   /** Returns the number of negative matches.
    *  @return the number of negative matches. */
   public double getNM()
   {
       return attributes.getNM();
   }

   /** Returns the response time.
    *  @return the response time. */
   public int getResponseTime()
   {
       return attributes.getResponseTime();
   }

   /** Returns the rule support.
    *  @return the rule support. */
   public double getRuleSupport()
   {
       return ruleSupport;
   }

   /** Returns the rule utility.
    *  @return the rule utility. */
   public double getUtility()
   {
       attributes.calcUtility( ruleCost );
       return attributes.getUtility();
   }

   /** Sets the rule conditon.
    *  @param cond the condition to set. */
   public void setCondition( short[][] cond )
   {
      for( int i = 0; i < cond.length; i++ )
       for( int j = 0; j < cond[ i ].length; j++ )
            conditions[ i ][ j ] = cond[ i ][ j ];
   }

   /*
   public void setSuggestedAction( short[] action )
   {
      //for( int i = 0; i < OUTPUT_DIM_NUM; i++ )
      //    suggestedAction[i] = action[i];
      suggestedAction = action;
   }
   */

   /** Sets the rule age.
    *  @param age the age to set. */
   public void setAge( int age )
   {
       attributes.setAge( age );
   }

   /** Sets the number of positive matches.
    *  @param pm the number of positive matches to set. */
   public void setPM( float pm )
   {
       attributes.setPM( pm );
   }

   /** Sets the number of negative matches.
    *  @param nm the number of negative matches to set. */
   public void setNM( float nm )
   {
       attributes.setNM( nm );
   }

   /*
   public void setRuleCost()
   {
       attributes.setCost( ruleCost );
   }
   */

   /** Sets the rule attributes.
    *  @param feat the rule attributes to set. */
   public void setAttributes( RuleAttributes feat )
   {
       attributes = (RuleAttributes)feat.clone();
   }

   /** Increases the rule age. */
   public void aging()
   {
       attributes.updateAge();
   }

   /** Discounts statistics. */
   public void discountStatistics()
   {
       attributes.discountMatches( DISCOUNT );
   }

   /** Resets statistics. */
   public void resetStatistics()
   {
       attributes.resetMatches();
   }

   /** Calculates the rule BLA.
    *  @param curTime current time. */
   public void calcBLA( int curTime )
   {
       attributes.calcBLA( curTime );
   }

   /** Calculates the top level decision time .
    *  @param chunkBLA the relevant chunk BLA. */
   public void calcTLDT( double chunkBLA )
   {
       attributes.calcTLDT( chunkBLA );
   }

   /** Calculates the response time.
    *  @param TLAT the top level actuation time. */
   public void calcRT( int TLAT )
   {
       attributes.calcTLRT( TLAT );
   }

   /** Updates routine.
    *  @param actionMatchFlag flag indicating action is matched or not.
    *  @param successFlag flag indicating current step is successfule or not.
    *  @param curTime current time. */
   public void update( boolean actionMatchFlag, boolean successFlag, int curTime )
   {
       attributes.setAge( 0 );
       if( blaOption == global.VARIED_OPTION )
           attributes.updateUsages( curTime );
       if( actionMatchFlag )
           attributes.updateMatches( successFlag );
   }

   /** Converts current state into the format of dim-vals (dimensions and their
    *  active values). inputConditions used for converting the original input state
    *  into input conditions.
    *  @param state current state. */
   public void convertToInputConditions( short[] state )
   {
        for( int i = 0; i < INPUT_DIM_NUM; i++ ){
             short[] values = new short[ inputDVs[i] ];
             int count = 0;
             for( int j = 0; j < inputDVs[i]; j++ )
                  if( state[ inputOffsets[i] + j ] == 1 )
                      values[ count++ ] = (short)j;
             inputConditions[ i ] = new short[ count ];
             for( int j = 0; j < count; j++ )
                  inputConditions[ i ][ j ] = values[ j ];
        }
   }

   /** Checks if this rule covers other rule by both condition and action.
    *  @return:
    *   -2: cover completely;
    *   -1: more than one dim is different
    * >= 0: the different dim index. */
   public int cover( TLRule rule )
   {
      if( rule == null )
          return -1;
      // check if the actions are the same.
      boolean flag = this.isActionMatched( rule.getAction() );
      if( !flag )
          return -1;
      //check if the conditions are the same.
      int retVal = -2;
      short[][] otherCondition = rule.getCondition();
      for( int i = 0; i < INPUT_DIM_NUM; i++ )
      {
           int count = 0;
           for( int j = 0; j < inputDVs[i]; j++ )
                if( conditions[i][j] == 0 && otherCondition[i][j] == 1 )
                    count++;
           if( count > 0 )
           {
               if( retVal == -2 )   // first different dim
                   retVal = i;
               else                // more than one dim different.
               {
                   retVal = -1;
                   break;
               }
           }
       }
       return retVal;
   }
 }