package clarion.acs.wm;
import clarion.acs.net.*;
import clarion.system.*;
import clarion.common.*;
import clarion.nacs.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is to implement the working memory action-decision network. It extends
 * the generic class <em>AcsNet</em>.
 *
    WM action format:
    dimension i :
         value 0 : do nothing for WM slot i
         value 1 : set action for WM slot i
         value 2 : reset action for WM slot i
    dimension size_WM :
         value 0 : DO_THING action
         value 1 : ENABLE_SLOTS action
         value 2 : RESET_ALL action
    in the above format, the priority order is :
    - DO_THING action
    - RESET_ALL action
    - ENABLE_SLOTS (enables actions specific to a WM slot)
    the priority order means that
    IF DO_THING action
    THEN do this aciton and hide all the other actions
    ELSE IF RESET_ALL action
         THEN do this aciton and hide ENABLE_SLOTS action
         ELSE do ENABLE_SLOTS action, that is, do actions specific to a WM slot.
 *
 *
 * <li> summary of the important methods. <br>
 * <ol>
 * <li> public void reasoning(): the process of reasoning for determining an working
 * memory action to perform given current input.
 * <li> public void perform(): the process of performing the working memory action.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * This class implements the working action-decision network which is one of three
 * basic action-decision networks.
 * <li> Usage of this class. <br>
 * It is instantiated inside an <em>ACS</em> object.
 * </ul>
 * @author Xi Zhang
 *
 */

public class WMNet extends AcsNet {

  /** working memory. */
  protected WM wm;
  /** retrieval results from NACS. */
  protected GKSChunk retrievalResults[];

  /** constructor.
   *  @param nIdx net index.
   *  @param g the global settings. */
  public WMNet( int nIdx, Global g )
  {
     super( nIdx, g );
     wm = new WM( global );
  }

  /** Sets the retrieval results from NACS.
   *  @param result the retrieval result from NACS. */
  public void setRetrievalResult( GKSChunk[] result )
  {
     retrievalResults = result;
  }

  /** reasoning for determining an action to perform given current input. */
  public void reasoning()
  {
     // get active net comps.
     if( global.getWmActionDecisionMode() == Global.ACTION_BY_TL ){
         for( int i = 0; i < COMP_NUM; i++ ){
              if( netCompFlags[ i ] ){
                  netComps[ i ].setStates( state, exState, gsState, wmState );
                  netComps[ i ].reasoning();
              }
         }
         // select which part: BL, RER, IRL or FR to decide action.
         chosenComp = Global.FR;
         // decide an action.
         decideAction( chosenComp );
         // set the chosen action.
         deliverAction();
     }
     else
         super.reasoning();
  }

  /*
   the WM action format is the following:
    - dimension 0 :
         value 0 : DO_NOTHING
         value 1 : SET
         value 2 : RESET_ONE
         value 3 : RESET_ALL
    - dimension 1 :
         value 0 : WM slot 0 is involved
         value 1 : WM slot 1 is involved
           .    .    .    .
           .    .    .    .
         value i : WM slot i is involved

    Actually, the actions: DO_NOTHING and RESET_ALL needn't parameters. Only the aciton SET use
    the dimension 1 as parameters.
  */
  public void perform()
  {
     System.out.println( "WMNet.java, performed action = " + chosenFullAction[0] );

     boolean updateFlag = false;
     if( chosenFullAction[Global.ACTION_NORMAL][0][0] == 0 ){  // do-nothing action.
         updateFlag = wm.actionDoNothing();
         return;
     }
     if( chosenFullAction[Global.ACTION_NORMAL][0][0] == 2 ){  // reset-one action.
         int itemIdx = chosenFullAction[Global.ACTION_NORMAL][1][0];
         updateFlag = wm.actionReset( itemIdx );
         return;
     }
     if( chosenFullAction[Global.ACTION_NORMAL][0][0] == 3 ){  // reset-all action.
         updateFlag = wm.actionResetAll();
         return;
     }
     int count = 0;
     int num = chosenFullAction[Global.ACTION_NORMAL][1].length;
     for( int i = 0; i < num; i++ ){
          int idx = chosenFullAction[Global.ACTION_NORMAL][1][i];
          if( wm.getFlagItemOn( idx ) ){
              updateFlag = wm.actionSet( idx, null );
              continue;
          }
          Chunk chunk = retrievalResults[count++];
          if( chunk != null )
              updateFlag = wm.actionSet( idx, chunk );
          if( count == retrievalResults.length )
              break;
     }
     // update current state.
     if( updateFlag )
         wm.updateState( wmState );
  }
}