package clarion.common;
import clarion.system.*;
import clarion.tools.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is the base class for a network component (knowledge type).
 * <li> summary of the important methods. <br>
 * <ol>
 * <li> abstract public short[][] decideAction( int compIdx ): decide an action
 * and return the decided action.
 * <li> abstract public void reasoning(): the reasoning routine defined in this
 * component.
 * <li> abstract public void update( boolean posFlag ): the update routine defined
 * in this component. the parameter <em>posFlag</em> indicates current step is
 * successful or not. needs overriding.
 * <li> abstract public void update( double reward ): the update routine defined
 * in this component. the parameter <em>reward</em> is the currently received
 * reinforcement. needs overriding.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * This class is the base class for knowledge representation.
 * <li> Usage of this class. <br>
 * It should be extended by the concrete knowledge representation such as IDN, RER,
 * IRL or FR knowledge types.
 * </ul>
 * @author Xi Zhang
 *
 */

public abstract class GeneralNetComp{

  /** global settings. */
  protected Global global;
  /** task-specific CLARION settings. */
  protected TaskClarion taskClarion;
  /** the subsystem index: ACS or MCS. */
  protected int subsysIdx;
  /** the  network index. */
  protected int netIdx;
  /** length of the input. */
  protected int INPUT_NUM;
  /** length of the output in terms of dims relevant to this net comp. */
  protected int OUTPUT_NUM;
  /** full length of the output in terms of all of output dims.
   *  That is, bottom-full length. */
  protected int FULL_OUTPUT_NUM;
  /** # of input dims. */
  protected int INPUT_DIM_NUM;
  /* # of output dims  */
  protected int OUTPUT_DIM_NUM;
  /** # OF All of output dims. That is, bottom-full #. */
  protected int FULL_OUTPUT_DIM_NUM;
  /** # of values in each input dim. */
  protected short[] inputDVs;
  /** # of values in each output dim. */
  protected short[] outputDVs;
  /** the start position of each dim in the input array. */
  protected short[] inputOffsets;
  /** the start position of each dim in the output array. */
  protected short[] outputOffsets;
  /** D-V info of bottom (IDN) outputs. */
  protected short[] fullOutputDVs;
  /** start position of each (bottom) output dim in one dimensional aray. */
  protected short[] fullOutputOffsets;
  /** output activations used for weight-sum combination.
   *  the first dim: ACTION_NORMAL OR ACTION_CONTROL action.
   *  the second dim: the activations of dim-vals of the dims relevent to this net comp. */
  protected double[][] outputActivations;
  /** the finally chosen action by integrating all of the NetComps.
   *  the first dim: ACTION_NORMAL or ACTION_CONTROL.
   *  the second dim: the relevant output dims specified by this net comp.
   *  Here, the output dims is just a subset of the full output dims.
   *  the third dim: the active values in one dim. */
  protected short[][][] chosenAction;
  /** current suggested action by <tt>this</tt> network component.
   *  the first dim: ACTION_NORMAL or ACTION_CONTROL.
   *  the second dim: the relevant output dims specified by this net comp.
   *  Here, the output dims is just a subset of the full output dims.
   *  the third dim: the active values in one dim. */
  protected short[][][] suggestedAction;
  protected int currentTime, responseTime;
  /** variables for action decision making. */
  protected double temperature;
  protected StochasticDecider actionDecider;

 /** returns the number of positive matches.
   * @return the number of positive matches. */
  abstract public double getPM();
 /** returns the number of positive matches.
   * @return the number of positive matches. */
  abstract public double getNM();
 /** returns the output activations including both NORMAL and CONTROL actions if any.
   * @return the output activations. */
  abstract public double[][] getOutputActivations();
 /** returns the output activations including both NORMAL and CONTROL actions if any.
   * @return activations the array to store the output activations. */
  abstract public void getOutputActivations( double[][] activations );
 /** decide an action and returns the decided action.
   * @return the decided action. */
  abstract public short[][][] decideAction();
 /** returns the response time given action time.
   * @return the response time. */
  abstract public int  getResponseTime( int AT );
 /** set the chosen action.
   * @param chosenFullAction the full action to fill in the chosen action. */
  abstract public void setChosenAction( short[][][] chosenFullAction );
 /** set the states.
   * @param state the value to set current state.
   * @param exState the value to set external state.
   * @param gsState the value to set goal structure state.
   * @param wmState the value to set working memory state.
   * */
  abstract public void setStates( double[] state, double[] exState, double[] gsState, double[] wmState );
 /** set the current time.
   * @param curTime the time to set. */
  abstract public void setCurrentTime( int curTime );
  /** the reasoning routine defined in this component. */
  abstract public void reasoning();
  /** the update routine defined in this component.
   *  @param posFlag indicates current step is successful or not. */
  abstract public void update( boolean posFlag );
  /** the update routine defined in this component.
   *  @param reward reinforement received by current step. */
  abstract public void update( double reward );
  /** calculate the response time given by action time. */
  abstract public void calcRT( int AT );
}