package clarion.coordinate;
import clarion.system.*;
import clarion.common.*;
import clarion.tools.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is to implement the stochastic combination of outputs from both
 * levels.
 * <li> summary of the important methods. <br>
 * <ol>
 * <li> public int selectComponent(): using "randomly selected" approach to select
 * one component on which the chosen actions are based, such as action chose is
 * based on BL, FR, RER, IRL etc.
 * <li> public int selectAction( int actNum ): Selects randomly one action from
 * the action-decision network (such as EX, GS, WM action-decision network)
 * using "randomly selected" approach based on a chosen component.
 * <li> public int selectAction( double[] utilArr ): Selects one action from the
 * action-decision network (such as EX, GS, WM action-decision network) using
 * "boltzmann distribution based on utility" approach based on a chosen component.
 * <li> public int selectAction( int netIdx, int knowledgeType, double[] utilArr ):
 * Selects one action from the action-decision network (such as EX, GS, WM
 * action-decision network) using "boltzmann distribution based on utility"
 * approach based on a chosen component.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * The class implements the stochastic combination of outputs from both levels,
 * one of the three basic combination methods.
 * <li> Usage of this class. <br>
 * It is instantiated inside a <em>AcsNet</em> object.
 * </ul>
 * @author Xi Zhang
 *
 */

public class StochasticComb {

   /** the global settings. */
   Global global;
   /** the number of components. */
   int componentNum;
   /** the number of action types. */
   int actionNum;
   /** the chosen component. */
   int chosenComponent;
   /** the chosen action. */
   int  chosenAction;
   /** the array of probabilities. */
   double probArr[];
   /** the array of beta parameters. */
   double betaArr[];
   /** the array of success rates. */
   double succArr[];
   /** the action decider. */
   StochasticDecider actionDecider;

   /** constructor.
    *  @param mode = 0: fixed method which means the probs are fixed,
    *         mode = 1: variable method which means the probs are got by calculation.
    *  @param compNum the number of components.
    *  @param arr the array of parameter values.
    *         mode = 0: carries the values of probabilities.
    *         mode = 1: carries the values of beta.  */
   public StochasticComb( int mode, int compNum, double[] arr )
   {
      componentNum = compNum;

      if( mode == 0 )      // fixed mode.
      {
          probArr = new double[ componentNum ];
          for( int i = 0; i < componentNum; i++ )
               probArr[ i ] = arr[ i ];
      }
      else                // variable mode.
      {
          probArr = new double[ componentNum ];
          betaArr = new double[ componentNum ];
          succArr = new double[ componentNum ];
          for( int i = 0; i < componentNum; i++ )
               betaArr[ i ] = arr[ i ];
      }

      actionDecider = new StochasticDecider( actionNum,
                          World.TASK_PARA_ARRAY[ World.TEMPERATURE ] );
   }

   /** constructor.
    *  @param mode = 0: fixed method which means the probs are fixed,
    *         mode = 1: variable method which means the probs are got by calculation.
    *  @param compNum the number of components.
    *  @param arr the array of parameter values.
    *         mode = 0: carries the values of probabilities.
    *         mode = 1: carries the values of beta.
    *   @param g the global settings. */
   public StochasticComb( int mode, int compNum, double[] arr, Global g )
   {
      componentNum = compNum;
      if( mode == 0 )      // fixed mode.
      {
          probArr = new double[ componentNum ];
          for( int i = 0; i < componentNum; i++ )
               probArr[ i ] = arr[ i ];
      }
      else                // variable mode.
      {
          probArr = new double[ componentNum ];
          betaArr = new double[ componentNum ];
          succArr = new double[ componentNum ];
          for( int i = 0; i < componentNum; i++ )
               betaArr[ i ] = arr[ i ];
      }

      global = g;
      actionDecider = new StochasticDecider();
   }

   /** Sets the success rates. used in variable combination method. */
   public void setSuccRate( double[] arr )
   {
      for( int i = 0; i < componentNum; i++ )
           succArr[ i ] = arr[ i ];
   }

   /** Calculates the probabilities. Used for variable combination. */
   public void calcProbabilities()
   {
      double sum = 0;
      for( int i = 0; i < componentNum; i++ )
           sum += betaArr[ i ] * succArr[ i ];

      for( int i = 0; i < componentNum; i++ )
           probArr[ i ] = betaArr[ i ] * succArr[ i ] / sum;
   }

   /** using "radomly selected" approach to select one component on which the
    *  chosen actions are based, such as action chosing is based on BL, FR, RER,
    *  IRL etc. */
   public int selectComponent()
   {
      double randomVal = Math.random();
      chosenComponent = 0;
      double threshold = 0.0;
      for( int i = 0; i < componentNum; i++ )
      {
           threshold += probArr[ i ];
           if( randomVal < threshold )
           {
               chosenComponent = i;
               break;
           }
      }
      return chosenComponent;
   }

   /** Selects randomly one action from the action-decision network (such as EX,
    *  GS, WM action-decision network) using "randomly selected" approach based
    *  on a chosen component.
    *  @param actNum the number of actions in that network. */
   public int selectAction( int actNum )
   {
      actionNum = actNum;
      double threshold = 0.0;
      double randomVal = Math.random();
      chosenAction = 0;
      double interval = 1.0 / actionNum;
      for( int i = 0; i < actionNum; i++ )
      {
           threshold += interval;
           if( randomVal < threshold )
           {
               chosenAction = i;
               break;
           }
      }
      return chosenAction;
   }

  /** Selects one action from the action-decision network (such as EX, GS, WM
   *  action-decision network) using "boltzmann distribution based on utility"
   *  approach based on a chosen component.
   *  @param utilArr the array of utilities of action from the chosen component
   *  in that network. */
   public int selectAction( double[] utilArr )
   {
      chosenAction = 0;
      chosenAction = actionDecider.getRecommendation( utilArr );
      return chosenAction;
   }

  /** Selects one action from the action-decision network (such as EX, GS, WM
   *  action-decision network) using "boltzmann distribution based on utility"
   *  approach based on a chosen component.
   *  @param netIdx the network index.
   *  @param knowledgeType the chosen component in that network.
   *  @param utilArr the array of utilities of action from the chosen component
   *  in that network. */
   public int selectAction( int netIdx, int knowledgeType, double[] utilArr )
   {
      chosenAction = 0;
      actionDecider.setTemperature( global.getARActionSelectionTemp( netIdx, knowledgeType ) );
      chosenAction = actionDecider.getRecommendation( utilArr );
      return chosenAction;
   }
}