package clarion.mcs;
import java.util.Vector;
import clarion.system.*;
import clarion.acs.*;
import clarion.nacs.*;
import clarion.common.*;
import clarion.tools.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is to implement the monitor buffer in the MCS subsystem of CLARION
 * model.
 * <li> summary of the important methods. <br>
 * <ol>
 * <li> public void calcAcsRelStrengths(): calculates the relative strengths of
 * top conclusions from ACS.
 * <li> public void calcNacsRelStrengths(): calculates the relative strengths of
 * top conclusions from NACS.
 * <li> public void calcAcsProgress(): calculates the progress of ACS.
 * <li> public void calcNacsProgress(): calculates the progress of NACS.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * This class is to represent the monitor buffer in the MCS subsystem of CLARION
 * model. As decribed in CLARION model, this buffer contains several sections:
 * the ACS performance section, the NACS performance section, the ACS learning
 * section, the NACS learning section and other section. This buffer can be used
 * to help filtering selection and regulation.
 * <li> Usage of this class. <br>
 * It is instantiated inside a <em>MCS</em> object.
 * </ul>
 * @author Xi Zhang
 *
 */

public class MonitorBuf {

   final int BL = Global.BL;
   final int TL = Global.TL;
   final int COMB = 2;
   /** global CLARION settings. */
   protected Global global;
   /** the parameters in the performance and learning section. see tutorial on
    *  the details of these parameters. */
   protected int c20 = 2, c21 = 2, c22 = 1, c23 = 1;
   protected int c24 = 2, c25 = 2, c26 = 1, c27 = 1;
   protected int c28 = 2, c29 = 2, c30 = 2, c31 = 2;
   /** the number of ACS networks. */
   protected int acsNetNum;
   /** the number of components in each ACS network. */
   protected int acsNetCompNum;
   /** the length of this monitor buffer. */
   protected int bufferLength;
   /** the content of the monitor buffer. */
   protected double[] buffer;
   //double[] acsStrengths[][], nacsStrengths[];
   /** relative strengths of the top conclusions from ACS. */
   protected double[] acsRelStrengths[][];
   /** relative strengths of the top conclusions from NACS. */
   protected double[] nacsRelStrengths[];
   /** progress of ACS. */
   protected double[] acsProgress[][];
   /** progress of NACS. */
   protected double[] nacsProgress[];
   /** improvement of ACS. */
   protected double[] acsImprove[][];
   /** improvement of ACS. */
   protected double[] nacsImprove[];
   /** the previous top conclusions from ACS. */
   protected Vector  preAcsChunks[][];
   /** the previous top conclusions from NACS. */
   protected Vector  preNacsChunks[];
   /** the current top conclusions from ACS. */
   protected Vector  curAcsChunks[][];
   /** the current top conclusions from NACS. */
   protected Vector  curNacsChunks[];

   /** constuctor.
    *  @param global the associated global settings. */
   public MonitorBuf( Global global )
   {
      this.global = global;
      c20 = global.getMonitorAcsTopConcNum();
      c24 = global.getMonitorAcsBottomConcNum();
      c28 = global.getMonitorAcsTopEpisodeNum();
      c30 = global.getMonitorAcsBottomEpisodeNum();
      c21 = global.getMonitorNacsTopConcNum();
      c25 = global.getMonitorNacsBottomConcNum();
      c29 = global.getMonitorNacsTopEpisodeNum();
      c31 = global.getMonitorNacsBottomEpisodeNum();
      acsNetNum = global.getAcsNetNum();
      acsNetCompNum = global.getAcsKnowledgeNum();
      // relative strengths.
      acsRelStrengths = new double[acsNetNum][][];
      for( int i = 0; i < acsNetNum; i++ ){
           int c = ( i == BL ) ? c24 : c20;
           acsRelStrengths[i] = new double[acsNetCompNum][c];
      }
      //nacsRelStrengths = new double[2][];
      nacsRelStrengths = new double[3][];
      nacsRelStrengths[TL] = new double[c21];
      nacsRelStrengths[BL] = new double[c25];
      nacsRelStrengths[COMB] = new double[ Math.max(c21, c25) ];
      // progress.
      acsProgress = new double[acsNetNum][][];
      for( int i = 0; i < acsNetNum; i++ ){
           int c = ( i == BL ) ? c26 : c22;
           acsProgress[i] = new double[acsNetCompNum][c];
      }
      nacsProgress = new double[2][];
      nacsProgress[TL] = new double[c23];
      nacsProgress[BL] = new double[c27];
      // improvements.
      acsImprove = new double[acsNetNum][2][];
      for( int i = 0; i < acsNetNum; i++ ){
           acsImprove[i][TL] = new double[c28];
           acsImprove[i][BL] = new double[c30];
      }
      nacsImprove = new double[2][];
      nacsImprove[TL] = new double[c29];
      nacsImprove[BL] = new double[c31];
      //init the bufer
      bufferLength = 0;
      bufferLength += CommonMethods.calcLength( acsRelStrengths );
      bufferLength += CommonMethods.calcLength( nacsRelStrengths );
      bufferLength += CommonMethods.calcLength( acsProgress );
      bufferLength += CommonMethods.calcLength( nacsProgress );
      bufferLength += CommonMethods.calcLength( acsImprove );
      bufferLength += CommonMethods.calcLength( nacsImprove );
      buffer = new double[ bufferLength ];
      // init variables for storing the chunks.
      preAcsChunks = new Vector[acsNetNum][acsNetCompNum];
      for( int i = 0; i < acsNetNum; i++ )
        for( int j = 0; j < acsNetCompNum; j++ ) {
             int c = ( i == BL ) ? c24 : c20;
             preAcsChunks[i][j] = new Vector( c );
      }
      curAcsChunks = new Vector[acsNetNum][acsNetCompNum];
      for( int i = 0; i < acsNetNum; i++ )
        for( int j = 0; j < acsNetCompNum; j++ ) {
             int c = ( i == BL ) ? c24 : c20;
             curAcsChunks[i][j] = new Vector( c );
      }
      //preNacsChunks = new Vector[2];
      preNacsChunks = new Vector[3];
      preNacsChunks[TL] = new Vector( c21 );
      preNacsChunks[BL] = new Vector( c25 );
      preNacsChunks[COMB] = new Vector( Math.max(c21, c25) );
      //curNacsChunks = new Vector[2];
      curNacsChunks = new Vector[3];
      curNacsChunks[TL] = new Vector( c21 );
      curNacsChunks[BL] = new Vector( c25 );
      curNacsChunks[COMB] = new Vector( Math.max(c21, c25) );
   }

   /** Returns the length of this buffer. */
   public int getLength()
   {
      return buffer.length;
   }

   /** Returns the top conclusions from ACS with specific network and the component
    *  in that network.
    *  @param netIdx the network index.
    *  @param compIdx the component index.
    *  @return the top conclusions. */
   public Vector getAcsChunks( int netIdx, int compIdx )
   {
      return curAcsChunks[netIdx][compIdx];
   }

   /** Returns the top conclusions from NACS with specific level.
    *  @param level the specific level.
    *  @return the top conclusions. */
   public Vector getNacsChunks( int level )
   {
      return curNacsChunks[level];
   }

   /** Returns the contents of this buffer.
    *  @return the contents of this buffer. */
   public double[] getBuffer()
   {
      int offset = 0;
      for( int i = 0; i < acsRelStrengths.length; i++ )
       for( int j = 0; j < acsRelStrengths[i].length; j++ )
        for( int k = 0; k < acsRelStrengths[i][j].length; k++ )
             buffer[ offset++ ] = acsRelStrengths[i][j][k];
      for( int i = 0; i < nacsRelStrengths.length; i++ )
       for( int j = 0; j < nacsRelStrengths[i].length; j++ )
             buffer[ offset++ ] = nacsRelStrengths[i][j];
      for( int i = 0; i < acsProgress.length; i++ )
       for( int j = 0; j < acsProgress[i].length; j++ )
        for( int k = 0; k < acsProgress[i][j].length; k++ )
             buffer[ offset++ ] = acsProgress[i][j][k];
      for( int i = 0; i < nacsProgress.length; i++ )
       for( int j = 0; j < nacsProgress[i].length; j++ )
             buffer[ offset++ ] = nacsProgress[i][j];
      for( int i = 0; i < acsImprove.length; i++ )
       for( int j = 0; j < acsImprove[i].length; j++ )
        for( int k = 0; k < acsImprove[i][j].length; k++ )
             buffer[ offset++ ] = acsImprove[i][j][k];
      for( int i = 0; i < nacsImprove.length; i++ )
       for( int j = 0; j < nacsImprove[i].length; j++ )
             buffer[ offset++ ] = nacsImprove[i][j];
      return buffer;
   }

   /** Returns the relative strengths of the top conclusions from ACS with the
    *  specific network and the component in that network.
    *  @param netIdx network NO.
    *  @param compIdx component index: IDN, RER, IRL, FR.
    *  @return the relative strengths of the top conclusions.  */
   public double[] getAcsRelStrengths( int netIdx, int compIdx )
   {
      calcRelStrengths( curAcsChunks[netIdx][compIdx], acsRelStrengths[netIdx][compIdx] );
      return acsRelStrengths[netIdx][compIdx];
   }

   /** Returns the relative strengths of the top conclusions from NACS with the
    *  specific level : TL or BL.
    *  @param level the specific level : TL or BL.
    *  @return the relative strengths of the top conclusions.  */
   public double[] getNacsRelStrengths( int level )
   {
      calcRelStrengths( curNacsChunks[level], nacsRelStrengths[level] );
      return nacsRelStrengths[ level ];
   }

   /** Returns the strengths of the top conclusions from ACS with the
    *  specific network and the component in that network.
    *  @param netIdx network NO.
    *  @param compIdx component index: IDN, RER, IRL, FR.
    *  @return the relative strengths of the top conclusions.
    *  @deprecated Replaced by getAcsRelStrengths( int netIdx, int compIdx ). */
   public double[] getAcsStrengths( int netIdx, int compIdx )
   {
      int size = curAcsChunks[netIdx][compIdx].size();
      double[] strengths = new double[ size ];
      for( int i = 0; i < size; i++ )
           strengths[i] = ((Chunk)curAcsChunks[netIdx][compIdx].elementAt(i)).getStrength();
      return strengths;
   }

   /** Returns the strengths of the top conclusions from NACS with the
    *  specific level : TL or BL.
    *  @param level the specific level : TL or BL.
    *  @return the relative strengths of the top conclusions.
    *  @deprecated Replaced by getNacsRelStrengths( int level ). */
   public double[] getNacsStrengths( int level )
   {
      int size = curNacsChunks[level].size();
      double[] strengths = new double[ size ];
      for( int i = 0; i < size; i++ )
           strengths[i] = ((Chunk)curNacsChunks[level].elementAt(i)).getStrength();
      return strengths;
   }

   /** Returns the progress of ACS with the specific network and the component in
    *  that network.
    *  @param netIdx network NO.
    *  @param compIdx component index: IDN, RER, IRL, FR.
    *  @return the progress. */
   public double[] getAcsProgress( int netIdx, int compIdx )
   {
      return acsProgress[netIdx][compIdx];
   }


   /** Returns the progress of NACS with the specific level.
    *  @param level the specific level: TL or BL.
    *  @return the progress. */
   public double[] getNacsProgress( int level )
   {
      return nacsProgress[ level ];
   }

   /** Returns the improvement of ACS with the specific network and level.
    *  @param netIdx network NO.
    *  @param level level: top or bottom.
    *  @return the progress. */
   public double[] getAcsImprovement( int netIdx, int level )
   {
      return acsImprove[ netIdx ][ level ];
   }

   /** Returns the improvement of NACS with the specific level.
    *  @param level level: top or bottom.
    *  @return the improvement. */
   public double[] getNacsImprovement( int level )
   {
      return nacsImprove[ level ];
   }

   /** Sets current top conclusions from ACS with specific network and component.
    *  @param netIdx the network NO.
    *  @param compIdx the component index.
    *  @param chunks the top conclusions from that component. */
   public void setCurAcsConclusions( int netIdx, int compIdx, Chunk[] chunks )
   {
      if( chunks == null )
          return;
      try{
          preAcsChunks[netIdx][compIdx].removeAllElements();
          // move conclusions from current to previous.
          int num = curAcsChunks[netIdx][compIdx].size();
          for( int i = 0; i < num; i++ )
               preAcsChunks[netIdx][compIdx].addElement( curAcsChunks[netIdx][compIdx].remove(0) );
          // fill in current conclusions.
          num = (compIdx == BL)? Math.min( c24, chunks.length ):Math.min( c20, chunks.length );
          for( int i = 0; i < num; i++ )
               curAcsChunks[netIdx][compIdx].addElement((Chunk)(chunks[i].clone()));
      }
      catch( CloneNotSupportedException e ){
      }
   }

   /** Sets current top conclusions from NACS with specific level.
    *  @param level the level: top or bottom.
    *  @param chunks the top conclusions from that component. */
   public void setCurNacsConclusions( int level, Chunk[] chunks )
   {
      if( chunks == null )
          return;
      try{
          preNacsChunks[level].removeAllElements();
          // move conclusions from current to previous.
          int num = curNacsChunks[level].size();
          for( int i = 0; i < num; i++ )
               preNacsChunks[level].addElement( curNacsChunks[level].remove(0) );
          // fill in current conclusions.
          num = Math.min( curNacsChunks[level].capacity(), chunks.length );
          for( int i = 0; i < num; i++ )
               curNacsChunks[level].addElement((Chunk)(chunks[i].clone()));

          /*
          System.out.println( "chunks in montor buffer in MonitorBuf.java = " );
          for( int i = 0; i < curNacsChunks[level].size(); i++ )
               System.out.println( curNacsChunks[level].elementAt(i) );
          */

      }
      catch( CloneNotSupportedException e ){
      }
   }

   /** Sets the success rate of an episode in ACS with the specific network and
    *  level.
    *  @param netIdx the network NO.
    *  @param level the specific level: top or bottom.
    *  @param succRate the success rate to set. */
   public void setAcsEpiSuccRate( int netIdx, int level, double succRate )
   {
      int len = acsImprove[netIdx][level].length;
      for( int i = 1; i < len; i++ )
           acsImprove[netIdx][level][i-1] = acsImprove[netIdx][level][i];
      acsImprove[netIdx][level][len-1] = succRate;
   }

   /** Sets the success rate of an episode in NACS with the specific level.
    *  @param level the specific level: top or bottom.
    *  @param succRate the success rate to set. */
   public void setNacsEpiSuccRate( int level, double succRate )
   {
      int len = nacsImprove[level].length;
      for( int i = 1; i < len; i++ )
           nacsImprove[level][i-1] = nacsImprove[level][i];
      nacsImprove[level][len-1] = succRate;
   }

   /** Calculates the relative strengths of top conclusions from ACS. */
   public void calcAcsRelStrengths()
   {
      for( int i = 0; i < acsNetNum; i++ )
        for( int j = 0; j < acsNetCompNum; j++ )
             calcRelStrengths( curAcsChunks[i][j], acsRelStrengths[i][j] );
   }

   /** Calculates the relative strengths of top conclusions from NACS. */
   public void calcNacsRelStrengths()
   {
      calcRelStrengths( curNacsChunks[BL], nacsRelStrengths[BL] );
      calcRelStrengths( curNacsChunks[TL], nacsRelStrengths[TL] );
      calcRelStrengths( curNacsChunks[COMB], nacsRelStrengths[COMB] );
   }

   /** Calculates the progress of ACS. */
   public void calcAcsProgress()
   {
      for( int i = 0; i < acsNetNum; i++ )
        for( int j = 0; j < acsNetCompNum; j++ )
             calcProgress( preAcsChunks[i][j], curAcsChunks[i][j], acsProgress[i][j] );
   }

   /** Calculates the progress of NACS. */
   public void calcNacsProgress()
   {
      calcProgress( preNacsChunks[BL], curNacsChunks[BL], nacsProgress[BL] );
      calcProgress( preNacsChunks[TL], curNacsChunks[TL], nacsProgress[TL] );
   }

   private void calcRelStrengths( Vector chunks, double[] strengths )
   {
      double sum = 0.0;

      //System.out.println( "chunk size = " + chunks.size() );
      //System.out.println( "strength len = " + strengths.length );

      for( int i = 0; i < chunks.size(); i++ )
           sum += ((Chunk)chunks.elementAt(i)).getStrength();
      for( int i = 0; i < chunks.size(); i++ )
           strengths[i] = ((Chunk)chunks.elementAt(i)).getStrength()/sum;
   }

   private void calcProgress( Vector preChunks, Vector curChunks, double[] progress )
   {
      if( preChunks.isEmpty() )
          return;
      int chgNum = preChunks.size();
      for( int i = 0; i < preChunks.size(); i++ ){
           Chunk chunk1 = (Chunk)(preChunks.elementAt(i));
           for( int j = 0; j < curChunks.size(); j++ ){
                Chunk chunk2 = (Chunk)(curChunks.elementAt(j));
                if( chunk1.equals( chunk2 ) ){
                    chgNum--;
                    break;
                }
           }
      }
      for( int i = 1; i < progress.length; i++ )
           progress[i-1] = progress[i];
      progress[progress.length-1] = (double)chgNum / preChunks.size();
   }
}