package clarion.nacs;
import java.awt.*;
import java.util.Arrays;
import java.util.Vector;
import clarion.common.*;
import clarion.system.*;
import clarion.tools.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is to implement an AMN network in bottom level of the NACS subsystem.
 * It implements the learning and reasoning processes inside AMN described in CLARION model.
 * <li> summary of the important methods. <br>
 * <ol>
 * <li> public void training(): the assimulation (training) process.
 * <li> public boolean retrieval( boolean multiFlag ): the retrieval process.
 * <li> public void update(): update the AMN.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * This class implements AMN part, the bottom level of NACS subsystem.
 * <li> Usage of this class. <br>
 * It is instantiated inside a <em>NACS</em> object.
 * </ul>
 * @author Xi Zhang
 */

public class AMNet extends BPNet{

  /** different types of AMN. */
  final int AAM_TYPE = 0, HAM_TYPE = 1;
  /** global settings. */
  Global global;
  /** amn net index. */
  int netIdx;
  /** GKS */
  GKS gks;
  /** associative assimulation is on or not, for offline training. */
  boolean assocAssimOn;
  /** the number of AMN passes for each iteration. */
  int passNum;
  /** EM is on or not. */
  boolean emOn;
  /** the probabliity of using EM to train the AMNs. */
  double  emTrainNacsProb;
  /** retrieval results. */
  Feature[] results;
  /** the # of dims of input. */
  int    inputDimNum;
  /** the # of dims of output. */
  int    outputDimNum;
  /** the # of values in each dim of input. */
  short  inputDVs[];
  /** the # of values in each dim of input. */
  short  outputDVs[];
  /** the offset of each dim in whole input. */
  short  inputOffsets[];
  /** the offset of each dim in whole output. */
  short  outputOffsets[];
  /** input's each feature's type and order in that type. */
  short  inputFeatInfo[][];
  /** output's each feature's type and order in that type. */
  short  outputFeatInfo[][];
  // need be added to description.
  /** threshold of noise. */
  double noiseThreshold;
  /** the response time. */
  int    RT;
  /** one pass time of associative mapping. */
  int    t_AMN;

  // old constructor.
  public AMNet( int inputNum, int hiddenNum, int outputNum, int dNum, int[] dValNums )
  {
      super( inputNum, hiddenNum, outputNum );
      inputDimNum = dNum;
      inputDVs = new short[ inputDimNum ];
      for( int i = 0; i < inputDimNum; i++ )
           inputDVs[ i ] = (short)dValNums[ i ];
      passNum = global.getAmnPassNum( netIdx );
  }

  // old constructor.
  public AMNet( int inputNum, int hiddenNum, int outputNum, int dNum, short[] dValNums )
  {
      super( inputNum, hiddenNum, outputNum );
      inputDimNum = dNum;
      inputDVs = new short[ inputDimNum ];
      for( int i = 0; i < inputDimNum; i++ )
           inputDVs[ i ] = (short)dValNums[ i ];
      passNum = global.getAmnPassNum( netIdx );
  }

  /** new constructor, # of input nodes should be equal to # of output nodes.
   *  @param nIdx the AMN index.
   *  @param g the associated global settings.
   *  @param gks the associated GKS.  */
  public AMNet( int nIdx, Global g, GKS gks )
  {
      super( g.getAmnInputNum(nIdx), nIdx, g, Global.NACS );
      netIdx = nIdx;
      global = g;
      this.gks = gks;
      emOn = global.isEMon();
      if( emOn )
          //emTrainNacsProb = global.getEmTrainProb();
          emTrainNacsProb = global.getAmnTrainedByEmProb();
      inputDVs = global.getAmnInputDimVals( netIdx );
      outputDVs = global.getAmnOutputDimVals( netIdx );
      inputDimNum = inputDVs.length;
      outputDimNum = outputDVs.length;
      inputOffsets = new short[ inputDimNum ];
      outputOffsets = new short[ outputDimNum ];
      CommonMethods.calcOffsets( inputOffsets, inputDVs );
      CommonMethods.calcOffsets( outputOffsets, outputDVs );
      inputFeatInfo = global.getAmnInputFeatureInfo( netIdx );
      outputFeatInfo = global.getAmnOutputFeatureInfo( netIdx );
      if( global.isRTOn() )
          t_AMN = global.getTaskClarion().getAmnPassTime();
      passNum = global.getAmnPassNum( netIdx );
  }

  /** reinitialization routine. */
  public void reinit()
  {
      super.reinitWeights();
  }

  /** Returns the retrieval results in terms of features.
   *  @return the retrieval results. */
  public Feature[] getResults()
  {
      return results;
  }

  /** Returns the output in terms of one dimensional array after filtering noise.
   *  @return the retrieval results. */
  public void getOutput( double[] arr )
  {
      for( int i = 0; i < arr.length; i++ )
           if( Output[ i ] < noiseThreshold )
               arr[ i ] = 0.0;
           else
               arr[ i ] = Output[ i ];
  }

  /** Returns the output in terms of two dimensional array after filtering noise.
   *  @return the retrieval results. */
  public void getOutput( double[][] arr )
  {
      int offset = 0;
      for( int i = 0; i < outputDimNum; i++ )
      {
           for( int j = 0; j <  outputDVs[ i ]; j++ )
           {
                if( Output[ offset + j ] < noiseThreshold )
                    arr[ i ][ j ] = 0.0;
                else
                    arr[ i ][ j ] = Output[ offset + j ];
           }
           offset += outputDVs[i];
      }
  }

  /** Returns the output in terms of features.
   *  @param multipFlag indicates allow multiple active values in one dimension.
   *  @return the output features. */
  public Feature[] getOutputFeatures( boolean multiFlag )
  {
      Feature[] outputFeatures = new Feature[ outputDimNum ];
      double[][] values = new double[ outputDimNum ][];
      int offset = 0;
      for( int i = 0; i < outputDimNum; i++ ){
           values[ i ] = new double[ outputDVs[ i ] ];
           for( int j = 0; j < outputDVs[ i ]; j++ ){
                if( Output[ offset + j ] < noiseThreshold )
                    values[ i ][ j ] = 0.0;
                else
                    values[ i ][ j ] = Output[ offset + j ];
           }
           short fType = outputFeatInfo[ i ][ 0 ];
           short fOrder = outputFeatInfo[ i ][ 1 ];
           String fLabel = global.getFeatureNames( fType, fOrder );
           if( !multiFlag ){
               int idx = CommonMethods.getMaxValueIdx( values[i] );
               for( int j = 0; j < values[i].length; j++ )
                    if( j != idx )
                        values[i][j] = 0.0;
           }
           outputFeatures[ i ] = new Feature( fLabel, fType, fOrder, values[ i ], global );
           offset += outputDVs[i];
      }
      return outputFeatures;
  }

  /** reset the response time. */
  public void resetRT()
  {
      RT = 0;
  }

  /** Returns the response time.
   *  @return the response time. */
  public int getRT()
  {
      return RT;
  }

  /** Sets noise threshold.
   *  @param th the value to set the noise threshold. */
  public void setNoiseThreshold( double th )
  {
      noiseThreshold = th;
  }

  /** Sets the flag for assimulation or not.
   *  @param falg the flag to set. */
  public void setAssocAssimOn( boolean flag )
  {
     assocAssimOn = flag;
  }

  public void setPassNum( int num )
  {
     passNum = num;
  }

  /** Sets the input with the specific info of active values of a chunk in terms
   *  of one dimensional array.
   *  @param arr the content used as input.
   *  @deprecated Replcaced by setInput( short[][] arr ). */
  public void setInput( int[] arr )
  {
      int offset = 0;
      for( int i = 0; i < arr.length; i++ ){
           Input[ offset + arr[i] ] = 1;
           offset += inputDVs[i];
      }
  }

  /** Sets the input with the specific info of active values of a chunk in terms
   *  of two dimensional array.
   *  @param arr the content used as input. */
  public void setInput( short[][] arr )
  {
      int offset = 0;
      for( int i = 0; i < arr.length; i++ ){
           for( int j = 0; j < inputDVs[i]; j++ )
                Input[ offset + j ] = arr[ i ][ j ];
           offset += inputDVs[i];
      }
  }

  /** Sets the input with the specific info of activation of a chunk in terms
   *  of two dimensional array.
   *  @param arr the content used as input. */
  public void setInput( double[][] arr )
  {
      int offset = 0;
      for( int i = 0; i < arr.length; i++ ){
           for( int j = 0; j < inputDVs[i]; j++ )
                Input[ offset + j ] = arr[ i ][ j ];
           offset += inputDVs[i];
      }
  }

  /** Sets the input with the specific info of active values of chunks in
   *  terms of three dimensional array.
   *  @param arr the content used as input. */
  public void setInput( short[][][] arr )
  {
      int offset = 0, dimOffset = 0;
      for( int i = 0; i < arr.length; i++ ){
           for( int j = 0; j < arr[i].length; j++ ){
                for( int k = 0; k < inputDVs[ dimOffset ]; k++ )
                     Input[ offset + k ] = arr[ i ][ j ][ k ];
                offset += inputDVs[ dimOffset ];
                dimOffset++;
           }
      }
  }

  /** Sets the input with the specific features.
   *  @param features the features used as input. */
  public void setInput( Feature[] features )
  {
      //System.out.println( "AMN input len = " + Input.length );

      CommonMethods.resetArray( Input, 0.0 );
      for( int i = 0; i < features.length; i++ ){
           if( features[i] == null )
               continue;
           int type = features[i].getType();
           int order = features[i].getOrder();
           int idx = searchFeature( type, order, inputFeatInfo );
           if( idx != -1 ){
                   short[] content = features[ i ].getContent();

                   //System.out.println( "feature len = " + content.length + ", offset = " + offset );

                   for( int j = 0; j < inputDVs[ idx ]; j++ ){

                        //if( offset + j >= Input.length || j >= content.length )
                        //    System.out.println( "feature len = " + content.length + ", offset = " + offset + ", j = " + j );


                        Input[ inputOffsets[idx] + j ] = content[ j ];
                   }
          }
      }
  }

  /** Sets the desired output with the specific info of active values of a chunk
   *  in terms of one dimensional array.
   *  @param arr the content used as output.
   *  @deprecated Replcaced by setOutput( short[][] arr ). */
  public void setDesiredOutput( short[] arr )
  {
      int offset = 0;
      for( int i = 0; i < arr.length; i++ ){
           DesiredOutput[ offset + arr[i] ] = 1;
           offset += outputDVs[i];
      }
  }

  /** Sets the desired output with the specific info of active values of a chunk
   *  in terms of two dimensional array.
   *  @param arr the content used as input. */
  public void setDesiredOutput( short[][] arr )
  {
      int offset = 0;
      for( int i = 0; i < arr.length; i++ ){
           for( int j = 0; j < outputDVs[i]; j++ )
                DesiredOutput[ offset + j ] = arr[ i ][ j ];
           offset += outputDVs[i];
      }
  }

  /** Sets the desired output with the specific features.
   *  @param features the features used as input. */
  public void setDesiredOutput( Feature[] features )
  { /*
      int offset = 0;
      for( int i = 0; i < outputDimNum; i++ ){
           if( i < features.length ){
               if( features[i] != null ){
                   short[] content = features[ i ].getContent();
                   for( int j = 0; j < outputDVs[ i ]; j++ )
                        DesiredOutput[ offset + j ] = content[ j ];
               }
               else
                   for( int j = 0; j < outputDVs[ i ]; j++ )
                        DesiredOutput[ offset + j ] = 0;
           }
           else
               for( int j = 0; j < outputDVs[ i ]; j++ )
                    DesiredOutput[ offset + j ] = 0;
           offset += outputDVs[ i ];
      }
      */

      //System.out.println( "AMN input len = " + Input.length );
      CommonMethods.resetArray( DesiredOutput, 0.0 );
      for( int i = 0; i < features.length; i++ ){
           if( features[i] == null )
               continue;
           int type = features[i].getType();
           int order = features[i].getOrder();
           int idx = searchFeature( type, order, outputFeatInfo );
           if( idx != -1 ){
               short[] content = features[ i ].getContent();

               //System.out.println( "feature len = " + content.length + ", offset = " + offset );

               for( int j = 0; j < outputDVs[ idx ]; j++ ){

                /*
                    if( j >= content.length )
                        System.out.println( "content len = " + content.length  + ", j = " + j );
                    else if( idx >= outputOffsets.length )
                             System.out.println( "outputOffsets len = " + outputOffsets.length + ", idx = " + idx );
                        else if (outputOffsets[idx] + j >= DesiredOutput.length )
                                 System.out.println( "DesiredOutput len = " + DesiredOutput.length +
                                                     ", outputOffsets[idx] + j = " + outputOffsets[idx] + j );
                */

                    DesiredOutput[ outputOffsets[idx] + j ] = content[ j ];
               }
          }
      }
  }

  /** the assimulation (training) process. */
  public void training()
  {
      forwardPass();
      backwardPass();
  }

  /** the retrieval process.
   *  @param multiFlag allow multiple active values in one dimension. */
  public boolean retrieval( boolean multiFlag )
  {
      RT = 0;
      for( int i = 0; i < passNum; i++ ){
           forwardPass();
           setInput( Output );
           if( global.isRTOn() )
               RT += t_AMN;
      }
      results = getOutputFeatures( multiFlag );
      return true;
  }

  /** update the AMN. */
  public void update()
  {
      //System.out.println( "AMN update, assocAssimOn = " + assocAssimOn );

      if( !assocAssimOn ) // associative rule assimilaiton is off.
          return;
      // offline training.
      if( emOn && Math.random() < emTrainNacsProb ){
          EMNacsSample[] nacsSamples = gks.getEmNacsSamples( netIdx );
          offlineTrain( nacsSamples );
      }
  }

/**************************** PRIVATE METHODS *********************************/
  /** offline training AMN using NACS EM. */
  private void offlineTrain( EMNacsSample[] samples )
  {
      int emIterNum = global.getAmnTrainedByEmIterNum();
      int emSampleNum = samples.length;
      /*
      System.out.println( "the iter num = " + emIterNum );
      System.out.println( "the offline NACS samples = " );
      for( int i = 0; i < samples.length; i++ )
            System.out.println( samples[i] );
      */
      Vector sands = new Vector( emSampleNum * emIterNum );
      for( int j = 0; j < emIterNum; j++ )
        for( int k = 0; k < emSampleNum; k++ )
             sands.addElement( new Integer(k) );

      /*
      System.out.println( "AMN update, vector num = " + sands.size() );
      System.out.println( "before training, given input - chunk 15, the outputs = \n" );
      setInput( global.getNACS().getGks().getChunk(15).getFeatures() );
      forwardPass();
      CommonMethods.showValues( getOutput() );
      */

      while( !sands.isEmpty() ) {
        int idx = ((Integer)(sands.remove((int)(sands.size()*Math.random())))).intValue();
        setInput( samples[idx].getInput().getFeatures() );
        forwardPass();
        setDesiredOutput( samples[idx].getOutput().getFeatures() );
        backwardPass();
      }

      /*
      System.out.println( "after training, given input - chunk 15, the outputs = \n" );
      setInput( global.getNACS().getGks().getChunk(15).getFeatures() );
      forwardPass();
      CommonMethods.showValues( getOutput() );
      */
      /*
      for( int i = 0; i < samples.length; i++ ){
           System.out.println( "the sample = \n" + samples[i] );
           setInput( samples[i].getInput().getFeatures() );
           System.out.println( "the output = \n" );
           forwardPass();
           CommonMethods.showValues( getOutput() );
      }
      */
  }

  private int searchFeature( int type, int order, short[][] info )
  {
      int idx = -1;
      for( int i = 0; i < info.length; i++ ){
           if( info[ i ][ 0 ] == type && info[ i ][ 1 ] == order ){
               idx = i;
               break;
           }
      }
      return idx;
  }
}
