package clarion.nacs;
import java.util.Vector;
import clarion.system.*;
import clarion.tools.linkedlist.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is to implement an associative rule group pointing to the same
 * conclusion chunk.
 * <li> summary of the important methods. <br>
 * <ol>
 * <li> public boolean insertRule( short[] condition, short conclusion ):
 * insert a rule with the specific condition and conclusion in terms of the
 * <li> public boolean removeRule( short[] condition, short conclusion )
 * Remove a rule with the specific condition and conclusion in terms of the
 * chunk indices on the chunk list in GKS.
 * <li> public double drawConclusion( Vector con ): draws conclusion triggered
 * by the specific list of chunk indices as condition.
 * <li> public void update(): update routine.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * This class provides an easier way to manage a group of assoiciative rules with
 * the same conclusion chunk and immplement the reasoning/learning processes in GKS.
 * <li> Usage of this class. <br>
 * It is instantiated inside its conclusion - a <em>GKSChunk</em> object.
 * </ul>
 * @author Xi Zhang
 */

public class AssocRuleGrp{

  /** the global settings. */
  Global global;
  /** the group ID. */
  int groupID;
  /** the maximal age a rule can have. */
  int MAX_RULE_AGE;
  /** EM is on or not. */
  boolean emOn;
  /** the threshold for rule extraction. */
  double threshold_ae;
  /** the conditions of the active rules. */
  short activeRuleConditions[][];
  /** the associative rule application time. */
  int assocRuleAppTime;
  /** current time. */
  long currentTime;
  /** the list of the associative rules. */
  LinkedList rules;

  /** constructor.
   *  @param id the group ID.
   *  @param g the associated global settings. */
  public AssocRuleGrp( int id, Global g )
  {
     groupID = id;
     global = g;
     emOn = global.isEMon();
     threshold_ae = global.getAssocExtThreshold();
     rules = new LinkedList();

  }

  /** Returns the group ID.
   *  @return the group ID.  */
  public int getGroupID()
  {
     return groupID;
  }

  /** Returns the number of the rules.
   *  @return the number of the rules.  */
  public int getRuleNum()
  {
     return rules.getCount();
  }

  /** Returns the rule list.
   *  @return the rule list. */
  public List getRules()
  {
     return rules;
  }

  /** Returns the conditions of the rules.
   *  @return the conditions of the rules. */
  public short[][] getRuleConditions()
  {
     int ruleNum = rules.getCount();
     if( ruleNum == 0 )
         return null;
     short[][] conditions = new short[ruleNum][];
     for( int i = 0; i < ruleNum; i++ ) {
          AssocRule rule = ( AssocRule )rules.getFrom( i );
          conditions[i] = new short[rule.getConditionNum()];
          rule.getCondition( conditions[i] );
     }
     return conditions;
  }

  /** Returns the conditions of the active rules.
   *  @return the conditions of the active rules. */
  public short[][] getActiveRuleConditions()
  {
     return activeRuleConditions;
  }

  /** Returns the rule application time.
   *  @return the rule application time. */
  public int getAssocRuleAppTime()
  {
     return assocRuleAppTime;
  }

  /** Sets GKS reasoning method.
    *  @param method the reasoning method to set. */
  public void setReasoningMethod( int method )
  {
     //reasoningMethod = method;
     for( int i = 0; i < rules.getCount(); i++ ){
          AssocRule rule = (AssocRule)( rules.getFrom( i ) );
          rule.setReasoningMethod( method );
     }
  }

  /** Sets the maximal age of a rule.
   *  @param age the age to set. */
  public void setMaxRuleAge( int age )
  {
     MAX_RULE_AGE = age;
  }

  /** Sets the current time.
   *  @param curTime current time. */
  public void setCurrentTime( long curTime )
  {
     currentTime = curTime;
  }

  /** Sets the rule list.
   *  @param list the rule list to set. */
  public void setRules( LinkedList list )
  {
     try{
         rules = (LinkedList)( list.clone() );
     }
     catch(  CloneNotSupportedException e ){
         System.out.println( "clone error " );
     }
  }

  /** insert a rule with the specific condition and conclusion in terms of the
   *  chunk indices on the chunk list in GKS.
   *  @param condition the rule condition.
   *  @param conclusion the rule conclusion. */
  public boolean insertRule( short[] condition, short conclusion )
  {
     AssocRule newRule = new AssocRule( condition, conclusion, global );
     for( int i = 0; i < rules.getCount(); i++ ){
          AssocRule rule = (AssocRule)( rules.getFrom( i ) );
          if( rule.equals( newRule ) ){   // the rule is evocated.
              rule.resetAge();
              if( global.isAssocBlaOn() && global.getAssocBlaOption() == Global.VARIED_OPTION )
                  rule.increaseUsage( currentTime );
              return false;
          }
     }
     rules.insertTail( newRule );
     return true;
  }

  /** Remove a rule with the specific condition and conclusion in terms of the
   *  chunk indices on the chunk list in GKS.
   *  @param condition the rule condition.
   *  @param conclusion the rule conclusion. */
  public boolean removeRule( short[] condition, short conclusion )
  {
     boolean retVal = false;
     int i = 0;
     while( i < rules.getCount() ) {
          AssocRule rule = (AssocRule)( rules.getFrom( i ) );
          // note: 'condition' will be sorted in this method.
          if( rule.isCondition( condition ) ) {
              rules.removeFrom( i );
              retVal = true;
          }
          else
              i++;
     }
     return retVal;
  }

  /** Removes any rule which condition contains the specific chunk index.
   *  @param chunkIdx the chunk index.
   *  @return true if such rules exist, false otherwise. */
  public boolean removeCondition( int chunkIdx )
  {
     boolean retVal = false;
     int i = 0;
     while( i < rules.getCount() )
     {
          AssocRule rule = (AssocRule)( rules.getFrom( i ) );
          if( rule.isInCondition( chunkIdx ) )
          {
              rules.removeFrom( i );
              retVal = true;
          }
          else
              i++;
     }
     return retVal;
  }

  /** Draws conclusion triggered by the specific list of chunk indices as condition.
   *  @con the specific list of chunk indices. */
  public double drawConclusion( Vector con )
  {
     if( global.isRTOn() )
         assocRuleAppTime = 0;
     double support = 0.0;
     for( int i = 0; i < rules.getCount(); i++ ) {
          AssocRule rule = ( AssocRule )rules.getFrom( i );
          double[] weights = global.getTaskClarion().getAssocWeights( rule.getConditionNum() );
          rule.calculateSupport( con, weights );
          double val = rule.getSupport();

          //System.out.println( "the support = " + val );

          if( val > support )
              support = val;
          if( val > 0.0 ){  // current rule is evocated.
              rule.resetAge();
              if( global.isAssocBlaOn() && global.getAssocBlaOption() == Global.VARIED_OPTION ){
                  rule.calcBLA( currentTime );
                  if( global.isRTOn() )
                      assocRuleAppTime += rule.calcAssocRuleAppTime();
                  rule.increaseUsage( currentTime );
              }
          }
     }
     // get all of the activated rules
     if( emOn ){
         short[][] conditions = new short[rules.getCount()][];
         int activeRuleCount = 0;
         for( int i = 0; i < rules.getCount(); i++ ) {
              AssocRule rule = ( AssocRule )rules.getFrom( i );
              if( rule.getSupport() > threshold_ae ){
                  conditions[activeRuleCount] = new short[rule.getConditionNum()];
                  rule.getCondition( conditions[activeRuleCount++] );
              }
         }
         activeRuleConditions = new short[activeRuleCount][];
         for( int i = 0; i < activeRuleCount; i++ ){
              activeRuleConditions[i] = new short[conditions[i].length];
              for( int j = 0; j < conditions[i].length; j++ ){
                   activeRuleConditions[i][j] = conditions[i][j];

                   //System.out.println( "the rule condition = " + conditions[i][j] );
              }
         }
     }
     return support;
  }

  /** Searches a rule with the specific condition in terms fo an array of chunk indices.
   *  @param condition the rule condition. */
  public boolean searchRule( short[] condition )
  {
     boolean retVal = false;
     if( global.isRTOn() )
         assocRuleAppTime = 0;
     for( int i = 0; i < rules.getCount(); i++ ) {
          AssocRule rule = (AssocRule)( rules.getFrom( i ) );
          if( rule.isCondition( condition ) ){
              rule.resetAge();
              if( global.isAssocBlaOn() && global.getAssocBlaOption() == Global.VARIED_OPTION ){
                  rule.calcBLA( currentTime );
                  if( global.isRTOn() )
                      assocRuleAppTime += rule.calcAssocRuleAppTime();
                  rule.increaseUsage( currentTime );
              }
              retVal = true;
              break;
          }
     }
     return retVal;
  }

/*************************** CHUNK UPDATE METHODS *****************************/
  /** update routine. */
  public void update()
  {
     /*
     for( int i = 0; i < rules.getCount(); i++ )
     {
          AssocRule rule = (AssocRule)( rules.getFrom( i ) );
          rule.increaseAge();
          if( rule.getAge() >= MAX_RULE_AGE )
          {
              rules.removeFrom( i );
              i--;
          }
     }
     */
     int i = 0;
     while( i < rules.getCount() )
     {
          AssocRule rule = (AssocRule)( rules.getFrom( i ) );
          rule.increaseAge();
          if( rule.getAge() >= MAX_RULE_AGE ){
              System.out.println( "the removed rule = " + rule );

              rules.removeFrom( i );
          }
          else
              i++;
     }
  }

/************************** OVERRIDED METHODS ********************************/
  /** methods from class <em>Object</em> */
  public String toString()
  {
     String prefix = Global.TABLE_STR + Global.TABLE_STR + Global.TABLE_STR;
     String str = "";
     for( int i = 0; i < rules.getCount(); i++ ){
          str += prefix + "Rule_" + String.valueOf( i + 1 ) + " is: ";
          str += (AssocRule)( rules.getFrom( i ) );
     }
     return str;
  }

  /** methods from class <em>Object</em> */
  public Object clone() throws CloneNotSupportedException
  {
      AssocRuleGrp ruleGrp = new AssocRuleGrp( groupID, global );
      ruleGrp.setRules( rules );
      return ruleGrp;
  }

}