package clarion.nacs;
import java.util.Arrays;
import java.util.Vector;
import clarion.system.*;
import clarion.common.*;
import clarion.tools.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is to implement GKS part of the NACS subsystem. It implements the
 * learning and reasoning processes inside GKS described in CLARION model.
 * <li> summary of the important methods. <br>
 * <ol>
 * <li> public boolean reasoning(): the reasoning process of GKS decrbied in CLARION
 * model.
 * <li> public void bottomUp( int amnIdx, Feature[] amnResult ): the process of
 * bottom-up activation described in CLARION model.
 * <li> public boolean formResults( int retrievalMode ): combine the result from
 * both GKS and AMN.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * This class implements GKS part, the top level of NACS subsystem.
 * <li> Usage of this class. <br>
 * It is instantiated inside a <em>NACS</em> object.
 * </ul>
 * @author Xi Zhang
 */

public class GKS{

  final int AMN_IDX = 0, GKS_IDX = 1;
  final int SCALE_VAL = 100;   // used for selecting chunks to remove.
  final double ALMOST_ZERO = 0.00001;
  /** global settings. */
  Global global;
  /** task-specific CLARION settings. */
  TaskClarion taskClarion;
  /** reasoning type. */
  int    reasoningType;
  /** parameters for learning in NACS. */
  double p_c, p_a, p_ce, p_ae, threshold_ce, threshold_ae;
  /** used for similarity reasoning type. */
  double threshold_simi;
  // variables for reasoning process.
  /** the reasoning method: forward chaining or forward chaining wiht SBR. */
  int reasoningMethod;
  /** # of reasoning iterations to get result. */
  int iterationNum;
  /** chunk activations by cue. */
  double gksActivations[];
  /** chunk strengths from GKS. */
  double gksConcStrengths[];
  /** chunk strengths from AMN. */
  double amnConcStrengths[];
  /** combined chunk strengths by GKS and AMN. */
  double chunkStrengths[];
  /** used for MCS's minotoring conclusions. */
  int c21, c25;
  /** the final combined results.  */
  GKSChunk[] combResults;
  /** all retrievals from both GKS and AMN for monitoring.  */
  Chunk[] allResults;
  /** the top level retrievals for monitoring.  */
  Chunk[] gksResults;
  /** the bottom level retrievals for monitoring.  */
  Chunk[] amnResults;
  // variables for chunks.
  /** maximum # of chunks allowed.  */
  int MAX_CHUNK_NUM;
  /** # of chunk type.  */
  int chunkTypeNum;
  /** current # of existing chunks in GKS.  */
  int curChunkNum;
  /** the index of chosen chunk.  */
  int chosenChunkIdx;
  // variables for retrieval.
  /** cue used for retrieval.  */
  Feature cue[];
  /** the label of the cue.  */
  String  cueLabel;
  /** the activation of the cue.  */
  double  cueActivation;
  /** type of cue. */
  short   cueType;
  /** conclusion dimension index. @deprecated */
  short   concDimIdx;
  // variables for BLA.
  /** current time.  */
  long   currentTime;
  // other variables.
  /** list of current existing chunks.  */
  Vector existChunkList;
  /** list of current empty chunks. */
  Vector emptyChunkList;
  /** all of chunks in GKS. */
  GKSChunk chunks[];
  /** whether EM is on.  */
  boolean emOn;
  /** the # of acs networks. */
  int  acsNetNum;
  /** the # of amn nets. */
  int  amnNetNum;
  /** EM for storing episodic chunks.  */
  EM[] acsEMs, nacsEMs;
  /** used for determining the final result. */
  StochasticDecider nodeDecider;
  /** used for training AMNs. */
  GKSChunk nullConclusion;
  // for RT.
  /** response time. */
  int responseTime;
  /** chunk BLA is on or not. */
  boolean chunkBlaOn;
  /** chunk option: constatant or variable. */
  int chunkBlaOption;

  public GKS( Global g )
  {
      global = g;
      taskClarion = global.getTaskClarion();
      // initialize variables for reasoning.
      reasoningType = global.getGksReasoningMethod();
      p_c = global.getChunkEncodeProb();
      p_a = global.getAssocEncodeProb();
      p_ce = global.getChunkExtProb();
      p_ae = global.getAssocExtProb();
      threshold_ce = global.getChunkExtThreshold();
      threshold_ae = global.getAssocExtThreshold();
      threshold_simi = global.getThreshold_sim();
      reasoningMethod = global.getGksReasoningMethod();
      iterationNum = global.getGksReasonIterNum();
      // initialize variables for chunks.
      MAX_CHUNK_NUM = global.getGksMaxChunkNum();
      gksActivations = new double[ MAX_CHUNK_NUM ];
      gksConcStrengths = new double[ MAX_CHUNK_NUM ];
      amnConcStrengths = new double[ MAX_CHUNK_NUM ];
      // initialize the lists for managing chunks.
      existChunkList = new Vector( MAX_CHUNK_NUM );
      emptyChunkList = new Vector( MAX_CHUNK_NUM );
      for( int i = 0; i < MAX_CHUNK_NUM; i++ ){
          Integer elem = new Integer( i );
          emptyChunkList.addElement( elem );
      }
      // initialize the chunks.
      chunks = new GKSChunk[ MAX_CHUNK_NUM ];
      //if( global.isAssocEncodeExternOn() || global.isChunkEncodeExternOn() )
       //   encodeExternal();
      emOn = global.isEMon();
      if( emOn ){
          acsNetNum = global.getAcsNetNum();
          acsEMs = new EM[ acsNetNum ];
          for( int i = 0; i < acsNetNum; i++ )
               acsEMs[ i ] = new EM( i, global, Global.ACS );
          amnNetNum = global.getAmnNum();
          nacsEMs = new EM[ amnNetNum ];
          for( int i = 0; i < amnNetNum; i++ )
               nacsEMs[ i ] = new EM( i, global, Global.NACS );
      }
      // initialize the conclusions for monitoring.
      if( global.isMcsOn() ){
          c21 = global.getMonitorNacsTopConcNum();
          c25 = global.getMonitorNacsBottomConcNum();
      }
      // initialize node decider.
      nodeDecider = new StochasticDecider( global.getGksOutputTemperature() );
      // initialize RT variables.
      chunkBlaOn = global.isNacsChunkBlaOn();
      chunkBlaOption = chunkBlaOn ? global.getChunkBlaOption() : Global.EMPTY;
  }

  /** reinitialization routine. */
  public void reinit()
  {
      currentTime = 0;
      existChunkList.removeAllElements();
      emptyChunkList.removeAllElements();
      for( int i = 0; i < MAX_CHUNK_NUM; i++ )
      {
          Integer elem = new Integer( i );
          emptyChunkList.addElement( elem );
      }
      chunks = new GKSChunk[ MAX_CHUNK_NUM ];
  }

  public void resetStrengths()
  {
      CommonMethods.resetArray( gksActivations, 0.0 );
      CommonMethods.resetArray( gksConcStrengths, 0.0 );
      CommonMethods.resetArray( amnConcStrengths, 0.0 );
      combResults = null;
      allResults = null;
      gksResults = null;
      amnResults = null;
  }

  /** reset the response time. */
  public void resetRT()
  {
     responseTime = 0;
  }

/***************************** GET METHODS **********************************/
  /** Returns the conbined retrieval results.
   *  @return the conbined retrieval results. */
  public GKSChunk[] getCombinedResults()
  {
      return combResults;
  }

  /** Returns the top level results for monitor buffer.
   *  @return the top level results for monitor buffer. */
  public Chunk[] getTopLevelMonitorConclusions()
  {
     return gksResults;
  }

  /** Returns the bottom level results for monitor buffer.
   *  @return the bottom level results for monitor buffer. */
  public Chunk[] getBottomLevelMonitorConclusions()
  {
     return amnResults;
  }

  /** Returns the all results for monitor buffer.
   *  @return the all results for monitor buffer. */
  public Chunk[] getAllMonitorConclusions()
  {
     return allResults;
  }

  /** Returns the ACS samples in EM.
   *  @param netIdx index on an ACS network.
   *  @param trainType the training type: for ACS or for AEM.
   *  @return the ACS samples in EM. */
  public EMAcsSample[] getEmAcsSamples( int netIdx, int trainType )
  {
     return acsEMs[netIdx].getAcsSamples( trainType );
  }

  /** Returns the NACS samples in EM.
   *  @param netIdx index on network.
   *  @return the NACS samples in EM. */
  public EMNacsSample[] getEmNacsSamples( int netIdx )
  {
     return nacsEMs[netIdx].getNacsSamples();
  }

  /** Returns a specific chunk in GKS.
   *  @param idx the index on a specific chunk in GKS.
   *  @return a specific chunk in GKS. */
  public GKSChunk getChunk( int idx )
  {
     return (idx >= 0) ? chunks[idx] : null;
  }

  /** Returns an index on a chunk with the specific label in GKS.
   *  @param label the chunk label.
   *  @return an index on a specific chunk. */
  public int getChunkIdx( String label )
  {
     curChunkNum = existChunkList.size();
     if( curChunkNum == 0 ){
         System.out.println( "no chunk in GKS (label)!" );
         return -1;
     }
     for( int i = 0; i < curChunkNum; i++ ){
          int idx = ( (Integer)existChunkList.elementAt( i ) ).intValue();
          if( chunks[ idx ].getLabel().compareTo( label ) == 0 ){
              chunks[ idx ].resetAge();
              if( chunkBlaOption == global.VARIED_OPTION )
                  chunks[ idx ].increaseUsage( currentTime );
              return idx;
          }
     }
     System.out.println( "no chunk with the label: " + label );
     return -1;
  }

  /** Returns an index on a chunk with the specific features in GKS.
   *  @param features the chunk features.
   *  @return an index on a specific chunk. */
  public int getChunkIdx( Feature[] features )
  {
     curChunkNum = existChunkList.size();
     if( curChunkNum == 0 ){
         System.out.println( "no chunk in GKS (feature)!" );
         return -1;
     }
     for( int i = 0; i < curChunkNum; i++ ){
          int idx = ( (Integer)existChunkList.elementAt( i ) ).intValue();
          if( chunks[ idx ].isEqual( features ) ){
              chunks[ idx ].resetAge();
              if( chunkBlaOption == global.VARIED_OPTION )
                  chunks[ idx ].increaseUsage( currentTime );
              return idx;
          }
     }
     //System.out.println( "no chunk with the features !" );
     return -1;
  }

  public short[][] getConditions( int conclusion )
  {
     if( conclusion == Global.EMPTY || chunks[conclusion] == null )
         return null;
     return chunks[conclusion].getRuleConditions();
  }

  /** Returns the response time.
   *  @return the response time. */
  public int getRT()
  {
     return responseTime;
  }

/******************************** SET METHODS  ********************************/
  /** Sets the cue for retrieval.
   *  @param content the cue features. */
  public void setRetrievalCue( Feature[] content )
  {
     cueType = (short)global.CHK_CUE;
     cueLabel = global.CHUNK_TYPE_LABELS[cueType];
     cue = new Feature[ content.length ];
     for( int i = 0; i < content.length; i++ )
          cue[ i ] = content[ i ];
  }

  /** Sets the cue for retrieval.
   *  @param content the cue features.
   *  @param concIdx the index on conclusion dimension in the cue.
   *  @deprecated Replaced by setRetrievalCue( Feature[] content ). */
  public void setRetrievalCue( Feature[] content, int concIdx )
  {
     cueType = (short)global.CHK_CUE;
     cueLabel = global.CHUNK_TYPE_LABELS[cueType];
     cue = new Feature[ content.length ];
     for( int i = 0; i < content.length; i++ )
          cue[ i ] = content[ i ];
     concDimIdx = (short)concIdx;
  }

  /** Sets the response time.
   *  @param time the value to set. */
  public void setCurrentTime( long time )
  {
     currentTime = time;
     for( int i = 0; i < existChunkList.size(); i++ ) {
          int idx = ( (Integer)existChunkList.elementAt( i ) ).intValue();
          chunks[ idx ].setCurrentTime( currentTime );
     }
  }

  /** Sets strength for the specificed chunk.
   *  @param idx the chunk index.
   *  @param strength the strength to set. */
  public void setChunkStrength( int idx, double strength )
  {
     chunks[idx].setStrength( strength );
  }

  /** Sets the # of iterations of GKS reasoning.
   *  @param num the number to set. */
  public void setGksIterNum( int num )
  {
     iterationNum = num;
  }

  /** Sets the GKS reasoning methods.
   *  @param method the method to set. */
  public void setGksReasoningMethod( int method )
  {
     reasoningMethod = method;
     curChunkNum = existChunkList.size();
     for( int i = 0; i < curChunkNum; i++ ){
          int idx = ( (Integer)existChunkList.elementAt( i ) ).intValue();
          chunks[ idx ].setReasoningMethod( reasoningMethod );
     }
  }

/********************** INSERT AND REMOVE CHUNK METHODS ***********************/
  // Vector 'con' should include the feature information.
  /** insert a specific chunk into GKS.
   *  @param chunk the chunk to insert. */
  public int insertChunk( GKSChunk chunk )
  {
     int availableChunkIdx = 0;
     try{
         if( emptyChunkList.isEmpty() ){
             System.out.println( "max # of chunks is =" + MAX_CHUNK_NUM );

             double[] inputs = new double[ MAX_CHUNK_NUM ];
             for( int i = 0; i < MAX_CHUNK_NUM; i++ )
                  inputs[ i ] = ( (double)chunks[ i ].getAge() ) / SCALE_VAL;
             availableChunkIdx = nodeDecider.getRecommendation( inputs );
             chunks[ availableChunkIdx ] = (GKSChunk)(chunk.clone());
         }
         else{
             availableChunkIdx = (( Integer )( emptyChunkList.remove( 0 ) )).intValue();
             existChunkList.addElement( new Integer( availableChunkIdx ) );
             if( chunks[ availableChunkIdx ] == null )
                 chunks[ availableChunkIdx ] = (GKSChunk)(chunk.clone());
         }
         chunks[ availableChunkIdx ].setAvailableFlag( true );
     }
     catch( Exception e ){
            System.out.println( "max # of chunks is =" + MAX_CHUNK_NUM );

            System.out.println( e + " happens in clone GKS chunk in GKS.java" );
     }

     return availableChunkIdx;
  }

  /** insert a specific chunk into GKS.
   *  @param idnIdx the index on an IDN.
   *  @param amnIdx the index on an AMN.
   *  @param type chunk type.
   *  @param label chunk label.
   *  @param features the chunk features. */
  public int insertChunk( int idnIdx, int amnIdx, int type, String label, Feature[] features )
  {
     int availableChunkIdx = 0;

     if( emptyChunkList.isEmpty() ){
         //System.out.println( MAX_CHUNK_NUM );

         double[] inputs = new double[ MAX_CHUNK_NUM ];
         for( int i = 0; i < MAX_CHUNK_NUM; i++ )
              inputs[ i ] = ( (double)chunks[ i ].getAge() ) / SCALE_VAL;
         availableChunkIdx = nodeDecider.getRecommendation( inputs );
         String newLabel = (!label.equals("")) ? label : "chunk " + availableChunkIdx;
         chunks[ availableChunkIdx ] = new GKSChunk( availableChunkIdx, type,
                                           newLabel, features, idnIdx, amnIdx, global );
     }
     else{
         availableChunkIdx = (( Integer )( emptyChunkList.remove( 0 ) )).intValue();
         existChunkList.addElement( new Integer( availableChunkIdx ) );
         String newLabel = (!label.equals("")) ? label : "chunk " + availableChunkIdx;
         if( chunks[ availableChunkIdx ] == null )
             chunks[ availableChunkIdx ] = new GKSChunk( availableChunkIdx, type,
                                           newLabel, features, idnIdx, amnIdx, global );
     }
     chunks[ availableChunkIdx ].setAvailableFlag( true );
     return availableChunkIdx;
  }

  /** remove a chunk with the specific index.
   *  @param chunkIdx the index on the chunk. */
  public void removeChunk( int chunkIdx )
  {
     Integer elem = new Integer( chunkIdx );
     existChunkList.removeElement( elem );
     emptyChunkList.addElement( elem );

     // remove all the rules which has chunkIdx as one of its condition.
     for( int i = 0; i < existChunkList.size(); i++ )
     {
          int idx = ( (Integer)existChunkList.elementAt( i ) ).intValue();
          chunks[ idx ].removeCondition( chunkIdx );
     }

     chunks[ chunkIdx ].setLabel( "" );
     chunks[ chunkIdx ].setAvailableFlag( false );
  }

  /** insert an associative with the condition and conclusion.
   *  @param condition the condition of the rule in terms of the indices of chunks.
   *  @param conclusion the conclusion of the rule in terms of the index of the chunk. */
  public void insertRule( short[] condition, short conclusion )
  {
     if( chunks[ conclusion ] == null || !chunks[ conclusion ].getAvailableFlag() )
     {
         System.out.println( " cannot insert this rule !" );
         return;
     }

     chunks[ conclusion ].insertRule( condition, conclusion );
  }

  /*
  public void removeRule( short[] condition, short conclusion )
  {
     if( chunks[ conclusion ] == null || !chunks[ conclusion ].getAvailableFlag() )
         System.out.println( " cannot remove such rule !" );

     chunks[ conclusion ].removeRule( condition, conclusion );
  }
  */

/***************************** RETRIEVAL METHOD *******************************/
  /** the reasoning process of GKS decrbied in CLARION model. */
  public boolean reasoning()
  {
     if( global.isRTOn() )
         responseTime = 0;
     curChunkNum = existChunkList.size();
     if( curChunkNum == 0 )
         return false;
     for( int i = 0; i < gksConcStrengths.length; i++ ){
          gksActivations[ i ] = 0.0;
          gksConcStrengths[ i ] = 0.0;
     }
     // activate chunks.
     int activateCount = activateChunks( cue, GKS_IDX );
     /*
     System.out.println( "activated num = " + activateCount );
     System.out.println( "gks activation = " );
     for( int i = 0; i < 20; i++ ){
          System.out.println( "chunk " + i + " = " + gksActivations[i] );
     }
     */

     // no activated chunks.
     if( activateCount == 0 )
         return false;
     //draw conclusion.
     drawConclusion();
     // fill in top level results.
     gksResults = fillInFinalResults( GKS_IDX, gksConcStrengths );

     /*
     System.out.println( "gks strengths = " );
     for( int i = 16; i < 20; i++ ){
          System.out.println( "chunk " + i + " = " + gksConcStrengths[i] );
     }
     */
     /*
     System.out.println( "gks conclusions = " );
     for( int i = 0; i < gksResults.length; i++ ){
          System.out.println( gksResults[i] );
     }
     */

     return true;
  }

  /** the process of bottom-up activation.
   *  @param amnIdx the index on an AMN.
   *  @param amnResult the retrieval result from the AMN. */
  public void bottomUp( int amnIdx, Feature[] amnResult )
  {
     /*
     System.out.println( "the amn retrieval result = " );
     for( int i = 0; i < amnResult.length; i++ )
          System.out.println( amnResult[i] );
     */

     for( int i = 0; i < amnConcStrengths.length; i++ )
          amnConcStrengths[ i ] = 0.0;
     curChunkNum = existChunkList.size();
     // activate all of the compatible chunks.
     activateChunks( amnResult, AMN_IDX );
     // fill in bottom level results.
     amnResults = fillInFinalResults( AMN_IDX, amnConcStrengths );

     /*
     System.out.println( "amn strengths = " );
     for( int i = 16; i < 20; i++ ){
          System.out.println( "chunk " + i + " = " + amnConcStrengths[i] );
     }
     */
     /*
     System.out.println( "amn  conclusions = " );
     for( int i = 0; i < amnResults.length; i++ ){
          System.out.println( amnResults[i] );
     }
     */

     // extract the explicit knowledge.
     extractExplicitKnowledge( amnIdx, amnResult );
  }

  /** combine the result from both GKS and AMN.
   *  @param retrievalMode the retrieval mode: retrieval all chunks or only one chunk. */
  public boolean formResults( int retrievalMode, int retrievalChunkType )
  {
     curChunkNum = existChunkList.size();
     combResults = null;
     allResults = null;
     // get results.
     chunkStrengths = new double[ curChunkNum ];
     short[] chunkIndices = new short[ curChunkNum ];
     for( int i = 0; i < curChunkNum; i++ ) {
          chunkStrengths[ i ] = 0.0;
          chunkIndices[ i ] = -1;
     }
     int concNum = 0;
     double threshold_r = global.getReasonThreshold();
     for( int i = 0; i < curChunkNum; i++ ){
          int idx = ( (Integer)existChunkList.elementAt( i ) ).intValue();
          double maxValue = Math.max( gksConcStrengths[idx], amnConcStrengths[idx] );
          if( maxValue > threshold_r && (retrievalChunkType == Global.ALL_CHK_TYPES ||
              retrievalChunkType == chunks[idx].getChunkType() ) ){
              chunkStrengths[ concNum ] = maxValue;
              chunkIndices[ concNum++ ] = (short)idx;
          }
     }
     // no conclusion has been drawn.
     if( concNum == 0 )
         return false;
     // fill in all the results from GKS and AMN for monitoring.
     allResults = fillInFinalResults( concNum, chunkStrengths, chunkIndices );

     /*
     System.out.println( "final strengths = " );
     for( int i = 0; i < 4; i++ ){
          System.out.println( "chunk " + (i+16) + " = " + chunkStrengths[i] );
     }
     */

     // obtain drawn conclusions.
     try{
         if( retrievalMode == Global.RETRIEVAL_ALL ){
             combResults = new GKSChunk[ concNum ];
             for( int i = 0; i < concNum; i++ ){
                  int idx = chunkIndices[ i ];
                  combResults[ i ] = (GKSChunk)(chunks[ idx ].clone());
                  combResults[ i ].setStrength( chunkStrengths[i] );
             }
         }
         if( retrievalMode == Global.RETRIEVAL_ONE ){
             combResults = new GKSChunk[ 1 ];
             int idx = nodeDecider.getRecommendation( chunkStrengths, 0, concNum );
             combResults[ 0 ] = (GKSChunk)(chunks[ chunkIndices[idx] ].clone());
             combResults[ 0 ].setStrength( chunkStrengths[idx] );
         }
     }
     catch( CloneNotSupportedException e ){
         System.out.println( "clone error " );
     }
     return true;
  }

  public boolean searchConditions( short[] conditions, short conclusion )
  {
    if( conclusion == Global.EMPTY || chunks[conclusion] == null )
         return false;
    int retrievalTime = 0;
    boolean flag = chunks[ conclusion ].searchRule( conditions );
    if( flag ){
        // retrieve conclusions.
        chunks[conclusion].resetAge();
        chunks[conclusion].calcRetrievalTime();
        retrievalTime += chunks[conclusion].getRetrievalTime()+chunks[conclusion].getAssocRuleAppTime();
    }
    if( global.isRTOn() )
        responseTime += retrievalTime;
    return flag;
  }

  /** search all of the conclusions with the specific condition in terms of indices
   *  of the chunks.
   *  @param condition the condition to trigger the search. */
  public Vector searchConclusions( short[] conditions )
  {
     curChunkNum = existChunkList.size();
     if( curChunkNum == 0 ){
         System.out.println( "No available chunks!" );
         return (Vector)null;
     }
     // retrieve the conditions.
     int retrievalTime = 0;
     for( int i = 0; i < conditions.length; i++ ) {
          //System.out.println("condtion=" + conditions[i]);
          if( conditions[i] >= 0 ){     // new revision. 111/01/04
              chunks[conditions[i]].resetAge();
              chunks[conditions[i]].calcRetrievalTime();
              retrievalTime += chunks[conditions[i]].getRetrievalTime();
          }
     }
     int concNum = 0;
     Vector vect = new Vector( curChunkNum );
     for( int i = 0; i < curChunkNum; i++ ){
          int idx = ( (Integer)existChunkList.elementAt( i ) ).intValue();
          if( chunks[ idx ].searchRule( conditions ) ){
              // retrieve conclusions.
              chunks[idx].resetAge();
              chunks[idx].calcRetrievalTime();
              retrievalTime += chunks[idx].getRetrievalTime()+chunks[idx].getAssocRuleAppTime();
              vect.addElement( new Integer(idx) );
              concNum++;
          }
     }
     Vector conclusions = new Vector( concNum );
     Vector sands = new Vector( concNum );
     for( int i = 0; i < concNum; i++ )
          sands.addElement( new Integer( i ) );
     for( int i = 0; i < concNum; i++ ){
          int idx = ((Integer)sands.remove((int)(Math.random()*sands.size()))).intValue();
          conclusions.addElement( vect.elementAt( idx ) );
     }
     if( global.isRTOn() )
         responseTime += retrievalTime;
     return conclusions;
  }

/***************************** GKS UPDATE METHODS *****************************/
  /** update GKS. */
  public void update()
  {
     curChunkNum = existChunkList.size();
     if( curChunkNum == 0 ){
         //System.out.println( "No available chunks!" );
         return;
     }
     // search all the chunks should be removed.
     int removeNum = 0;
     int[] removeChunks = new int[ curChunkNum ];
     for( int i = 0; i < curChunkNum; i++ ){
          int idx = ( (Integer)existChunkList.elementAt( i ) ).intValue();
          boolean needRemoved = chunks[ idx ].update();
          if( needRemoved ){
              //System.out.println("remove chunk = " + chunks[idx] );
              removeChunks[ removeNum++ ] = idx;
          }
     }
     // remove all of the less used chunks.
     for( int i = 0; i < removeNum; i++ )
          removeChunk( removeChunks[ i ] );
     // update EM if exist.
     if( emOn )
         checkEM();
  }

  /** add chunk with the specific type and index into EM.
   *  @param subsys the chunk comes from.
   *  @param idx the index on the chunk. */
  public void updateEM( int subsys, int idx )
  {
     if( !emOn )
         return;
     if( subsys == Global.NACS ){
         int amnIdx = chunks[idx].getAmnIdx();
         if( amnIdx != global.EMPTY )
             nacsEMs[ amnIdx ].insertItem( chunks[idx] );
     }
     if( subsys == Global.ACS ){
         int idnIdx = chunks[idx].getIdnIdx();
         if( idnIdx != global.EMPTY )
             acsEMs[ idnIdx ].insertItem( chunks[idx] );
     }
  }

  /** add associative rules with the specific condition and conclusion in terms of
   *  features into EM.
   *  @param subsys ACS or NACS.
   *  @param netIdx index to an network.
   *  @param condition the condition features.
   *  @param conclusion the conclusion features. */
  public void updateEM( int subsys, int netIdx, Feature[] condition, Feature[] conclusion )
  {
     if( !emOn )
         return;
     Chunk[] ruleChunk = new Chunk[2];
     if( subsys == Global.NACS ){
         ruleChunk[0] = new Chunk( Global.CHK_CUE, Global.CHUNK_TYPE_LABELS[global.CHK_CUE], condition,
                                   global );
         ruleChunk[1] = new Chunk( Global.CHK_RETRIEVAL, Global.CHUNK_TYPE_LABELS[global.CHK_RETRIEVAL],
                                   conclusion, global );
         nacsEMs[netIdx].insertItem( new Chunk( Global.CHK_ASSOC, Global.CHUNK_TYPE_LABELS[global.CHK_ASSOC],
                                     ruleChunk, global) );
     }
     if( subsys == Global.ACS ){
         ruleChunk[0] = new Chunk( Global.CHK_STATE, Global.CHUNK_TYPE_LABELS[global.CHK_ASSOC], condition,
                                   global );
         ruleChunk[1] = new Chunk( Global.CHK_ACTION, Global.CHUNK_TYPE_LABELS[global.CHK_ASSOC], conclusion,
                                   global );
         acsEMs[netIdx].insertItem( new Chunk( Global.CHK_ACTION_RULE, Global.CHUNK_TYPE_LABELS[global.CHK_ACTION_RULE],
                                    ruleChunk, global) );
     }
  }

/**************************** OVERRIDED METHODS *****************************/
  /** methods from class <em>Object</em> */
  public String toString()
  {
     curChunkNum = existChunkList.size();

     String str = "GKS( # of chunks = " + String.valueOf( curChunkNum ) + " ) = \n";
     str += "{ \n";
     for( int i = 0; i < curChunkNum; i++ ){
          int idx = ( (Integer)existChunkList.elementAt( i ) ).intValue();
          str += chunks[ idx ];
     }
     str += "} \n";

     return str;
  }

/**************************** OTHER METHODS *****************************/
  /** compute the BLA of the chosen chunk. */
  public void calcChosenChunkBLA()
  {
     chunks[ chosenChunkIdx ].calcBLA( currentTime );
  }

  /** Fills in the EM with the chunks and associative rules in GKS for assimilating
   *  the explicit knowledge. */
  public void fillInEmForAssimilation()
  {
     createNullConclusion();
     curChunkNum = existChunkList.size();
     for( int i = 0; i < curChunkNum; i++ ){
          int idx = ( (Integer)existChunkList.elementAt( i ) ).intValue();
          short[][] conditions = chunks[idx].getRuleConditions();
          if( conditions != null )
              updateEM( conditions, idx );
          else{
              conditions = new short[1][1];
              conditions[0][0] = (short)idx;
              updateEM( conditions, -1 );
          }
     }
  }

/******************************* PRIVATE METHODS ******************************/
  /* encode the external given knowledge: chunks and associative rules. */
  private void encodeExternal()
  {
     GKSChunk[] gksChunks = taskClarion.getExternalNacsChunks();
     for( int i = 0; i < gksChunks.length; i++ ){
          if( Math.random() < p_c )
              this.insertChunk( gksChunks[i] );

          //System.out.println( gksChunks[i] );
     }
     /*
     for( int i = 0; i < assocRuleGroups.length; i++ ){
          int chunkIdx = assocRuleGroups[i].getGroupID();
          chunks[ chunkIdx ].setRules( assocRuleGroups[i] );

          System.out.println("the rule group " + assocRuleGroups[i] );
     }
     */
  }

  /** create the null conclusion chunk. */
  private void createNullConclusion()
  {
     int featureNum = global.getGksOutputDimNum();
     Feature[] featList = new Feature[ featureNum ];
     for( int i = 0; i < featureNum; i++ ){
          int featType = global.getGksInputFeatureInfo()[i][0];
          int featOrder = global.getGksInputFeatureInfo()[i][1];
          int featLength = global.getFeatureLength( featType, featOrder );
          String featLabel = global.getFeatureNames( featType, featOrder );
          short[] content = new short[ featLength ];
          for( int j = 0; j < content.length; j++ )
               content[j] = (short)0;
          featList[i] = new Feature( featLabel, (short)featType, (short)featOrder, content, global );
     }
     nullConclusion = new GKSChunk( curChunkNum, Global.CHK_ASSOC, "null conclusion", featList, 0, 0, global );
  }

  /** activate chunks triggered by features in a specific level.
   *  @param features the features to trigger the activation.
   *  @param whichLevel the level: GKS or AMN. */
  private int activateChunks( Feature[] features, int whichLevel )
  {
     int activateCount = 0;
     for( int i = 0; i < curChunkNum; i++ ){
          int idx = ( (Integer)existChunkList.elementAt( i ) ).intValue();
          // if the chunk is activated.
          boolean bVal = chunks[ idx ].activateChunk( features, whichLevel );
          // the chunk is evocated, set strength of the activated chunk.
          if( bVal ) {
              if( whichLevel == GKS_IDX ) {   // GKS.
                  //gksConcStrengths[ idx ] = chunks[ idx ].getStrength();
                  gksActivations[ idx ] = chunks[ idx ].getStrength();
                  //if( emOn )
                  //    updateEM( Global.NACS, idx );
              }
              else {   // AMN.
                  amnConcStrengths[ idx ] = chunks[ idx ].getStrength();
                  //if( emOn && amnConcStrengths[idx] > threshold_ce )
                  //    updateEM( Global.NACS, idx );
              }
              activateCount++;
              // the chunk is evocated, reset age.
              chunks[ idx ].resetAge();
              // the chunk is evocated, record this time being used.
              // update RT, need further change.
              if( global.isRTOn() ){
                  chunks[idx].calcRetrievalTime();
                  int rt = chunks[idx].getRetrievalTime();
                  if( responseTime < rt )
                      responseTime = rt;
              }
              if( chunkBlaOption == global.VARIED_OPTION )
                  chunks[ idx ].increaseUsage( currentTime );
          }
     }
     return activateCount;
  }

  /** draw conclusions. */
  private void drawConclusion()
  {
     Vector deducedChunks = new Vector( curChunkNum );    // deduced chunks.
     Vector newDeducedChunks = new Vector( curChunkNum ); // new deduced chunks.
     // load all of the activated chunks.
     for( int i = 0; i < curChunkNum; i++ ){
          int idx = ( (Integer)existChunkList.elementAt( i ) ).intValue();
          if( Math.abs( gksActivations[idx] - 1.0 ) < ALMOST_ZERO ||
              ( global.isSimilarityOn() && gksActivations[idx]>0.0 ) )
             deducedChunks.addElement( new Integer( idx ) );
          chunks[idx].setInEmOn( false );
     }
     // draw conclusions.
     for( int i = 0; i < iterationNum; i++ ){
          newDeducedChunks.removeAllElements();
          int RT = 0;
          for( int j = 0; j < curChunkNum; j++ ){
               int idx = ((Integer)existChunkList.elementAt( j )).intValue();
               // calculate support from rule-based and similarity-based.
               if( chunks[idx].getRuleNum() == 0 )
                   continue;
               chunks[ idx ].calculateStrength( deducedChunks );
               double strength = chunks[ idx ].getStrength();
               // if a new deduced chunk or not.
               if( strength > gksConcStrengths[ idx ] ){
                   // new deduced chunk, the chunk is evocated, reset age.
                   if( gksConcStrengths[ idx ] == 0.0 ){
                       newDeducedChunks.addElement( new Integer( idx ) );
                       chunks[ idx ].resetAge();
                   }
                   chunks[ idx ].setStrength( strength );
                   gksConcStrengths[ idx ] = strength;
                   // update EM if EM is on.
                   if( emOn && strength > threshold_ae && !chunks[idx].isInEmOn() ){
                       chunks[ idx ].setInEmOn( true );
                       short[][] activeConditions = chunks[idx].getActiveRuleConditions();
                       updateEM( activeConditions, idx );
                   }
                   // update RT, need further change.
                   if( global.isRTOn() ){
                       chunks[idx].calcRetrievalTime();
                       int rt = chunks[idx].getRetrievalTime()+chunks[idx].getAssocRuleAppTime();
                       if( RT < rt )
                           RT = rt;
                   }
                   if( chunkBlaOption == global.VARIED_OPTION )
                       chunks[ idx ].increaseUsage( currentTime );
               }
          }
          // no new chunk is concluded.
          if( newDeducedChunks.size() == 0 )
              break;
          deducedChunks.removeAllElements();
          for( int j = 0; j < newDeducedChunks.size(); j++ )
               deducedChunks.addElement( newDeducedChunks.elementAt(j) );
          if( global.isRTOn() )
              responseTime += RT;
     }
     // update conclusion strengths.
     for( int j = 0; j < curChunkNum; j++ ){
          int idx = ((Integer)existChunkList.elementAt( j )).intValue();
          if( chunks[ idx ].getRuleNum() == 0 )
              gksConcStrengths[idx] = 0.0;
     }
  }

  /** extract the explicit knowledge activated by the result from the specific AMN.
   *  @param amnIdx the index on a specific AMN.
   *  @param amnResult the result from the specific AMN. */
  private void  extractExplicitKnowledge( int amnIdx, Feature[] amnResult )
  {
     // chunk from AMN is not allowed.
     if( !global.isChunkFromAmnOn() || !global.isAssocExtractOn() )
         return;
     // calculate the activation of AMN result.
     double amnActivation = calcActivation( amnResult );
     cueActivation = calcActivation( cue );
     int cueChunkIdx = -1;
     int amnResultChunkIdx = -1;
     int amnCombChunkIdx = -1;
     if( amnActivation > threshold_ce )
     {   //  chunk extraction considered.
         if( Math.random() < p_ce ){
             // extract cue.
             cueChunkIdx = getChunkIdx( cue );
             if( cueChunkIdx == -1 )
                 cueChunkIdx = insertChunk( global.EMPTY, amnIdx, cueType, cueLabel, cue );
             // extract AMN result.
             amnResultChunkIdx = getChunkIdx( amnResult );
             if( amnResultChunkIdx == -1 ){
                 int resultType = global.CHK_RETRIEVAL;
                 String resultLabel = global.CHUNK_TYPE_LABELS[resultType];
                 amnResultChunkIdx = insertChunk( global.EMPTY, amnIdx, resultType, resultLabel, amnResult );
             }
             // extract cue and AMN result together.
             Feature[] amnCombResult = new Feature[ cue.length + amnResult.length ];
             for( int i = 0; i < cue.length; i++ )
                  amnCombResult[ i ] = (Feature)(cue[ i ].clone());
             for( int i = 0; i < amnResult.length; i++ )
                  amnCombResult[ cue.length + i ] = (Feature)(amnResult[ i ].clone());
             amnCombChunkIdx = getChunkIdx( amnCombResult );
             if( amnCombChunkIdx == -1 ) {
                 int combType = global.CHK_COMB;
                 String combLabel = global.CHUNK_TYPE_LABELS[combType];
                 amnCombChunkIdx = insertChunk( global.EMPTY, amnIdx, combType, combLabel, amnCombResult );
             }
         }
     }
     if( !global.isAssocExtractOn() )
         return;
     if( amnActivation > threshold_ae )
     {  //  associative rule extraction considered.
         if( Math.random() < p_ae ){
             // extract cue.
             if( cueChunkIdx == -1 )
                 cueChunkIdx = insertChunk( global.EMPTY, amnIdx, cueType, cueLabel, cue );
             // extract AMN result.
             if( amnResultChunkIdx == -1 ){
                 int resultType = global.CHK_RETRIEVAL;
                 String resultLabel = global.CHUNK_TYPE_LABELS[resultType];
                 amnResultChunkIdx = insertChunk( global.EMPTY, amnIdx, resultType, resultLabel, amnResult );
             }
             // associative rule extraction
             curChunkNum = existChunkList.size();
             for( int i = 0; i < curChunkNum; i++ ) {
                  int idx = ((Integer)existChunkList.elementAt( i )).intValue();
                  if( amnConcStrengths[ idx ] > 0.0 ){
                      short[] condition = new short[1];
                      condition[ 0 ] = (short)cueChunkIdx;
                      chunks[ idx ].insertRule( condition, (short)idx );
                  }
             }
         }
     }
  }

  /** calculate the activation of the features. */
  private double calcActivation( Feature[] features )
  {
     double actVal = 0.0;
     int activationMode = global.getActivationMeasure();

     if( activationMode == 0 )  // minimum measurement.
     {
         actVal = 1.0;
         for( int i = 0; i < features.length; i++ )
         {
              double[] activations = features[ i ].getActivations();
              for( int j = 0; j < activations.length; j++ )
                   if( activations[ j ] > 0.0 && activations[ j ] < actVal )
                       actVal = activations[ j ];
         }
     }
     else  // average measurement.
     {
         for( int i = 0; i < features.length; i++ )
         {
              int count = 0;
              double val = 0.0;
              double[] activations = features[ i ].getActivations();
              for( int j = 0; j < activations.length; j++ )
              {
                   if( activations[ j ] > 0.0 )
                   {
                       val += activations[ j ];
                       count++;
                   }
              }
              if( count == 0 )
                  count++;
              val /= count;

              actVal += val;
         }
         actVal /= features.length;
     }
     return actVal;
  }

  /** add associative rules with the specific condition and conclusion in terms of
   *  chunks into EM.
   *  @param condition the condition chunk.
   *  @param conclusion the conclusion chunk. */
   /*
  private void updateEM( GKSChunk condition, GKSChunk conclusion )
  {
     if( !emOn )
         return;
     GKSChunk[] ruleChunk = new GKSChunk[2];
     try{
         ruleChunk[0] = (GKSChunk)(condition.clone());
         ruleChunk[1] = (GKSChunk)(conclusion.clone());
         nacsEMs[condition.getAmnIdx()].insertItem(
             new Chunk(Global.CHK_ASSOC,Global.CHUNK_TYPE_LABELS[global.CHK_ASSOC],ruleChunk,global) );
     }
     catch( Exception e ){
     }
  }
  */

  /** add associative rules with the specific condition and conclusion in terms of
   *  indices of chinks into EM.
   *  @param condition the condition chunk.
   *  @param conclusion the conclusion chunk. */
  private void updateEM( short[][] condIndices, int concIndex )
  {
     if( !emOn )
         return;
     try{

       //System.out.println( "the condition indices len = " + condIndices.length );

       for( int i = 0; i < condIndices.length; i++ ){
            int chunkNum = condIndices[i].length + 1;
            GKSChunk[] ruleChunk = new GKSChunk[chunkNum];

            //System.out.println( "the condition chunk idx = " + condIndices[i][0] );

            //condition chunks.
            for( int j = 0; j < chunkNum-1; j++ )
                 ruleChunk[j] = (GKSChunk)(chunks[condIndices[i][j]].clone());
            //conclusion chunks.
            if( concIndex >= 0 )
                ruleChunk[chunkNum-1] = (GKSChunk)(chunks[concIndex].clone());
            else
                ruleChunk[chunkNum-1] = nullConclusion;
            int amnIdx = concIndex >= 0 ? chunks[concIndex].getAmnIdx() : 0;
            //combined chunk.
            Chunk chunk = new Chunk( Global.CHK_ASSOC,Global.CHUNK_TYPE_LABELS[Global.CHK_ASSOC],
                                     ruleChunk, global );
            nacsEMs[amnIdx].insertItem( chunk );
       }

       //if( concIndex < 0 )
       //    System.out.println( "null conclusion is performed" );

     }
     catch( Exception e ){
            System.out.println( e );
     }
  }

  /** check EM if activations of some chunks are lower than the threshold. */
  public void checkEM()
  {
     for( int i = 0; i < acsEMs.length; i++ )
          acsEMs[i].checkValid();
     for( int i = 0; i < nacsEMs.length; i++ )
          nacsEMs[i].checkValid();
  }

  /** Returns top conclusions in terms of strengths from the specific level.
   *  @param level the level: GKS or AMN.
   *  @param strengths the strengths of the result chunks. */
  private Chunk[] fillInFinalResults( int level, double[] strengths )
  {
     int resultNum = 0;
     //double threshold = (global.isMcsOn()) ? 0.0 : global.getReasonThreshold();
     double threshold = global.getReasonThreshold();
     for( int i = 0; i < strengths.length; i++ )
          if( strengths[i] > threshold )
              resultNum++;
     SortableElement[] sortableElements = new SortableElement[ resultNum ];
     int num = 0;
     for( int i = 0; i < strengths.length; i++ ){
          if( strengths[i] > threshold )
              sortableElements[num++] = new SortableElement( strengths[i], i );
     }
     Arrays.sort( sortableElements );
     int c = (level == GKS_IDX)?Math.min( c21, resultNum ):Math.min( c25, resultNum );
     int len = (global.isMcsOn())? c : resultNum;
     Chunk[] results = new Chunk[ len ];
     try{
         for( int i = 0; i < len; i++ ){
              int idx =  sortableElements[resultNum-i-1].getChunkIdx();
              results[i] = (GKSChunk)(chunks[idx].clone());
              results[i].setStrength( strengths[idx] );
         }
     }
     catch( Exception e ){
            System.out.println( e + ", clone chunk error in GKS.java." );
     }
     return results;
  }

  /** Returns top conclusions in terms of strengths from the specific level.
   *  @param concNum the number of result chunks.
   *  @param strengths the strengths of the result chunks.
   *  @param indices the indices of the result chunks*/
  private Chunk[] fillInFinalResults( int concNum, double[] strengths, short[] indices )
  {
     int resultNum = concNum;
     SortableElement[] sortableElements = new SortableElement[ resultNum ];
     for( int i = 0; i < resultNum; i++ )
          sortableElements[i] = new SortableElement( strengths[i], indices[i] );
     Arrays.sort( sortableElements );
     int c = Math.min( Math.max(c21, c25), resultNum );
     int len = (global.isMcsOn())? c : resultNum;

     //System.out.println( "result len = " + len );

     Chunk[] results = new Chunk[ len ];
     try{
         for( int i = 0; i < len; i++ ){
              int idx =  sortableElements[resultNum-i-1].getChunkIdx();
              results[i] = (GKSChunk)(chunks[idx].clone());
              double strength = sortableElements[resultNum-i-1].getKeyValue();
              results[i].setStrength( strength );
         }
     }
     catch( Exception e ){
            System.out.println( e + ", clone chunk error in GKS.java." );
     }
     return results;
  }
}
