package clarion.nacs;
import java.util.Vector;
import java.util.Arrays;
import clarion.system.*;
import clarion.common.*;
import clarion.tools.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is to implement the NACS subsystem of CLARION model.
 * <li> summary of the important methods. <br>
 * <ol>
 * <li> public void observeState( Feature[] state ): observes and decompose current
 * state.
 * <li> public void run( int idx ): run the learning/reasoning processes in NACS.
 * <li> public void update( short[][][] chosenActions, Feature[] nState, double[] rewards ):
 * update the NACS subsystem.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * This class implements NACS subsystem and on the top of NACS implementation
 * structure and all the other classes relevant to NACS is declared inside it.
 * <li> Usage of this class. <br>
 * It is instantiated inside a <em>Clarion</em> object.
 * </ul>
 * @author Xi Zhang
 *
 */

public class NACS {

  final int BL = Global.BL, TL = Global.TL, COMB = 2;
  final double ALMOST_ZERO = 0.0000001;
  /** global settings. */
  protected Global global;
  /** the GKS component. */
  protected GKS    gks;
  /** the AMN component. */
  protected AMNet  amns[];
  /** the index on the active AMN. */
  protected int   activeAmnIdx;
  /** the number of ACS network. */
  protected int   acsNetNum;
  /** the numbers of dimensions in ACS state. */
  protected int[] acsStateDimNums;
  /** the numbers of dimensions in ACS action. */
  protected int[] acsActionDimNums;
  /** the numbers of dimensions in ACS reward. */
  protected int[] acsRewardDimNums;
  /** the nubmers of dimensions in a step chunk. */
  protected int[] acsStepFeatNums;
  /** flag for experience-specific chunks is on or not. */
  protected boolean esChunkOn;
  /** flag for AEM is on or not. */
  protected boolean aemOn;
  /** flag for allowing multiple active AMN or not. */
  protected boolean amnMultiFlag;
  /** the probability of EM training AEM. */
  protected double  aemTrainedByEmProb;
  /** the AEM components. */
  protected AbsEM  absEMs[];
  /** the active dim indices to ACS input or output. */
  protected short[] nacsActiveInputDimInfo;
  /** the active dim indices to ACS input or output. */
  protected short[] nacsActiveOutputDimInfo;
  /** current active input features to NACS. */
  protected Feature[] nacsActiveInput;
  /** current active output features from NACS. */
  protected Feature[] nacsActiveOutput;
  /** current state in format of an array of features. */
  protected Feature[] curState;
  /** current action in format of an array of features. */
  protected Feature[] curAction;
  /** next state in format of an array of features. */
  protected Feature[] nextState;
  /** current reinforcement in format of an array of features. */
  protected Feature[] reward;
  /** retrieval results in format of an array of GKSChunks. */
  //protected GKSChunk results[];
  protected Object results[];
  /** used to compose condition chunks. */
  protected short[][] conditionDimInfo;

  public NACS(){}

  public NACS( Global g )
  {
      global = g;
      global.setNACS( this );
      gks = new GKS( global );
      amns = new AMNet[ global.getAmnNum() ];
      for( int i = 0; i < amns.length; i++ )
           amns[ i ] = new AMNet( i, global, gks );
      amnMultiFlag = global.isAmnMultiFlagOn();
      // initialize variables relative to acs.
      acsNetNum = global.getAcsNetNum();
      acsStateDimNums = new int[ acsNetNum ];
      acsActionDimNums = new int[ acsNetNum ];
      acsRewardDimNums = new int[ acsNetNum ];
      acsStepFeatNums = new int[ acsNetNum ];
      for( int i = 0; i < acsNetNum; i++ ){
           acsStateDimNums[i] = global.getIdnInputDimVals(i).length;
           acsActionDimNums[i] = global.getIdnOutputDimVals(i).length;
           acsRewardDimNums[i] = 1;
           if( global.getIdnLearningOption(i) == global.QL )
               acsStepFeatNums[i]= 2*acsStateDimNums[i]+acsActionDimNums[i]+acsRewardDimNums[i];
           else
               acsStepFeatNums[i]= acsStateDimNums[i]+acsActionDimNums[i]+acsRewardDimNums[i];
      }
      esChunkOn = global.isESChunkOn();
      aemOn = global.isAEMon();
      if( aemOn ){
          aemTrainedByEmProb = global.getAemTrainedByEmProb();
          absEMs = new AbsEM[ acsNetNum ];
          for( int i = 0; i < acsNetNum; i++ )
               absEMs[ i ] = new AbsEM( i, global );
      }
  }

  /** check if the process of retrieval is done or not at this step. */
  public boolean isRetrieved()
  {
     return true;
  }

  /** Returns the retrieval results.
   *  @return the retrieval results. */
  public Object[] getResults()
  {
      return results;
  }

  /** Returns the all of the conclusions from the monitor buffer.
   *  @return the all of the conclusions from the monitor buffer. */
  public Chunk[][] getAllConclusions()
  {
      Chunk[][] chunks = new Chunk[3][];
      chunks[TL] = gks.getTopLevelMonitorConclusions();
      chunks[BL] = gks.getBottomLevelMonitorConclusions();
      chunks[COMB] = gks.getAllMonitorConclusions();
      /*
      System.out.println( "all conclusions from GKS = " );
      for( int i = 0; i < chunks[TL].length; i++ )
           System.out.println( chunks[TL][i] );

      System.out.println( "all conclusions from AMN = " );
      for( int i = 0; i < chunks[BL].length; i++ )
           System.out.println( chunks[BL][i] );
       */

      return chunks;
  }

  /** Returns AEM samples.
   *  @param netIdx the index on an ACS net.
   *  @return AEM samples. */
  public AbsEMSample[] getAemSamples( int netIdx )
  {
      return absEMs[ netIdx ].getSamples();
  }

  /** Returns ACS samples in EM.
   *  @param netIdx the index on an ACS net.
   *  @param trainType the train type: for ACS or AEM.
   *  @return ACS samples in EM. */
  public EMAcsSample[] getEmAcsSamples( int netIdx, int trainType )
  {
      return gks.getEmAcsSamples( netIdx, trainType );
  }

  /** Returns NACS samples in EM.
   *  @param netIdx the index on an AMN net.
   *  @return NACS samples in EM. */
  public EMNacsSample[] getEmNacsSamples( int netIdx )
  {
      return gks.getEmNacsSamples( netIdx );
  }

  /** Returns the response time. */
  public int getRT()
  {
     //System.out.println( "GKS RT = " + gks.getRT() );
     //System.out.println( "AMN RT = " + amns[activeAmnIdx].getRT() );
     return gks.getRT() + amns[activeAmnIdx].getRT();
  }

  /** Returns the GKS component. */
  public GKS getGks()
  {
     return gks;
  }

  /** Returns the GKS component. */
  public AMNet[] getAmns()
  {
     return amns;
  }

  /** Sets current time. */
  public void setCurrentTime( long curTime )
  {
     gks.setCurrentTime( curTime );
  }

/********************* SET METHODS FOR CUE PARAMETERS ***********************/
  /** Sets the cue as input to NACS.
   *  @deprecated Replaced by setGksCue( Feature[] cue ). */
  public void setGksCue( Feature[] cue, int concIdx )
  {
     //gks.setRetrievalCue( cue, concIdx );
  }

  /** Sets the cue as input to NACS.
   *  @param cue the value to set. */
  public void setGksCue( Feature[] cue )
  {
     gks.setRetrievalCue( cue );
  }

  /** Sets the cue as input to an AMN net.
   *  @param idx index on an AMN network.
   *  @param input the value to set. */
  public void setAmnCue( int idx, Feature[] input )
  {
     amns[ idx ].setInput( input );
  }

  /** Sets the active AMN network.
   *  @param amnIdx the AMN index to set. */
  public void setActiveAmn( int amnIdx )
  {
     activeAmnIdx = amnIdx;
  }

  /** Sets the dimension information of condition chunks.
   *  @param info the information to set. */
  public void setConditionDimInfo( short[][] info )
  {
      conditionDimInfo = info;
  }

  /** pre-training a specific AMN.  */
  public void preTrainAmn( int amnIdx )
  {
      gks.fillInEmForAssimilation();
      amns[amnIdx].update();
  }

  /** Observes and decompose current state. need be refined further.*/
  public void observeState( Feature[] state )
  {
     //if( global.isEMon() )
     //    gks.checkEM();
     curState = new Feature[ state.length ];
     for( int i = 0; i < curState.length; i++ )
          curState[i] = (Feature)(state[ i ].clone());
     if( esChunkOn ) {
         // store current state as a chunk.
         int chunkIdx = gks.insertChunk( 0, 0, global.CHK_STATE,
                                         global.CHUNK_TYPE_LABELS[global.CHK_STATE], curState );
         if( global.isEMon() )
             gks.updateEM( Global.ACS, chunkIdx );
         //System.out.println( "current state = " + gks.getChunk(chunkIdx) );
     }
  }

  /** Performs NACS control action. */
  public void performAction( short[][][] chosenActions )
  {
     // reset result.
     results = null;
     // if RT is on, reset the RT in the previous step.
     if( global.isRtOn() ){
         gks.resetRT();
         amns[ activeAmnIdx ].resetRT();
     }
     // reset strengths in GKS.
     gks.resetStrengths();
     // analyse the NACS control action.
     short[][] action = chosenActions[Global.ACTION_CONTROL];
     short[][] curNormalAction = chosenActions[Global.ACTION_NORMAL];
     int activationMode = (action[ControlActionFormat.ACT_REP]==null) ? 0 :
                          action[ControlActionFormat.ACT_REP][0];
     if( activationMode == ControlActionFormat.ACT_REP_DO_NOTHING )
         return;
     // determine the 'input to' and 'output from' dimensions of NACS.
     nacsActiveInput = determineInputOutputDimensions( 0, action, curNormalAction );
     nacsActiveOutput = determineInputOutputDimensions( 1, action, curNormalAction );
     if( activationMode == ControlActionFormat.ACT_REP_ACT_NACS ){
         performActionEncodeExternalKnowledge( action, curNormalAction );
         performActionSetChunkStrength( action );
         performActionInferenceRetrieval( action );
         performActionAssimilateExpKnowledge( action );
     }
     else{  // reporting.
         performActionReporting( action );
     }
  }

  /** update NACS. need be refined further.
   *  @param chosenActions the chosen actions by ACS.
   *  @param nState the next state.
   *  @param rewards the currently received reinforcements.
   */
  public void update( short[][][][] chosenActions, Feature[] nState, double[] rewards )
  {
      gks.update();
      amns[ activeAmnIdx ].update();
      if( esChunkOn ) {
          // store the current action, current step as a chunk.
          nextState = nState;
          for( int i = 0; i < acsNetNum; i++ ){
               if( chosenActions[i][0][0][0] == Global.EMPTY )
                   continue;
               curAction = new Feature[chosenActions[i][Global.ACTION_NORMAL].length];
               for( int j = 0; j < curAction.length; j++ ){
                    int type = global.getIdnOutputFeatureInfo(i)[j][0];
                    int order = global.getIdnOutputFeatureInfo(i)[j][1];
                    int len = global.getFeatureLength( type, order );
                    String label = global.getFeatureNames( type, order );
                    short[] vals = CommonMethods.copy( chosenActions[i][Global.ACTION_NORMAL][j] );
                    curAction[j] = new Feature( label, (short)type, (short)order, (short)len, vals, global );
               }
               reward = new Feature[1];
               double[] actVals = new double[1];
               actVals[0] = rewards[i];
               String label = global.getFeatureNames( (short)Global.FT_SYS, (short)Global.FT_SYS_REWARD );
               reward[0] = new Feature( label, (short)Global.FT_SYS, (short)Global.FT_SYS_REWARD, actVals, global );
               Feature[] curStep = fetchAcsStep( i );
               int chunkIdx = gks.insertChunk( i, global.EMPTY, global.CHK_ACTION,
                                               global.CHUNK_TYPE_LABELS[ global.CHK_ACTION ], curAction );
               if( global.isEMon() )
                   gks.updateEM( Global.ACS, chunkIdx );
               chunkIdx = gks.insertChunk( i, global.EMPTY, global.CHK_STEP,
                                           global.CHUNK_TYPE_LABELS[global.CHK_STEP], curStep );
               //System.out.println( "current step = " + gks.getChunk(chunkIdx) );
               if( global.isEMon() )
                   gks.updateEM( Global.ACS, chunkIdx );
          }
      }
      // offline training AEM using EM.
      if( aemOn && Math.random() < aemTrainedByEmProb )
          for( int i = 0; i < acsNetNum; i++ )
               offlineTrainAem( i );
  }

  public String testAem( int acsNetIdx )
  {
      return absEMs[ acsNetIdx ].testing();
  }

/************************ PRIVATE METHODS *************************************/
  private Feature[] determineInputOutputDimensions( int type, short[][] controlAction,
                                                    short[][] curNormalAction )
  {
     short[] inDim = (type == 0) ? controlAction[ControlActionFormat.PARA_IN_IN] :
                      controlAction[ControlActionFormat.PARA_OUT_IN];
     short[] outDim = (type == 0) ? controlAction[ControlActionFormat.PARA_IN_OUT] :
                       controlAction[ControlActionFormat.PARA_OUT_OUT];
     Feature[] features;
     if( inDim != null || outDim != null ){
         if( inDim != null ){
             if( type == 0 )
                 nacsActiveInputDimInfo = CommonMethods.copy( inDim );
             else
                 nacsActiveOutputDimInfo = CommonMethods.copy( inDim );
             features = new Feature[ inDim.length ];
             for( int i = 0; i < inDim.length; i++ ){

                  //System.out.println("inDim[i] = " + inDim[i] + "curState len = " + curState.length );

                  features[i] = (Feature)(curState[inDim[i]].clone());
             }
         }
         else {
                if( type == 0 )
                    nacsActiveInputDimInfo = CommonMethods.copy( outDim );
                else
                    nacsActiveOutputDimInfo = CommonMethods.copy( outDim );
                features = new Feature[ outDim.length ];
                for( int i = 0; i < outDim.length; i++ ){
                     short idx = outDim[i];
                     String label = global.getFeatureNames( Global.FT_OUTPUT, idx );
                     int len = global.getFeatureLength( (short)Global.FT_OUTPUT, idx );
                     short[] actVals = CommonMethods.copy( curNormalAction[idx] );
                     features[i] = new Feature( label, (short)Global.FT_OUTPUT, idx, (short)len, actVals, global );
                }
         }
     }
     else{
          short[][] dimInfo = (type == 0) ? global.getGksInputFeatureInfo() :
                              global.getGksOutputFeatureInfo();
          features = new Feature[ dimInfo.length ];
          if( type == 0 )
              nacsActiveInputDimInfo = new short[dimInfo.length];
          else
              nacsActiveOutputDimInfo = new short[dimInfo.length];
          for( int i = 0; i < dimInfo.length; i++ ){
               short dimType = dimInfo[i][0];
               short idx = dimInfo[i][1];
               if( type == 0 )
                   nacsActiveInputDimInfo[i] = idx;
               else
                   nacsActiveOutputDimInfo[i] = idx;
               if( dimType == Global.FT_INPUT )
                   features[i] = (Feature)(curState[idx].clone());
               else{
                   String label = global.getFeatureNames( Global.FT_OUTPUT, idx );
                   int len = global.getFeatureLength( (short)Global.FT_OUTPUT, idx );
                   short[] actVals = CommonMethods.copy( curNormalAction[idx] );
                   features[i] = new Feature( label, (short)Global.FT_OUTPUT, idx, (short)len, actVals, global );
               }
          }
     }
     return features;
  }

  private void performActionEncodeExternalKnowledge( short[][] action, short[][] curNormalAction )
  {
    int encodeMode = ( action[ControlActionFormat.ENCODE_EXT] == null ) ? 0 :
                       action[ControlActionFormat.ENCODE_EXT][0];
    if( encodeMode == ControlActionFormat.ENCODE_EXT_DO_NOTHING )
        return;
    activeAmnIdx = (action[ControlActionFormat.PARA_AMNS]==null) ? 0 :
                    action[ControlActionFormat.PARA_AMNS][0];
    if( encodeMode == ControlActionFormat.ENCODE_EXT_RULES_USER ){
        global.getTaskClarion().performEncodeExternalKnowledge( action );
        return;
    }
    // encode the condition chunks.
    Arrays.sort( nacsActiveInputDimInfo );
    Arrays.sort( nacsActiveOutputDimInfo );
    int conditionNum = conditionDimInfo.length;
    short[] conditions = new short[ conditionNum ];
    for( int i = 0; i < conditionNum; i++ ){
         int featNum = conditionDimInfo[i].length;
         Feature[] features = new Feature[featNum];
         for( int j = 0; j < featNum; j++ ){
              short featIdx = conditionDimInfo[i][j];
              int idx = Arrays.binarySearch( nacsActiveInputDimInfo, featIdx );
              if( idx >= 0 )
                  features[j] = (Feature)(nacsActiveInput[idx].clone());
              else
                  return;
         }
         conditions[i] = (short)gks.getChunkIdx( features );
         if( conditions[i] == global.EMPTY ){
             String label = "chunk " + conditions[i];
             conditions[i] = (short)gks.insertChunk( 0, activeAmnIdx, Global.CHK_ASSOC, label, features );
         }
    }
    // encode the conclusion chunk.
    short conclusion = (short)gks.getChunkIdx( nacsActiveOutput );
    if( conclusion == global.EMPTY ){
        String label = "chunk " + conclusion;
        conclusion = (short)gks.insertChunk( 0, activeAmnIdx, Global.CHK_ASSOC, label, nacsActiveOutput );
    }
    // encode associative rule.
    gks.insertRule( conditions, conclusion );
  }

  private void performActionSetChunkStrength( short[][] action )
  {
    int setStrengthMode = (action[ControlActionFormat.SET_STRENTH]==null) ? 0 :
                          action[ControlActionFormat.SET_STRENTH][0];
    if( setStrengthMode == ControlActionFormat.SET_STRENTH_DO_NOTHING )
        return;
    int chunkIdx = (action[ControlActionFormat.PARA_CHK_SET]==null) ? 0 :
                   action[ControlActionFormat.PARA_CHK_SET][0];
    if( gks.getChunk( chunkIdx ) == null )
        return;
    double strength = (action[ControlActionFormat.PARA_STRENGTH]==null) ? 0.5 :
                      action[ControlActionFormat.PARA_STRENGTH][0] * 0.1;
    gks.setChunkStrength( chunkIdx, strength );
  }

  private void performActionInferenceRetrieval( short[][] action )
  {
    /** retrieval mode. (no retrieval, retrieval_one, retrieval_all). */
    int retrievalMode = (action[ControlActionFormat.INFER_RETR]==null) ? 0 :
                         action[ControlActionFormat.INFER_RETR][0];

    //System.out.println( "retrieval mode = " + retrievalMode );


    if( retrievalMode == ControlActionFormat.INFER_RETR_DO_NOTHING )
         return;
    int levelMode = (action[ControlActionFormat.CHOOSE_LEVEL]==null) ? 0:
                    action[ControlActionFormat.CHOOSE_LEVEL][0];
    activeAmnIdx = (action[ControlActionFormat.PARA_AMNS]==null) ? 0 :
                    action[ControlActionFormat.PARA_AMNS][0];
    // set GKS reasoning method.
    int reasoningMethod = (action[ControlActionFormat.PARA_GKS_RSN_MTHD]==null) ? 0 :
                          action[ControlActionFormat.PARA_GKS_RSN_MTHD][0];
    gks.setGksReasoningMethod( reasoningMethod );
    // set # of iterations of GKS reasoning.
    int num = (action[ControlActionFormat.PARA_GKS_RSN_ITER]==null )? 0 :
               action[ControlActionFormat.PARA_GKS_RSN_ITER][0];
    int iterNum = (num == Global.GKS_ITER_UNLIMIT_IDX) ? Integer.MAX_VALUE : num+1;
    gks.setGksIterNum( iterNum );
    // set # of AMN passes for each iteration.
    int passNum = (action[ControlActionFormat.PARA_AMN_PASS]==null) ? 1 :
                   action[ControlActionFormat.PARA_AMN_PASS][0] + 1;
    amns[ activeAmnIdx ].setPassNum( passNum );
    int combMode = (action[ControlActionFormat.PARA_COMB_MODE]==null) ? 0 :
                   action[ControlActionFormat.PARA_COMB_MODE][0];


    //System.out.println( this.getClass().getName() + ", levelMode = " + levelMode );
    //System.out.println( this.getClass().getName() + ", combMode = " + combMode );


    // set cue for both GKS and AMN.
    this.setGksCue( nacsActiveInput );
    this.setAmnCue( activeAmnIdx, nacsActiveInput );
    switch( levelMode ){
       case ControlActionFormat.CHOOSE_LEVEL_GKS:
            gks.reasoning();
            if( combMode != ControlActionFormat.COMB_DEFAULT ){
                results = gks.getTopLevelMonitorConclusions();
                return;
            }
            break;
       case ControlActionFormat.CHOOSE_LEVEL_AMN:
            amns[ activeAmnIdx ].retrieval( amnMultiFlag );
            // retrieval result of AMN is sent up.
            Feature[] blRetrievals = amns[ activeAmnIdx ].getResults();
            if( combMode == ControlActionFormat.COMB_DEFAULT )
                gks.bottomUp( activeAmnIdx, blRetrievals );
            else{
                results = blRetrievals;

                //System.out.println( "the result len in NACS.java = " + results.length );

                return;
            }
            break;
       case ControlActionFormat.CHOOSE_LEVEL_BOTH:
            if( combMode == ControlActionFormat.COMB_DEFAULT ){
                gks.reasoning();
                amns[ activeAmnIdx ].retrieval( amnMultiFlag );
                // retrieval result of AMN is sent up.
                blRetrievals = amns[ activeAmnIdx ].getResults();
                gks.bottomUp( activeAmnIdx, blRetrievals );
            }
            if( combMode == ControlActionFormat.COMB_USER_BUR )
                global.getTaskClarion().combineByBottomUpRectification();
            if( combMode == ControlActionFormat.COMB_USER_TDG )
                global.getTaskClarion().combineByTopDownGuidance();
            break;
    }
    // get the final results.
    if( combMode == ControlActionFormat.COMB_DEFAULT ){
        short[] chunkTypes = action[ControlActionFormat.PARA_CHK_TYPES];
        int retrievalChunkType = (chunkTypes[0]==ControlActionFormat.CHK_TYPES_ALL) ?
              Global.ALL_CHK_TYPES : (chunkTypes[0]==ControlActionFormat.CHK_TYPES_STATE ||
              chunkTypes[0]==ControlActionFormat.CHK_TYPES_STATE_NOINPUT) ? Global.CHK_STATE:
              Global.CHK_ACTION;
        gks.formResults( retrievalMode, retrievalChunkType );
        results = gks.getCombinedResults();
    }
    else
        results = global.getTaskClarion().getNacsResults();


    //if( combMode == 3 ){
        /*
        System.out.println( "levelMode = " + levelMode );
        System.out.println( "combMode = " + combMode );
        System.out.println( "retrievalMode = " + retrievalMode );
        System.out.println( "reasoningMethod = " + reasoningMethod );
        System.out.println( "activeAmnIdx = " + activeAmnIdx );
        System.out.println( "iterNum = " + iterNum );
        System.out.println( "passNum = " + passNum );
        for( int i = 0; i < nacsActiveInput.length; i++ )
             System.out.println( "nacsActiveInput" + "[" + i + "] = " + nacsActiveInput[i] );
        System.out.println( "results len = " + results.length );
        */
        /*
        System.out.println( "results = " );
        for( int i = 0; i < results.length; i++ )
             System.out.println( results[i] );
        */
    //}
  }

  private void performActionAssimilateExpKnowledge( short[][] action )
  {
    activeAmnIdx = (action[ControlActionFormat.PARA_AMNS]==null) ? 0 :
                   action[ControlActionFormat.PARA_AMNS][0];
    int assimMode = (action[ControlActionFormat.ASSIM_EXP]==null) ? 0 :
                    action[ControlActionFormat.ASSIM_EXP][0];
    amns[ activeAmnIdx ].setAssocAssimOn( assimMode != ControlActionFormat.ASSIM_EXP_DO_NOTHING );
    if( assimMode == ControlActionFormat.ASSIM_EXP_SYS_ASSIM ){
        // AMN is trained right away (for one iteration).
        amns[ activeAmnIdx ].setInput( nacsActiveInput );
        amns[ activeAmnIdx ].setDesiredOutput( nacsActiveOutput );
        amns[ activeAmnIdx ].training();
        //gks.fillInEmForAssimilation();
        // the given association is stored into the EM.
        gks.updateEM( Global.NACS, activeAmnIdx, nacsActiveInput, nacsActiveOutput );
    }
    if( assimMode == ControlActionFormat.ASSIM_EXP_USER_ASSIM )
        global.getTaskClarion().performNacsAssimilation( action );
  }

  private void performActionReporting( short[][] action )
  {
    global.getACS().setRetrievalResult( (GKSChunk[])results );
  }

  /** decompose the step chunk into different parts. */
  private Feature[] fetchAcsStep( int netIdx )
  {
     int dimNum = acsStepFeatNums[ netIdx ];
     Feature[] features = new Feature[dimNum];
     int dimOffset = 0;
     for( int i = 0; i < curState.length; i++ )
         features[ dimOffset+i] = (Feature)(curState[ i ].clone());
     dimOffset += curState.length;
     for( int i = 0; i < curAction.length; i++ )
         features[ dimOffset+i] = (Feature)(curAction[ i ].clone());
     dimOffset += curAction.length;
     if( global.getIdnLearningOption(netIdx) == global.QL ){
         for( int i = 0; i < nextState.length; i++ )
              features[ dimOffset+i] = (Feature)(nextState[ i ].clone());
         dimOffset += nextState.length;
     }

     //System.out.println( "dimNum = " + dimNum + ", dimOffset = " + dimOffset );

     for( int i = 0; i < reward.length; i++ )
          features[ dimOffset+i ] = (Feature)(reward[ i ].clone());
     return features;
  }

  /** offline training AEM using EM. */
  private void offlineTrainAem( int acsNetIdx )
  {
     EMAcsSample[] acsSamples = gks.getEmAcsSamples( acsNetIdx, global.AEM );
     int sampleNum = acsSamples.length;
     int iterNum = global.getAemTrainedByEmIterNum();
     if( sampleNum < global.getAemTrainedByEmItemNum() ){
         //System.out.println("insuffient EM item # training AEM!" );
         return;
     }
     Vector sands = new Vector( sampleNum * iterNum );
     for( int j = 0; j < iterNum; j++ )
       for( int k = 0; k < sampleNum; k++ )
            sands.addElement( new Integer(k) );
     while( !sands.isEmpty() ){
       int idx = ((Integer)(sands.remove((int)(sands.size()*Math.random())))).intValue();
       //System.out.println("current sample = " + acsSamples[idx] );
       absEMs[ acsNetIdx ].offlineTrain( acsSamples[idx] );
     }

     //debugging and testing.
     //for( int i = 0; i < sampleNum; i++ )
     //     System.out.println( acsSamples[i] );
     //absEMs[ acsNetIdx ].testing();
  }
}