package clarion.options.transfer;
import javax.swing.*;
import java.io.*;
import clarion.system.*;
import clarion.options.*;
import clarion.options.transfer.subsys.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is to manage the processes of loading all of CLARION-relevant option
 * settings from the relevant parameter setting files and initializing the relevant
 * global variables and saving user's new configuration to the persistent storage.
 * <li> important methods <br>
 * <ol>
 * <li> public void load(): implement the process of loading options from persistent
 * storage.
 * <li> public void save(): implement the process of save the options to persistent
 * storage or to the associated <tt>Global</tt> object.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model. <br>
 * This class implemnts the process of loading and saving CLARION configurations.
 * <li> Usage of this class. <br>
 * Each <em>TransferManager</em> object is associated with the GUI specific to a
 * single <em>TaskAgent</em> object. This class is instantiated in the class
 * <em>Task</em> or <em>GuiTemplate</em>.
 * </ul>
 * @author Xi Zhang
 *
 **/

public class TransferManager{
  /** constants indicates save options to associated <tt>Global</tt> object or
   * persistent storage */
  protected final int SAVE_TO_GLOBAL = 0, SAVE_TO_FILE = 1;
  /** numbers of components in CLARION and Task-Specific configuration. */
  protected final int TOTAL_COMP_NUM = 8, CLARION_COMP_NUM = 6, TASK_COMP_NUM = 2;
  /** constants indicate the subsystems in configuration of CLARION and task-specific */
  protected final int ACS = 0, NACS = 1, SUPER = 2, RT = 3, DISPLAY = 4, DIMS = 5, TS = 6, SUBSYS = 7;
  /** constants for configuration of ACS subsystem. */
  protected final int ACS_COMP_NUM = 4;
  protected final int ACS_COMP = 0, ACS_NET = 1, ACS_WM = 2, ACS_GS = 3;
  protected final int ACS_NET_OPT_NUM = 4;
  protected final int EX_NET = 0, WM_NET = 1, GS_NET = 2, ACS_COR = 3;
  protected final int ACS_NET_COMP_NUM = 5;
  protected final int IDN = 0, RER = 1, IRL = 2, FR = 3, COMP = 4;
  /** constants for configuration of NACS subsystem. */
  protected final int NACS_COMP_NUM = 6;
  protected final int NACS_COMP = 0, NACS_GKS = 1, NACS_AMN = 2, NACS_EM = 3, NACS_AEM = 4, NACS_COR = 5;
  /** constants for configuration of Supervisory subsystem. */
  protected final int SUPER_COMP_NUM = 3;
  protected final int SUPER_COMP = 0, SUPER_MS = 1, SUPER_MCS = 2;
  /** constants for configuration of Response Time Calculation. */
  protected final int RT_COMP_NUM = 1;
  /** constants for configuration of task specific. */
  protected final int DISPLAY_COMP_NUM = 1;
  /** constants indicates one component or choice is selected or not. */
  protected final int ON = 0, OFF = 1;
  protected Global global;
 // save to global or file
  protected int saveCodes[];
  /** the transfers for transfer data of subsystem. */
  protected Transfer transfers[];
  /** variables for locating files. */
  protected static String sysPath = Global.getSettingDir() + "\\DEFAULT";
  protected static File sysDir = new File( sysPath );
  protected String sysSettingPath = sysPath + "\\subjectGroup0";
  protected File sysSettingDir = new File( sysSettingPath );
  protected String subsysPath;

  /** constructor. instantiated with an associated <tt>Global</tt> object. */
  public TransferManager( Global global )
  {
     this.global = global;
     saveCodes = new int[2];
     // initialize the transfers.
     transfers = new Transfer[TOTAL_COMP_NUM];
     transfers[ACS] = new AcsTransfer( this, global );
     transfers[NACS] = new NacsTransfer( this, global );
     transfers[SUPER] = new SuperTransfer( this, global );
     transfers[RT] = new RtTransfer( this, global );
     transfers[DISPLAY] = new InterTransfer( this, global );
     transfers[DIMS] = new DimsTransfer( this, global );
     transfers[TS] = new TaskSpecTransfer( this, global );
     transfers[SUBSYS] = new SubsysTransfer( this, global );
  }

  /** Get the names of existing simulated tasks.
   * @return the names of tasks. */
  public String[] searchCurrentTasks()
  {
     String path = global.getSettingDir();
     File file = new File( path );
     int len = file.list().length;
     String[] names = new String[ len ];
     for( int i = 0; i < len; i++ )
          names[i] = file.list()[i];
     return names;
  }

  /** Get the names of existing setting group in a simulated task.
   * @return the names of setting groups. */
  public String[] searchCurrentSettings()
  {
     String path = global.getSettingDir() + "\\" + global.getTaskName();
     File file = new File( path );
     int len = file.list().length;
     String[] names = new String[ len ];
     for( int i = 0; i < len; i++ )
          names[i] = file.list()[i];
     return names;
  }

  /** Get an array of particular options.
   * @param dim1 index on a subsystem.
   * arguments <tt>dim2</tt> through <tt>dim6</tt> are all indices on a 6-dimension
   * array which stores all options of a particular subsystem. */
  public Option[] getOptions( int dim1, int dim2, int dim3, int dim4, int dim5, int dim6 )
  {
     return transfers[dim1].getOptions( dim2, dim3, dim4, dim5, dim6);
  }

  /** Get a particular option.
   * @param dim1 index on a subsystem.
   * arguments <tt>dim2</tt> through <tt>dim7</tt> are all indices on a 6-dimension
   * array which stores all options of a particular subsystem. */
  public Option getOptions( int dim1, int dim2, int dim3, int dim4, int dim5, int dim6, int dim7 )
  {
     return transfers[dim1].getOptions( dim2, dim3, dim4, dim5, dim6, dim7 );
  }

  /** Check if NACS subsystem is selected or not. */
  public boolean isNacsOn()
  {
     return ((SubsysTransfer)transfers[SUBSYS]).isNacsOn();
  }

  /** Check if Supervisory subsystem is selected or not. */
  public boolean isSupervisoryOn()
  {
     return ((SubsysTransfer)transfers[SUBSYS]).isSupervisoryOn();
  }

  /** Check if Response Time is selected or not. */
  public boolean isRtOn()
  {
     return ((SubsysTransfer)transfers[SUBSYS]).isRtOn();
  }

  /** Get the number of features in the specified dimension type.
   *  @param type the dimension type.
   *  @return the number of features in the specified dimension type. */
  public int getFeatureNum( int type )
  {
     return ((DimsTransfer)transfers[DIMS]).getFeatureNum( type );
  }

  /** Get the name of a feature in the specified type and order in that type.
   *  @param type the dimension type.
   *  @param idx the order in that type.
   *  @return the name of a feature. */
  public String getFeatureName( int type, int idx )
  {
     return ((DimsTransfer)transfers[DIMS]).getFeatureName( type, idx );
  }

  /** Get the order of a feature in a type with the particular name. return -1 if
   *  the feature is not in the type.
   *  @param type the dimension type.
   *  @param name the feature name.
   *  @return the order of a feature in a type. */
  public int searchFeatureName( int type, String name )
  {
     return ((DimsTransfer)transfers[DIMS]).searchFeatureName( type, name );
  }

  /** Get the Goal Structure type.
   *  0: goal stack, 1: goal list, 2: not selected.
   *  @return the Goal Structure type. */
  public int getGsOption()
  {
     return ((AcsTransfer)transfers[ACS]).getGsOption();
  }

  /** Get the Working Memory option to check if it is selected or not.
   *  @return the Working Memory option. */
  public boolean getWmOption()
  {
     return ((AcsTransfer)transfers[ACS]).getWmOption();
  }

  /** Get the number of ACS external networks.
   *  @return the number of ACS external networks. */
  public int getAcsExNetNum()
  {
     return ((AcsTransfer)transfers[ACS]).getAcsExNetNum();
  }

  /** Get the Goal Structure IDN option to check if the IDN is selected or not.
   *  @return the Goal Structure IDN option. */
  public boolean isGsIdnOn()
  {
     return ((AcsTransfer)transfers[ACS]).isGsIdnOn();
  }

  /** Get the Working Memory IDN option to check if the IDN is selected or not.
   *  @return the Goal Structure IDN option. */
  public boolean isWmIdnOn()
  {
     return ((AcsTransfer)transfers[ACS]).isWmIdnOn();
  }

  /** Get ACS component option in one ACS network to check if the component is
   * selected or not.
   *  @param netType : external, GS or WM network.
   *  @param netidx : the index on one type of networks.
   *  @param type: the component type IDN, RER, IRL or FR.
   *  @return ACS component option. */
  public boolean isAcsCompOn( int netType, int netIdx, int type )
  {
     return ((AcsTransfer)transfers[ACS]).isAcsCompOn( netType, netIdx, type );
  }

  /** Get the number of a specified component allowed in one ACS network.
   *  @param netType : external, GS or WM network.
   *  @param netidx : the index on one type of networks.
   *  @param type: the component type IDN, RER, IRL or FR.
   *  @return the number of a specified component allowed in one ACS network. */
  public int getAcsCompFileNum( int netType, int netIdx, int type )
  {
     return ((AcsTransfer)transfers[ACS]).getAcsCompFileNum( netType, netIdx, type );
  }

  /** Get the GKS option to check if the GKS component is selected or not.
   *  @return the GKS option. */
  public boolean isGksOn()
  {
     return ((NacsTransfer)transfers[NACS]).isGksOn();
  }

  /** Get the number of AMN networks in the NACS subystem.
   *  @return the number of AMN networks in the NACS subystem. */
  public int getAmnNum()
  {
     return ((NacsTransfer)transfers[NACS]).getAmnNum();
  }

  /** Get the EM option to check if the EM component is selected or not.
   *  @return the EM option. */
  public boolean isEmOn()
  {
     return ((NacsTransfer)transfers[NACS]).isEmOn();
  }

  /** Get the AEM option to check if the AEM component is selected or not.
   *  @return the AEM option. */
  public boolean isAemOn()
  {
     return ((NacsTransfer)transfers[NACS]).isAemOn();
  }

  /** Get the type of an AMN network in the NACS subystem.
   *  @return the type of an AMN network in the NACS subystem.
   *  @deprecated */
  public int getAmnType( int netIdx )
  {
     return ((NacsTransfer)transfers[NACS]).getAmnType( netIdx );
  }

  /** Get the GKS reasoning method.
   *  @return the GKS reasoning method. */
  public int getGksReasoningType()
  {
     return ((NacsTransfer)transfers[NACS]).getGksReasoningType();
  }

  /** Get the MS option to check if the MS component is selected or not in supervisory
   * subsystem.
   * @return the MS option. */
  public boolean isMsOn()
  {
     return ((SuperTransfer)transfers[SUPER]).isMsOn();
  }

  /** Get the MCS option to check if the MCS component is selected or not in supervisory
   * subsystem.
   * @return the MCS option. */
  public boolean isMcsOn()
  {
     return ((SuperTransfer)transfers[SUPER]).isMcsOn();
  }

  /** Set the associated <tt>Global</tt> object with this manager.
   * @param global the <tt>Global</tt> object to set. */
  public void setGlobal( Global global )
  {
     this.global = global;
     for( int i = 0; i < transfers.length; i++ )
          transfers[i].setGlobal( global );
  }

  /** Set the mode of saving options:
   * 0: save to <tt>Global</tt> object; 1: save to persistent storage.
   * @param index indicates CLARION or Task-Specific configuration.
   * @param code the mode of saving options. */
  public void setSaveCode( int index, int code )
  {
     saveCodes[index] = code;
     int start = (index == Global.CLARION_OPTIONS) ? 0 : CLARION_COMP_NUM;
     int end = (index == Global.CLARION_OPTIONS) ? CLARION_COMP_NUM : TOTAL_COMP_NUM;
     for( int i = start; i < end; i++ )
          transfers[i].setSaveCode( code );
  }

  /** Load configurations from persistent storage.
   * @param index indicates CLARION or Task-Specific configuration. */
  public void load( int index )
  {
     String taskPath = global.getSettingDir() + "\\" + global.getTaskName();
     File taskDir = new File( taskPath );
     String taskSettingPath = taskPath + "\\" + global.getTaskSettingName();
     File taskSettingDir = new File( taskSettingPath );
     if( !taskDir.exists() ){   // a new task.
         System.out.println( "A new task, copy from DEFAULT opitons." );
         taskDir.mkdir();
         taskSettingDir.mkdir();
         copyFileStructure( sysSettingDir, taskSettingDir );
     }
     else if( !taskSettingDir.exists() ){  // a new settting.
              System.out.println( "A new task settings, copy from an existing setting opitons." );
              taskSettingDir.mkdir();
              String[] settingNames = searchCurrentSettings();
              String existTaskSettingPath = taskPath + "\\" + settingNames[0];
              File existTaskSettingDir = new File( existTaskSettingPath );
              copyFileStructure( existTaskSettingDir, taskSettingDir );
     }

     if( index == Global.CLARION_OPTIONS ) {
         transfers[SUBSYS].load();
         //System.out.println( "after subsys");
         transfers[DIMS].load();
         //System.out.println( "after dims");
         transfers[ACS].load();
         //System.out.println( "after acs");
         transfers[NACS].load();
         //System.out.println( "after nacs");
         transfers[SUPER].load();
         //System.out.println( "after super");
         transfers[RT].load();
         //System.out.println( "after rt");
     }
     else{
         transfers[DISPLAY].load();
         //System.out.println( "after diplay");
         transfers[TS].load();
         //System.out.println( "after TS");
     }
  }

  /** save configurations.
   * @param index indicates CLARION or Task-Specific configuration. */
  public void save( int index )
  {
     if( saveCodes[index] == SAVE_TO_FILE ){
         // create a new user setting directory.
         String path = global.getSettingDir() + "\\" + global.getTaskName();
         File taskDir = new File( path );
         File[] subjectGroups = taskDir.listFiles();
         File newUserDir = new File( path + "\\" + global.getTaskSettingName() );
         newUserDir.mkdir();
     }

     if( index == Global.CLARION_OPTIONS ) {
         // save subsystem-selection settings.
         transfers[SUBSYS].save();
         //System.out.println( "after subsys");
         // save current settings.
         transfers[DIMS].save();
         //System.out.println( "after dims");
         transfers[ACS].save();

         //System.out.println( "In TransManager.java, after ACS");

         if( this.isNacsOn() )
             transfers[NACS].save();
         //System.out.println( "after NACS");
         if( this.isSupervisoryOn() )
             transfers[SUPER].save();
         //System.out.println( "after super");
         if( this.isRtOn() )
             transfers[RT].save();
         //System.out.println( "after RT");
     }
     else{
         transfers[DISPLAY].save();
         //System.out.println( "after display");
         transfers[TS].save();
         //System.out.println( "after TS");
     }

     // update the Global variables or options.
     if( saveCodes[index] == SAVE_TO_GLOBAL ){
         int start = (index == Global.CLARION_OPTIONS) ? 0 : CLARION_COMP_NUM;
         int end = (index == Global.CLARION_OPTIONS) ? CLARION_COMP_NUM : TOTAL_COMP_NUM;
         for( int i = start; i < end; i++ )
              transfers[i].update();
     }
     //System.out.println( "after update");
  }

  private void copyFileStructure( File from, File to )
  {
     if( from.isDirectory() ){
         File[] files = from.listFiles();
         for( int i = 0; i < files.length; i++ ){
              try{
                   File subDir = new File( to, files[i].getName() );
                   boolean succFlag = false;
                   if( files[i].isDirectory() )
                       succFlag = subDir.mkdir();
                   if( files[i].isFile() )
                       succFlag = subDir.createNewFile();
                   //System.out.println( subDir.getAbsolutePath() + ": " + succFlag );
                   copyFileStructure( files[i], subDir );
              }
              catch ( IOException e ) {
                 JOptionPane.showMessageDialog( null, "File Operation Error in creating!",
                      "Invalid File Name", JOptionPane.ERROR_MESSAGE );
              }
        }
     }
     if( from.isFile() ){
         try {
              RandomAccessFile fromFile = new RandomAccessFile( from, "rw" );
              RandomAccessFile toFile = new RandomAccessFile( to, "rw" );
              fromFile.seek( 0 );
              toFile.seek( 0 );
              while( fromFile.getFilePointer() < fromFile.length() )
                     toFile.writeByte( fromFile.readByte() );
              fromFile.close();
              toFile.close();
         }
         catch ( IOException e ) {
                 JOptionPane.showMessageDialog( null, "File Operation Error in copying!",
                      "Invalid File Name", JOptionPane.ERROR_MESSAGE );
         }
     }
  }
}
