package clarion.system;
import java.util.Vector;
import clarion.acs.*;
import clarion.nacs.*;
import clarion.acs.gs.*;
import clarion.acs.wm.*;
import clarion.mcs.*;
import clarion.ms.*;
import clarion.common.*;
import clarion.coordinate.*;
import clarion.tools.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is to control the overall learning and reasoning process of CLARION.
 * <li> summary of the important methods. <br>
 * <ol>
 * <li> public void training() is to implement the process of learning described in
 * CLARION.
 * <li> public void testing() is to test the learning results of the learning agents
 * after the learning process.
 * <li> public void performActions() is to let all of the ACS networks perform their
 * selected actions including the EX,GS,WM networks if exist.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * This class is to implement the overall learning and reasoning process described
 * in CLARION. In the implementation of CLARION model, a CLARION is embedded
 * in each learning agent as its learning & reasoning mechanism, and it is
 * instantiated inside that agent.
 * <li> Usage of this class. <br>
 * All of the functionalities of CLARION model can be achieved by calling the methods
 * in <tt>this</tt> class. This class is only instantiated inside a <tt>TaskAgent</tt>
 * object.
 * </ul>
 * @author Xi Zhang
 *
 */

public class Clarion {

   /** The task agent this Clarion Object is associated with. */
   protected TaskAgent agent;
   /** The global settings this Clarion Object is associated with. */
   protected Global  global;
   /** The task-specific clarion settings the associated task agent has. */
   protected TaskClarion taskClarion;
   /** action-centered subsystem. */
   protected ACS     acs;
   /** non-action-centered subsystem. */
   protected NACS    nacs;
   /** Meta Cognition Subsystem.   */
   protected MCS     mcs;
   /** Motivational Subsystem. */
   protected MS      ms;
   /** drives described in CLARION, used in MS/MCS subsystem. */
   protected Drives  drives;
   /** coordinator for coordinating the External action and Goal Structure action. */
   protected ExGScoordinator coordinator;
   /** action type num including EX, GS and WM. */
   protected int     actionTypeNum;
   /** sensory input to ACS. */
   protected Feature sensoryInput[];
   /** input to NACS.   */
   protected Feature nacsInput[];
   /** desired output depending specific tasks.  */
   protected Feature desiredOutput[][];
   /** the selected actions from ACS. */
   protected short[] chosenActions[][][];
   /** reinforcement for each ACS network. */
   protected double  reinforcements[];
   /** retrieval results from GKS. */
   //protected GKSChunk retrievalResults[];
   protected Object retrievalResults[];

   /** Class constructor for specifying the associated global settings and task agent */
   public Clarion( Global g, TaskAgent taskAgent )
   {
      global = g;
      global.setClarion( this );
      agent = taskAgent;
      taskClarion = global.getTaskClarion();
      // nacs.
      if( global.isNacsOn() )
          nacs = new NACS( g );
      // acs.
      acs = new ACS( g );
      // initialize supervisory subsystem.
      if( global.isSupervisorySubsysOn() ){
          if( global.isMcsOn() ){
              drives = new Drives();
              mcs = new MCS( acs, nacs, drives, global );
          }
          if( global.isMsOn() )
              ms = new MS( global, global.getGS() );
      }
      // initialize coordinator.
      coordinator = new ExGScoordinator( agent, global );
      actionTypeNum = global.getAcsNetNum();
      chosenActions = new short[ actionTypeNum ][2][][];
      for( int i = 0; i< actionTypeNum; i++ ){
           if( !global.isExNet( i ) )
               chosenActions[i][Global.ACTION_NORMAL] = new short[ global.getIdnOutputDimNum(i) ][];
           else
               chosenActions[i][Global.ACTION_NORMAL] = new short[ global.getFeatureNums()[Global.FT_OUTPUT] ][];
          chosenActions[i][Global.ACTION_CONTROL] = new short[Global.ACT_DIM_NUM_NACS_CONTROL][];
      }
      reinforcements = new double[ actionTypeNum ];
   }

   //=========================== Get and Set METHODS ==========================/
   /** Returns the ACS part in this <em>Clarion</em> object. */
   public ACS getACS()
   {
      return acs;
   }

   /** Returns the NACS part in this <em>Clarion</em> object. */
   public NACS getNACS()
   {
      return nacs;
   }

   /** Returns current sensory input. */
   public Feature[] getSensoryInput()
   {
      return sensoryInput;
   }

   /** Returns the RER rules in the foramt of string array generated in the specified network in ACS.
       @param netIdx the specified network in ACS.
       @return the RER rules in the foramt of string array generated in the specified network in ACS. */
   public String[] getRerRules( int netIdx )
   {
      return acs.getRerRules( netIdx );
   }

   /** Returns the IRL rules generated in the specified network in ACS.
       @param netIdx the specified network in ACS.
       @return the IRL rules generated in the specified network in ACS in the foramt of string array. */
   public String[] getIrlRules( int netIdx )
   {
      return acs.getIrlRules( netIdx );
   }

   /** Returns the FR Rules generated in the specified network in ACS.
       @param netIdx the specified network in ACS.
       @return the FR rules generated in the specified network in ACS in the foramt of string array. */
   public String[] getFixRules( int netIdx )
   {
      return acs.getFixRules( netIdx );
   }

   /** Returns the response time at current step. */
   public int getRT()
   {
      int RT = 0;
      for( int i = 0; i < actionTypeNum; i++ ){
           if( acs.isEligibleExNet(i) )
           { // an external action.
               if( global.isNacsOn() ) { // action for controlling NACS
                   RT = acs.getRT( nacs.getRT() );
               }
               else   // normal ACS action.
                   RT = acs.getRT();
           }
      }
      return RT;
   }

   /** Returns the retrieval results from NACS at current step. */
   //public GKSChunk[] getRetrievalResults()
   public Object[] getRetrievalResults()
   {
      boolean retVal = this.isRetrievalAction();
      //System.out.println( "retrieval action = " + retVal );
      if( retVal )
          return retrievalResults;
      return null;
   }

   /** Set current time used for calculating the response time */
   public void setCurrentTime( long curTime )
   {
      acs.setCurrentTime( curTime );
      if( global.isNacsOn() )
          nacs.setCurrentTime( curTime );
   }

   public void setNacsConditionDimInfo( short[][] conditionInfo )
   {
      nacs.setConditionDimInfo( conditionInfo );
   }

  /** Implement the process of learning described in CLARION. */
   public void training()
   {
      sensoryInput = agent.getCurSensoryInput();
      acs.observeState( sensoryInput );
      if( !agent.isInLearningTest() && global.isNacsOn() ){
          //nacsInput = agent.getCurNacsInput();
          //nacs.observeState( nacsInput );
          nacs.observeState( sensoryInput );
      }
      acs.reasoning();
      //CommonMethods.resetArray( chosenActions, Global.EMPTY );
      chosenActions  = acs.getChosenActions();
      performActions();
      // controlled by MCS, ( may be revised ).
      if( global.isMcsOn() )
          fillInMonitorBuffer();
      // update subsystems.
      sensoryInput = agent.getCurSensoryInput();
      desiredOutput = agent.getCurDesiredOutput();
      acs.setCurDesiredOutput( desiredOutput );
      if( !agent.isInLearningTest() ){
          acs.update( sensoryInput, reinforcements );
          if( global.isNacsOn() ){
              if( isRetrievalAction() )
                  retrievalResults = nacs.getResults();
              nacsInput = agent.getCurNacsInput();
              nacs.update( chosenActions, nacsInput, reinforcements );
          }
      }
   }

   /** Test the learning results of the learning agents after the learning process. */
   public void testing()
   {
      sensoryInput = agent.getCurSensoryInput();
      acs.observeState( sensoryInput );
      if( !agent.isInLearningTest() && global.isNacsOn() ){
          //nacsInput = agent.getCurNacsInput();
          //nacs.observeState( nacsInput );
          nacs.observeState( sensoryInput );
      }
      acs.reasoning();
      //CommonMethods.resetArray( chosenActions, Global.EMPTY );
      chosenActions  = acs.getChosenActions();
      performActions();
      // controlled by MCS, ( may be revised ).
      if( global.isMcsOn() )
          fillInMonitorBuffer();
      //if( global.isNacsOn() )
      //    retrievalResults = nacs.getResults();
      if( !agent.isInLearningTest() ){
          acs.update( sensoryInput, reinforcements );
          if( global.isNacsOn() ){
              if( isRetrievalAction() )
                  retrievalResults = nacs.getResults();
              nacsInput = agent.getCurNacsInput();
              nacs.update( chosenActions, nacsInput, reinforcements );
          }
      }
   }

   public String testAem( int acsNetIdx )
   {
      return nacs.testAem( acsNetIdx );
   }

   /** Let all of the ACS networks perform their selected actions including the EX,GS,WM
       networks if exist. */
   public void performActions()
   {
      Chunk[] wmChunks = global.isWmOn() ? global.getWM().getItems():null;
      GoalChunk goal = global.getGsOption()!=Global.GS_NON ? global.getGS().getActiveItem() : null;
      //all ACS networks perform their actions including EX,GS,WM if exist.
      coordinator.setChosenActions( chosenActions );
      coordinator.coordinate();
      Feature[] newInput = agent.getCurSensoryInput();
      Chunk[] newWmChunks = global.isWmOn() ? global.getWM().getItems() : null;
      GoalChunk newGoal = global.getGsOption()!=Global.GS_NON ? global.getGS().getActiveItem() : null;
      for( int i = 0; i < actionTypeNum; i++ )
           if( acs.isEligibleExNet(i) ){
               reinforcements[i] = taskClarion.getReinforcement( i, sensoryInput, wmChunks, goal,
                                   chosenActions[i], newInput, newWmChunks, newGoal);
           }
   }

   /************************ Get and Set METHODS ******************************/
   private void fillInMonitorBuffer()
   {
      MonitorBuf buffer = global.getMCS().getMonitorBuffer();
      Chunk[][][] acsConclusions = acs.getAllConclusions();
      for( int i = 0; i < acsConclusions.length; i++ )
        for( int j = 0; j < acsConclusions[i].length; j++ )
             buffer.setCurAcsConclusions( i, j, acsConclusions[i][j] );
      if( global.isNacsOn() && isRetrievalAction() ){
          Chunk[][] nacsConclusions = nacs.getAllConclusions();
          for( int i = 0; i < nacsConclusions.length; i++ )
               buffer.setCurNacsConclusions( i, nacsConclusions[i] );
      }
   }

   private boolean isRetrievalAction()
   {
      boolean retVal = false;
      for( int i = 0; i < chosenActions.length; i++ ){
           int CTRL_IDX = Global.ACTION_CONTROL;
           int ACT_IDX = ControlActionFormat.ACT_REP;
           int INFER_IDX = ControlActionFormat.INFER_RETR;
           if( global.isExNet(i) && chosenActions[i][CTRL_IDX][ACT_IDX][0] != ControlActionFormat.ACT_REP_DO_NOTHING &&
               chosenActions[i][CTRL_IDX][INFER_IDX][0] != ControlActionFormat.INFER_RETR_DO_NOTHING ){
               retVal = true;
               break;
           }
      }
      return retVal;
   }
}
