package clarion.tools;
import java.lang.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is to implement the process of action-decision making using Boltzman
 * probability distribution to choose one action among all of its elements.
 * <li> summary of the important methods. <br>
 * <ol>
 * <li> public short[] getRecommendation( double[] els, short[] dimVals ):
 * returns the action recommendation given the arrray of action activations
 * and the array of value number info of each dimension.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * This class implements the process of action decision by Boltzman probability
 * distribution which is the basic method for action decision-making. See tutorial
 * on its details.
 * <li> Usage of this class. <br>
 * It is instantiated inside any class where there needs a stochastic action
 * decider to make an action decision.
 * </ul>
 * @author Xi Zhang
 *
 */

public class StochasticDecider
{
    /** the size of elements in this decider. */
    protected int size = 0;
    //protected double threshold = 0.0005;
    /** threshold. */
    protected double threshold = 0.0;
    /** temperature for action-decision making. */
    protected double temperature;
    protected double BoltzmannSum = 0;
    /** the array of elements. */
    protected double[] elementsArr; // used for const num of elems

    public StochasticDecider()
    {
    }

    /** constructor.
     *  @param nSize the size of elements.
     *  @param temper the temperature of this decider. */
    public StochasticDecider( int nSize, double temper )
    {
        size = nSize;
        temperature = temper;
        elementsArr = new double[ nSize ];
    }

    /** constructor.
     *  @param temper the temperature of the decider. */
    public StochasticDecider( double temper )
    {
        temperature = temper;
    }

    /** Sets the size of elments in this decider.
     *  @param length the size to set. */
    public void setSize( int length )
    {
        size = length;
    }

    /** Sets the temperature of this decider.
     *  @param temper the temperature to set. */
    public void setTemperature( double temper )
    {
        temperature = temper;
    }

    /** Sets the elements in this decider.
     *  @param els the elements to set. */
    public void setElements( double[] els )
    {
        BoltzmannSum = 0;
        for ( int i = 0; i < size; i++ ) {
              elementsArr[i] = Math.exp( ( els[i] ) / temperature );
              BoltzmannSum += elementsArr[i];
        }
    }

    /** Returns the action recommendation.
     *  @return the action recommendation. */
    public int getRecommendation()
    {
        int decision = 0;
        double lottery = Math.random() * BoltzmannSum;
        double currentSum = elementsArr[0];

        while( ( lottery - currentSum ) > threshold )
        {
           decision++;
           currentSum += elementsArr[decision];
        }
        return decision;
    }

    /** Returns the action recommendation given the array of action activations.
     *  @param els  the array of action activations.
     *  @return the index of the recommended action. */
    public int getRecommendation( double[] els )
    {
        size = els.length;
        double[] inputArr = new double[ size ];

        // fill in the input array of Boltzmann distribution.
        BoltzmannSum = 0;
        for ( int i = 0; i < size; i++ )
        {
           inputArr[i] = Math.exp( ( els[i] ) / temperature );
           BoltzmannSum += inputArr[i];
        }

        // make a decision according to Boltzmann distribution.
        int decision = 0;
        double lottery = Math.random() * BoltzmannSum;
        double currentSum = inputArr[0];
        while( ( lottery - currentSum ) > threshold ){
               decision++;
               currentSum += inputArr[ decision ];
        }

        return decision;
    }

    /** Returns the action recommendation given the array of action activations
     *  and a limit on the length of elements in the array from position 0 those
     *  are eligible candidates.
     *  @param els the array of action activations.
     *  @param len the length of eligible candidates.
     *  @return the index of the recommended action. */
    public int getRecommendation( double[] els, int len )
    {
        size = len;
        double[] inputArr = new double[ size ];

        // fill in the input array of Boltzmann distribution.
        BoltzmannSum = 0;
       	for ( int i = 0; i < size; i++ ){
         	    inputArr[i] = Math.exp( ( els[i] ) / temperature );
         	    BoltzmannSum += inputArr[i];
        }

        // make a decision according to Boltzmann distribution.
        int decision = 0;
        double lottery = Math.random() * BoltzmannSum;
        double currentSum = inputArr[0];
        while( ( lottery - currentSum ) > threshold ) {
               decision++;
               currentSum += inputArr[ decision ];
        }

        return decision;
    }

    /** Returns the action recommendation given the array of action activations
     *  and a limit on start position and end position in the array within the
     *  range the elements are eligible candidates.
     *  @param els  the array of action activations.
     *  @param start the start position.
     *  @param end the end position.
     *  @return the index of the recommended action which is the offset from start. */
    public int getRecommendation( double[] els, int start, int end )
    {
        size = end - start;
        double[] inputArr = new double[ size ];
        // fill in the input array of Boltzmann distribution.
        BoltzmannSum = 0;
        for( int i = start; i < end; i++ ){
            inputArr[i-start] = Math.exp( ( els[i] ) / temperature );
            BoltzmannSum += inputArr[i-start];
        }
        // make a decision according to Boltzmann distribution.
        int decision = 0;
        double lottery = Math.random() * BoltzmannSum;
        double currentSum = inputArr[0];
        while( ( lottery - currentSum ) > threshold ){
            decision++;
            currentSum += inputArr[ decision ];
        }
        return decision;
    }

   /** Returns the action recommendation given the arrray of action activations
    *  and the array of value number info of each dimension.
    *  @param els the array of action activations.
    *  @param dimVals the array of value number info of each dimension.
    *  @return an array of indices of recommended value in each dimension. */
    public short[] getRecommendation( double[] els, short[] dimVals  )
    {
       short[] decision = new short[ dimVals.length ];
       int offset = 0;
       for( int i = 0; i < decision.length; i++ ){
            decision[i] = (short)getRecommendation( els, offset, offset + dimVals[i] );
            offset += dimVals[i];
       }
       return decision;
    }

    /** Returns action recommendation with multiple active values in some
     *  dimensions. needs be further implemented.  */
    public short[][] getMultiRecommendation( double[] els, short[] dimVals  )
    {
       return null;
    }
}